/* meko
 * Copyright (C) 2001 Scott James Remnant <scott@netsplit.com>.
 * All Rights Reserved.
 *
 * meko-connection-chooser.c
 *  - MekoConnectionChooser class
 *  - A dialog box that displays a pretty splash screen-like logo and gives
 *    the user an opportunity to choose which saved Say2 connection they wish
 *    to connect to
 * -- 
 * $Id: meko-connection-chooser.c,v 1.7.4.5 2001/11/24 01:25:51 scott Exp $
 * 
 * This file is distributed according to the GNU General Public
 * License.  For full details, read the top of 'main.c' or the
 * file called COPYING that was distributed with this code.
 */

#include <meko.h>

#include <glib.h>
#include <gtk/gtkobject.h>
#include <gtk/gtksignal.h>
#include <gtk/gtkwidget.h>
#include <gtk/gtkbutton.h>
#include <gtk/gtktogglebutton.h>
#include <gtk/gtkentry.h>
#include <gtk/gtkoptionmenu.h>
#include <gtk/gtkmenu.h>
#include <gtk/gtkmenuitem.h>
#include <libgnome/gnome-defs.h>
#include <libgnome/gnome-i18n.h>
#include <libgnome/gnome-util.h>
#include <libgnomeui/gnome-dialog.h>
#include <libgnomeui/gnome-stock.h>
#include <glade/glade-xml.h>
#include <say2/say2.h>

#include "meko-connection-chooser.h"


/* Signals */
enum {
	CONSTRUCT,
	LAST_SIGNAL
};


/* Forward prototypes for required functions */
static void meko_connection_chooser_init(MekoConnectionChooser *);
static void meko_connection_chooser_class_init(MekoConnectionChooserClass *);

/* Forward prototypes for GtkObject methods */
static void meko_connection_chooser_set_arg(GtkObject *, GtkArg *, guint);
static void meko_connection_chooser_get_arg(GtkObject *, GtkArg *, guint);
static void meko_connection_chooser_destroy(GtkObject *);

/* Forward prototypes for MekoConnectionChooser methods */
static void meko_real_connection_chooser_construct(MekoConnectionChooser *);

/* Forward prototypes for MekoConnectionChooser method wrappers */
static void meko_connection_chooser_construct(MekoConnectionChooser *);

/* Forward prototypes for internal functions */
static void _connection_active_cb(GtkMenuItem *, MekoConnectionChooser *);


/* Our parent class */
static GnomeDialogClass *parent_class = NULL;

/* Signal array */
static guint chooser_signals[LAST_SIGNAL] = { 0 };


/* meko_connection_chooser_get_type
 * Returns our GtkType id, allocating it if necessary
 */
guint
meko_connection_chooser_get_type(void)
{
	static guint chooser_type = 0;

	if (!chooser_type) {
		static const GtkTypeInfo chooser_info = {
			"MekoConnectionChooser",
			sizeof(MekoConnectionChooser),
			sizeof(MekoConnectionChooserClass),
			(GtkClassInitFunc)meko_connection_chooser_class_init,
			(GtkObjectInitFunc)meko_connection_chooser_init,
			/* reserved_1 */ NULL,
			/* reserved_2 */ NULL,
			(GtkClassInitFunc)NULL
		};

		chooser_type = gtk_type_unique(GNOME_TYPE_DIALOG,
					       &chooser_info);
	}

	return chooser_type;
}

/* meko_connection_chooser_class_init
 * Initialise the object class, setting our arguments, signals and overriding
 * methods of parent object classes
 */
static void
meko_connection_chooser_class_init(MekoConnectionChooserClass *klass)
{
	GtkObjectClass *object_class;

	g_return_if_fail(klass != NULL);

	object_class = (GtkObjectClass *)klass;
	parent_class = gtk_type_class(GNOME_TYPE_DIALOG);

	chooser_signals[CONSTRUCT] = gtk_signal_new(
		"construct", GTK_RUN_FIRST, object_class->type,
		GTK_SIGNAL_OFFSET(MekoConnectionChooserClass, construct),
		gtk_marshal_NONE__NONE,
		GTK_TYPE_NONE, 0
	);

	gtk_object_class_add_signals(object_class, chooser_signals,
				     LAST_SIGNAL);

	object_class->set_arg = meko_connection_chooser_set_arg;
	object_class->get_arg = meko_connection_chooser_get_arg;
	object_class->destroy = meko_connection_chooser_destroy;

	klass->construct = meko_real_connection_chooser_construct;
}

/* meko_connection_chooser_init
 * Initialise an object instance, set default values and emit the construct
 * signal to have the dialog itself built
 */
static void
meko_connection_chooser_init(MekoConnectionChooser *chooser)
{
	g_return_if_fail(chooser != NULL);

	chooser->hostname = NULL;
	chooser->port = NULL;

	chooser->entry_id = NULL;
	chooser->entry_password = NULL;

	/* Set the window stuff */
	gtk_window_set_title(GTK_WINDOW(chooser), _("Welcome to Meko!"));
	gtk_window_set_wmclass(GTK_WINDOW(chooser), "connection-chooser",
			       "Meko");
	gtk_window_set_policy(GTK_WINDOW(chooser), FALSE, FALSE, TRUE);
	gtk_window_set_position(GTK_WINDOW(chooser), GTK_WIN_POS_CENTER);
	
	/* Add the buttons we want */
	gnome_dialog_append_button(GNOME_DIALOG(chooser),
				   GNOME_STOCK_BUTTON_OK);
	gnome_dialog_append_button(GNOME_DIALOG(chooser),
				   GNOME_STOCK_BUTTON_CLOSE);
	gnome_dialog_set_default(GNOME_DIALOG(chooser), 0);
	
	/* Call construct to get stuff from Glade and stuff it in */
	meko_connection_chooser_construct(chooser);
}

/* meko_connection_chooser_new
 * Create a new object intsance, just a wrapper around gtk_type_new really
 */
GtkWidget *
meko_connection_chooser_new(void)
{
	GtkWidget *widget;

	widget = GTK_WIDGET(gtk_type_new(MEKO_TYPE_CONNECTION_CHOOSER));

	return widget;
}

/* meko_connection_chooser_destroy
 * Destroy an object instance, freeing and unreferencing anydata we've
 * been hanging on to
 */
static void
meko_connection_chooser_destroy(GtkObject *object)
{
	MekoConnectionChooser *chooser;

	g_return_if_fail(object != NULL);
	g_return_if_fail(MEKO_IS_CONNECTION_CHOOSER(object));

	chooser = MEKO_CONNECTION_CHOOSER(object);

	/* Chain up */
	if (GTK_OBJECT_CLASS(parent_class)->destroy)
		GTK_OBJECT_CLASS(parent_class)->destroy(object);
}

/* meko_connection_chooser_set_arg
 * Set object arguments
 */
static void
meko_connection_chooser_set_arg(GtkObject *object, GtkArg *arg, guint arg_id)
{
	MekoConnectionChooser *chooser;

	g_return_if_fail(object != NULL);
	g_return_if_fail(MEKO_IS_CONNECTION_CHOOSER(object));

	chooser = MEKO_CONNECTION_CHOOSER(object);
	switch (arg_id) {
	}
}

/* meko_connection_chooser_get_arg
 * Get object arguments
 */
static void
meko_connection_chooser_get_arg(GtkObject *object, GtkArg *arg, guint arg_id)
{
	MekoConnectionChooser *chooser;

	g_return_if_fail(object != NULL);
	g_return_if_fail(MEKO_IS_CONNECTION_CHOOSER(object));

	chooser = MEKO_CONNECTION_CHOOSER(object);
	switch (arg_id) {
	default:
		arg->type = GTK_TYPE_INVALID;
		break;
	}
}

/* meko_real_connection_chooser_construct
 * Construct the dialog from the Glade source
 */
static void
meko_real_connection_chooser_construct(MekoConnectionChooser *chooser)
{
	GladeXML  *gui;
	GtkWidget *widget;
	gchar	  *filename;

	g_return_if_fail(chooser != NULL);

	/* Load the interface from the MEKO_GLADEDIR directory */
	filename = g_concat_dir_and_file(MEKO_GLADEDIR,
					 "connection-chooser.glade");
	gui = glade_xml_new(filename, NULL);
	g_free(filename);

	/* Grab the vbox from the glade interface, and stuff it into ourselves
	 */
	widget = glade_xml_get_widget(gui, "dialog-vbox");
	if (widget != NULL) {
		gtk_widget_ref(widget);
		gtk_container_remove(GTK_CONTAINER(widget->parent), widget);
		gtk_box_pack_start(GTK_BOX(GNOME_DIALOG(chooser)->vbox), widget,
				   TRUE, TRUE, 0);
		gtk_widget_unref(widget);
		gtk_widget_show(widget);
	}

	/* Create the option menu entries */
	widget = glade_xml_get_widget(gui, "optionmenu-connection");
	if ((widget != NULL) && GTK_IS_OPTION_MENU(widget)) {
		GtkWidget *menu, *item;

		/* Create the popup menu and attach signals to everything
		 * so we can get the values.  Sometimes I think the GTK+
		 * programmers set out to deliberately make life difficult
		 */
		menu = gtk_menu_new();

		/* Lag.net is where it's at */
		item = gtk_menu_item_new_with_label("say2.lag.net");
		gtk_object_set_data(GTK_OBJECT(item), "hostname",
				    "say2.lag.net");
		gtk_object_set_data(GTK_OBJECT(item), "port", "3797");
		gtk_signal_connect(GTK_OBJECT(item), "activate",
				   GTK_SIGNAL_FUNC(_connection_active_cb),
				   (gpointer)chooser);
		gtk_menu_append(GTK_MENU(menu), item);
		gtk_widget_show(item);

		/* This doesn't get done by the set_history below */
		gtk_menu_item_activate(GTK_MENU_ITEM(item));

		/* Let them connect to localhost too :) */
		item = gtk_menu_item_new_with_label("localhost");
		gtk_object_set_data(GTK_OBJECT(item), "hostname", "localhost");
		gtk_object_set_data(GTK_OBJECT(item), "port", "3797");
		gtk_signal_connect(GTK_OBJECT(item), "activate",
				   GTK_SIGNAL_FUNC(_connection_active_cb),
				   (gpointer)chooser);
		gtk_menu_append(GTK_MENU(menu), item);
		gtk_widget_show(item);

		gtk_option_menu_set_menu(GTK_OPTION_MENU(widget), menu);
		gtk_option_menu_set_history(GTK_OPTION_MENU(widget), 0);
	}

	/* Get the other two widgets out and save them */
	widget = glade_xml_get_widget(gui, "entry-id");
	if ((widget != NULL) && GTK_IS_ENTRY(widget))
		chooser->entry_id = widget;
	widget = glade_xml_get_widget(gui, "entry-password");
	if ((widget != NULL) && GTK_IS_ENTRY(widget))
		chooser->entry_password = widget;
	widget = glade_xml_get_widget(gui, "checkbutton-ssl");
	if ((widget != NULL) && GTK_IS_TOGGLE_BUTTON(widget))
		chooser->checkbox_ssl = widget;

#ifdef HAVE_OPENSSL
        if (!socket_ssl_capable())
#endif
                gtk_widget_set_sensitive(chooser->checkbox_ssl, FALSE);

	/* Don't need glade anymore */
	gtk_object_unref(GTK_OBJECT(gui));

	/* Focus the first entry box */
	gtk_widget_grab_focus(chooser->entry_id);
}

/* meko_connection_chooser_construct
 * Emit the construct signal
 */
static void
meko_connection_chooser_construct(MekoConnectionChooser *chooser)
{
	g_return_if_fail(chooser != NULL);

	gtk_signal_emit(GTK_OBJECT(chooser), chooser_signals[CONSTRUCT]);
}

/* meko_connection_chooser_get_selection
 * Return the values of all the entry boxes and options
 */
void
meko_connection_chooser_get_selection(MekoConnectionChooser *chooser,
				      gchar **hostname, gchar **port,
				      gchar **nick, gchar **password)
{
	g_return_if_fail(chooser != NULL);

	/* Easy stuff, it's stored in the object */
	*hostname = chooser->hostname;
	*port = chooser->port;

	/* Grab the current textual values out of the entrys */
	if (chooser->entry_id != NULL) {
		*nick = gtk_entry_get_text(GTK_ENTRY(chooser->entry_id));
	} else {
		*nick = NULL;
	}
	if (chooser->entry_password != NULL) {
		*password = gtk_entry_get_text(
			GTK_ENTRY(chooser->entry_password)
		);
	} else {
		*password = NULL;
	}
}

gboolean
meko_connection_chooser_get_ssl_flag(MekoConnectionChooser *chooser)
{
	g_return_val_if_fail(chooser != NULL, FALSE);

	if (chooser->checkbox_ssl == NULL)
		return FALSE;

        return gtk_toggle_button_get_active(
		GTK_TOGGLE_BUTTON(chooser->checkbox_ssl)
	);
}

/* _connection_active_cb
 * Update the structure with the settings of the menu item they just
 * selected
 */
static void
_connection_active_cb(GtkMenuItem *item, MekoConnectionChooser *chooser)
{
	g_return_if_fail(item != NULL);
	g_return_if_fail(chooser != NULL);

	chooser->hostname = gtk_object_get_data(GTK_OBJECT(item), "hostname");
	chooser->port = gtk_object_get_data(GTK_OBJECT(item), "port");
}
