/* meko
 * Copyright (C) 2001 Scott James Remnant <scott@netsplit.com>.
 * All Rights Reserved.
 *
 * meko-main-window.c
 *  - MekoMainWindow class
 *  - A GnomeApp derived window containing a MekoChatInterface which doubles
 *    for both a server window and a channel window.
 * -- 
 * $Id: meko-main-window.c,v 1.3.4.2 2001/11/24 00:58:42 scott Exp $
 * 
 * This file is distributed according to the GNU General Public
 * License.  For full details, read the top of 'main.c' or the
 * file called COPYING that was distributed with this code.
 */

#include <meko.h>

#include <glib.h>
#include <gdk/gdk.h>
#include <gtk/gtkmain.h>
#include <gtk/gtkobject.h>
#include <gtk/gtksignal.h>
#include <gtk/gtkwidget.h>
#include <gtk/gtkcontainer.h>
#include <libgnome/gnome-defs.h>
#include <libgnome/gnome-i18n.h>
#include <libgnomeui/gnome-app.h>

#include "say2/say2-server.h"
#include "say2/say2-client.h"
#include "util/say2/client-label.h"
#include "widgets/client-list/meko-client-list.h"
#include "widgets/chat-interface/meko-chat-interface.h"
#include "gui/private-window/meko-private-window.h"

#include "meko-main-window.h"

#define IDLE_TIME 300

/* Forward prototypes for required functions */
static void meko_main_window_init(MekoMainWindow *);
static void meko_main_window_class_init(MekoMainWindowClass *);

/* Forward prototypes for GtkObject methods */
static void meko_main_window_set_arg(GtkObject *, GtkArg *, guint);
static void meko_main_window_get_arg(GtkObject *, GtkArg *, guint);
static void meko_main_window_destroy(GtkObject *);

/* Forward prototypes for internal functions */
static void _server_authenticated_cb(Say2Server *, MekoMainWindow *);
static void _server_disconnected_cb(Say2Server *, MekoMainWindow *);
static void _client_list_activate_cb(MekoClientList *, Say2Client *,
				     MekoMainWindow *);
static void _private_window_destroy_cb(MekoPrivateWindow *, Say2Client *);
static gint _idle_cb(MekoMainWindow *);
static void _entry_activate_cb(GtkWidget *, MekoMainWindow *);

/* Our parent class */
static GnomeAppClass *parent_class = NULL;


/* meko_main_window_get_type
 * Returns our GtkType id, allocating it if necessary
 */
guint
meko_main_window_get_type(void)
{
	static guint main_window_type = 0;

	if (!main_window_type) {
		static const GtkTypeInfo main_window_info = {
			"MekoMainWindow",
			sizeof(MekoMainWindow),
			sizeof(MekoMainWindowClass),
			(GtkClassInitFunc)meko_main_window_class_init,
			(GtkObjectInitFunc)meko_main_window_init,
			/* reserved_1 */ NULL,
			/* reserved_2 */ NULL,
			(GtkClassInitFunc)NULL
		};

		main_window_type = gtk_type_unique(GNOME_TYPE_APP,
						   &main_window_info);
	}

	return main_window_type;
}

/* meko_main_window_class_init
 * Initialise the object class, overriding methods of parent object classes
 */
static void
meko_main_window_class_init(MekoMainWindowClass *klass)
{
	GtkObjectClass *object_class;

	g_return_if_fail(klass != NULL);

	object_class = (GtkObjectClass *)klass;
	parent_class = gtk_type_class(GTK_TYPE_OBJECT);

	object_class->set_arg = meko_main_window_set_arg;
	object_class->get_arg = meko_main_window_get_arg;
	object_class->destroy = meko_main_window_destroy;
}

/* meko_main_window_init
 * Initialise an object instance, setting up window attributes and placing
 * a ChatInterface inside
 */
static void
meko_main_window_init(MekoMainWindow *main_window)
{
	GtkWidget *entry;

	g_return_if_fail(main_window != NULL);

	/* Constuct the application */
	gnome_app_construct(GNOME_APP(main_window), "meko", _("Meko"));
	
	/* Set the window stuff */
	gtk_window_set_wmclass(GTK_WINDOW(main_window),
			       "main-window", "Meko");
	gtk_window_set_policy(GTK_WINDOW(main_window), FALSE, TRUE, FALSE);
	gtk_window_set_default_size(GTK_WINDOW(main_window), 600, 300);
	gtk_container_set_border_width(GTK_CONTAINER(main_window), 2);

	/* Create the interface inside */
	main_window->interface = meko_chat_interface_new(NULL, NULL);
	meko_chat_interface_recv_msg_set(
		MEKO_CHAT_INTERFACE(main_window->interface),
		SAY2_MSG_CONNECTING,
		SAY2_MSG_CONNECTION_FAILED,
		SAY2_MSG_CONNECTED,
		SAY2_MSG_NEED_LOGIN,
		SAY2_MSG_DISCONNECTED,
		SAY2_MSG_ERROR,
		SAY2_MSG_BAD_ENAML,
		SAY2_MSG_BAD_COMMAND,
		SAY2_MSG_BAD_NICKNAME,
		SAY2_MSG_BAD_ADDRESS,
		SAY2_MSG_SERVER_INFO,
		SAY2_MSG_MOTD,
		SAY2_MSG_WELCOME,
		SAY2_MSG_LOGOUT,
		SAY2_MSG_REPLICANT_NEW,
		SAY2_MSG_REPLICANT_GONE,
		SAY2_MSG_ATTRIBUTES_SET,
		SAY2_MSG_PREFS,
		SAY2_MSG_NICK_NOT_PRESENT,
		SAY2_MSG_CLIENT_INVALID,
		SAY2_MSG_MEMBERS,
		SAY2_MSG_WALL,
		SAY2_MSG_USER_INVALID,
		SAY2_MSG_USER_INFO,
		SAY2_MSG_USER_ATTRIBUTES,
		SAY2_MSG_USER_STATUS,
		SAY2_MSG_USER_NICK,
		SAY2_MSG_USER_BOOTED,
		SAY2_MSG_USER_TEXT,
		SAY2_MSG_USER_ACT,
		SAY2_MSG_USER_TEXT_TO,
		SAY2_MSG_USER_ACT_TO,
		SAY2_MSG_CHAN_INVALID,
		SAY2_MSG_CHAN_MEMBERS,
		SAY2_MSG_CHAN_SIGNON,
		SAY2_MSG_CHAN_SIGNOFF,
		SAY2_MSG_CHAN_TEXT,
		SAY2_MSG_CHAN_ACT,
		SAY2_MSG_LAST
	);
	gnome_app_set_contents(GNOME_APP(main_window), main_window->interface);
	gtk_widget_show(main_window->interface);
	
	/* Give the entry widget the focus */
	entry = MEKO_CHAT_INTERFACE(main_window->interface)->entry;
	gtk_widget_grab_focus(entry);
	gtk_signal_connect(GTK_OBJECT(entry), "activate",
			   GTK_SIGNAL_FUNC(_entry_activate_cb), main_window);
	main_window->idle_timeout = 0;
}

/* meko_main_window_new
 * Create a new object intsance, just a wrapper around gtk_type_new really
 */
GtkWidget *
meko_main_window_new(void)
{
	GtkWidget *widget;

	widget = GTK_WIDGET(gtk_type_new(MEKO_TYPE_MAIN_WINDOW));

	return widget;
}

/* meko_main_window_destroy
 * Destroy an object instance, just chain up
 */
static void
meko_main_window_destroy(GtkObject *object)
{
	MekoMainWindow *main_window;

	g_return_if_fail(object != NULL);
	g_return_if_fail(MEKO_IS_MAIN_WINDOW(object));

	main_window = MEKO_MAIN_WINDOW(object);

	/* Chain up */
	if (GTK_OBJECT_CLASS(parent_class)->destroy)
		GTK_OBJECT_CLASS(parent_class)->destroy(object);
}

/* meko_main_window_set_arg
 * Set object arguments
 */
static void
meko_main_window_set_arg(GtkObject *object, GtkArg *arg, guint arg_id)
{
	MekoMainWindow *main_window;

	g_return_if_fail(object != NULL);
	g_return_if_fail(MEKO_IS_MAIN_WINDOW(object));

	main_window = MEKO_MAIN_WINDOW(object);
	switch (arg_id) {
	}
}

/* meko_main_window_get_arg
 * Get object arguments
 */
static void
meko_main_window_get_arg(GtkObject *object, GtkArg *arg, guint arg_id)
{
	MekoMainWindow *main_window;

	g_return_if_fail(object != NULL);
	g_return_if_fail(MEKO_IS_MAIN_WINDOW(object));

	main_window = MEKO_MAIN_WINDOW(object);
	switch (arg_id) {
	default:
		arg->type = GTK_TYPE_INVALID;
		break;
	}
}

/* meko_main_window_set
 * Set what server this should be bound to
 */
void
meko_main_window_set(MekoMainWindow *main_window, Say2Server *server)
{
	const gchar *hostname;
	gint         port;
	gchar       *title;

	g_return_if_fail(main_window != NULL);
	g_return_if_fail(server != NULL);

	gtk_object_get(GTK_OBJECT(server), "hostname", &hostname,
		       "port", &port, NULL);
	title = g_strdup_printf(_("Meko: on %s:%d"), hostname, port);
	gtk_window_set_title(GTK_WINDOW(main_window), title);
	g_free(title);

	meko_chat_interface_set_server(
		MEKO_CHAT_INTERFACE(main_window->interface), server
	);

	gtk_signal_connect(GTK_OBJECT(server), "authenticated",
			   GTK_SIGNAL_FUNC(_server_authenticated_cb),
			   main_window);
	gtk_signal_connect(GTK_OBJECT(server), "disconnected",
			   GTK_SIGNAL_FUNC(_server_disconnected_cb),
			   main_window);
}

/* _server_authenticated_cb
 * Server has been successfully authenticated, change ourselves so we are
 * on it's party line
 */
static void
_server_authenticated_cb(Say2Server *server, MekoMainWindow *main_window)
{
	GtkWidget   *client_list;
	const gchar *label, *hostname;
	gint         port;
	gchar	    *title;

	g_return_if_fail(main_window != NULL);
	g_return_if_fail(server != NULL);

	label = meko_say2_client_label(SAY2_CLIENT(server->party_line));
	gtk_object_get(GTK_OBJECT(server), "hostname", &hostname,
		       "port", &port, NULL);
	title = g_strdup_printf(_("Meko: %s on %s:%d"), label, hostname, port);
	gtk_window_set_title(GTK_WINDOW(main_window), title);
	g_free(title);

	meko_chat_interface_set_client(
		MEKO_CHAT_INTERFACE(main_window->interface),
		SAY2_CLIENT(server->party_line)
	);

	client_list = meko_chat_interface_get_client_list(
		MEKO_CHAT_INTERFACE(main_window->interface)
	);
	gtk_signal_connect(GTK_OBJECT(client_list), "activate",
			   GTK_SIGNAL_FUNC(_client_list_activate_cb),
			   main_window);

	/* begin idle timeout */
	main_window->idle_timeout = gtk_timeout_add(
		IDLE_TIME * 1000, (GtkFunction)_idle_cb, (gpointer)main_window
	);
}

/* _server_disconnected_cb
 * Server has disconnected, so update our title
 */
static void
_server_disconnected_cb(Say2Server *server, MekoMainWindow *main_window)
{
	g_return_if_fail(server != NULL);
	g_return_if_fail(main_window != NULL);

	gtk_window_set_title(GTK_WINDOW(main_window), _("Meko"));
}

/* _client_list_activate_cb
 * Open a private message window for the client the user double clicked on,
 * or raise the existing one to the top
 */
static void
_client_list_activate_cb(MekoClientList *client_list, Say2Client *client,
			 MekoMainWindow *main_window)
{
	GtkWidget  *private_window;
	Say2Server *server;

	g_return_if_fail(client_list != NULL);
	g_return_if_fail(client != NULL);
	g_return_if_fail(main_window != NULL);
	g_return_if_fail(SAY2_IS_USER(client));

	server = meko_chat_interface_get_server(
		MEKO_CHAT_INTERFACE(main_window->interface)
	);

	private_window = gtk_object_get_data(GTK_OBJECT(client),
					     "private_window");
	if (private_window == NULL) {
		private_window = meko_private_window_new();
		meko_private_window_set(MEKO_PRIVATE_WINDOW(private_window),
					server, SAY2_USER(client));
		gtk_object_set_data(GTK_OBJECT(client), "private_window",
				    private_window);
		gtk_signal_connect(GTK_OBJECT(private_window), "destroy",
				   GTK_SIGNAL_FUNC(_private_window_destroy_cb),
				   client);
		gtk_widget_show(private_window);
	} else {
		gdk_window_raise(private_window->window);
	}
}

/* _private_window_destroy_cb
 * Private window is closed, so forget about it
 */
static void
_private_window_destroy_cb(MekoPrivateWindow *private_window,
			   Say2Client *client)
{
	g_return_if_fail(private_window != NULL);
	g_return_if_fail(client != NULL);

	gtk_object_set_data(GTK_OBJECT(client), "private_window", NULL);
}

/* _entry_activate_cb
 * Text entry of the main window was activated, so reset the idle counter
 * (or reset our idle status if we've been idle)
 */
static void
_entry_activate_cb(GtkWidget *widget, MekoMainWindow *main_window)
{
	MekoChatInterface *chat_interface;
	Say2Server	  *server;
	Say2User	  *me;
	Say2UserStatus	   status;

	g_return_if_fail(main_window != NULL);

	chat_interface = MEKO_CHAT_INTERFACE(main_window->interface);
	g_return_if_fail(chat_interface != NULL);

	server = meko_chat_interface_get_server(chat_interface);
	g_return_if_fail(server != NULL);

	me = say2_server_get_me(server);
	g_return_if_fail(me != NULL);

	gtk_object_get(GTK_OBJECT(me), "status", &status, NULL);
	if (status == SAY2_USER_IDLE)
		say2_server_send(server, "status", "mode:%s", "active");

	if (main_window->idle_timeout != 0)
		gtk_timeout_remove(main_window->idle_timeout);
	main_window->idle_timeout = gtk_timeout_add(
		IDLE_TIME * 1000, (GtkFunction)_idle_cb, (gpointer)main_window
	);
}

/* _idle_cb
 * Our idle time has elapsed, set our status to idle if we're online
 */
static gint
_idle_cb(MekoMainWindow *main_window)
{
	MekoChatInterface *chat_interface;
	Say2Server	  *server;
	Say2User	  *me;
	Say2UserStatus	   status;

	g_return_val_if_fail(main_window != NULL, FALSE);
	main_window->idle_timeout = 0;

	chat_interface = MEKO_CHAT_INTERFACE(main_window->interface);
	g_return_val_if_fail(chat_interface != NULL, FALSE);

	server = meko_chat_interface_get_server(chat_interface);
	g_return_val_if_fail(server != NULL, FALSE);

	me = say2_server_get_me(server);
	g_return_val_if_fail(me != NULL, FALSE);

	gtk_object_get(GTK_OBJECT(me), "status", &status, NULL);
	if (status != SAY2_USER_ACTIVE)
		return FALSE;

	say2_server_send(server, "status", "mode:%s duration:%d", "idle",
			 IDLE_TIME);
	return FALSE;
}
