/* meko
 * Copyright (C) 2001 Scott James Remnant <scott@netsplit.com>.
 * All Rights Reserved.
 *
 * meko-private-window.c
 *  - MekoPrivateWindow class
 *  - A window containing a MekoChatInterface specifically for the purpose
 *    of having a private conversation with someone.
 * -- 
 * $Id: meko-private-window.c,v 1.5.4.1 2001/11/23 22:19:55 scott Exp $
 * 
 * This file is distributed according to the GNU General Public
 * License.  For full details, read the top of 'main.c' or the
 * file called COPYING that was distributed with this code.
 */

#include <meko.h>

#include <glib.h>
#include <gdk/gdk.h>
#include <gtk/gtkobject.h>
#include <gtk/gtksignal.h>
#include <gtk/gtkwidget.h>
#include <gtk/gtkcontainer.h>
#include <libgnome/gnome-defs.h>
#include <libgnome/gnome-i18n.h>

#include "say2/say2-server.h"
#include "say2/say2-client.h"
#include "util/say2/client-label.h"
#include "widgets/chat-interface/meko-chat-interface.h"

#include "meko-private-window.h"


/* Forward prototypes for required functions */
static void meko_private_window_init(MekoPrivateWindow *);
static void meko_private_window_class_init(MekoPrivateWindowClass *);

/* Forward prototypes for GtkObject methods */
static void meko_private_window_set_arg(GtkObject *, GtkArg *, guint);
static void meko_private_window_get_arg(GtkObject *, GtkArg *, guint);
static void meko_private_window_destroy(GtkObject *);


/* Our parent class */
static GtkWindowClass *parent_class = NULL;


/* meko_private_window_get_type
 * Returns our GtkType id, allocating it if necessary
 */
guint
meko_private_window_get_type(void)
{
	static guint private_window_type = 0;

	if (!private_window_type) {
		static const GtkTypeInfo private_window_info = {
			"MekoPrivateWindow",
			sizeof(MekoPrivateWindow),
			sizeof(MekoPrivateWindowClass),
			(GtkClassInitFunc)meko_private_window_class_init,
			(GtkObjectInitFunc)meko_private_window_init,
			/* reserved_1 */ NULL,
			/* reserved_2 */ NULL,
			(GtkClassInitFunc)NULL
		};

		private_window_type = gtk_type_unique(GTK_TYPE_WINDOW,
						      &private_window_info);
	}

	return private_window_type;
}

/* meko_private_window_class_init
 * Initialise the object class, overriding methods of parent object classes
 */
static void
meko_private_window_class_init(MekoPrivateWindowClass *klass)
{
	GtkObjectClass *object_class;

	g_return_if_fail(klass != NULL);

	object_class = (GtkObjectClass *)klass;
	parent_class = gtk_type_class(GTK_TYPE_OBJECT);

	object_class->set_arg = meko_private_window_set_arg;
	object_class->get_arg = meko_private_window_get_arg;
	object_class->destroy = meko_private_window_destroy;
}

/* meko_private_window_init
 * Initialise an object instance, setting up window attributes and placing
 * a ChatInterface inside
 */
static void
meko_private_window_init(MekoPrivateWindow *private_window)
{
	GtkWidget *entry;

	g_return_if_fail(private_window != NULL);

	/* Set the window stuff */
	gtk_window_set_title(GTK_WINDOW(private_window),
			     _("Private conversation"));
	gtk_window_set_wmclass(GTK_WINDOW(private_window),
			       "private-window", "Meko");
	gtk_window_set_policy(GTK_WINDOW(private_window), FALSE, TRUE, FALSE);
	gtk_window_set_default_size(GTK_WINDOW(private_window), 300, 150);
	gtk_container_set_border_width(GTK_CONTAINER(private_window), 2);

	/* Create the interface inside */
	private_window->interface = meko_chat_interface_new(NULL, NULL);
	meko_chat_interface_recv_msg_set(
		MEKO_CHAT_INTERFACE(private_window->interface),
		SAY2_MSG_CONNECTED,
		SAY2_MSG_DISCONNECTED,
		SAY2_MSG_USER_STATUS,
		SAY2_MSG_USER_NICK,
		SAY2_MSG_USER_TEXT,
		SAY2_MSG_USER_ACT,
		SAY2_MSG_USER_TEXT_TO,
		SAY2_MSG_USER_ACT_TO,
		SAY2_MSG_LAST
	);
	gtk_container_add(GTK_CONTAINER(private_window),
			  private_window->interface);
	gtk_widget_show(private_window->interface);

	entry = MEKO_CHAT_INTERFACE(private_window->interface)->entry;
	gtk_widget_grab_focus(entry);
}

/* meko_private_window_new
 * Create a new object intsance, just a wrapper around gtk_type_new really
 */
GtkWidget *
meko_private_window_new(void)
{
	GtkWidget *widget;

	widget = GTK_WIDGET(gtk_type_new(MEKO_TYPE_PRIVATE_WINDOW));

	return widget;
}

/* meko_private_window_destroy
 * Destroy an object instance, just chain up
 */
static void
meko_private_window_destroy(GtkObject *object)
{
	MekoPrivateWindow *private_window;

	g_return_if_fail(object != NULL);
	g_return_if_fail(MEKO_IS_PRIVATE_WINDOW(object));

	private_window = MEKO_PRIVATE_WINDOW(object);

	/* Chain up */
	if (GTK_OBJECT_CLASS(parent_class)->destroy)
		GTK_OBJECT_CLASS(parent_class)->destroy(object);
}

/* meko_private_window_set_arg
 * Set object arguments
 */
static void
meko_private_window_set_arg(GtkObject *object, GtkArg *arg, guint arg_id)
{
	MekoPrivateWindow *private_window;

	g_return_if_fail(object != NULL);
	g_return_if_fail(MEKO_IS_PRIVATE_WINDOW(object));

	private_window = MEKO_PRIVATE_WINDOW(object);
	switch (arg_id) {
	}
}

/* meko_private_window_get_arg
 * Get object arguments
 */
static void
meko_private_window_get_arg(GtkObject *object, GtkArg *arg, guint arg_id)
{
	MekoPrivateWindow *private_window;

	g_return_if_fail(object != NULL);
	g_return_if_fail(MEKO_IS_PRIVATE_WINDOW(object));

	private_window = MEKO_PRIVATE_WINDOW(object);
	switch (arg_id) {
	default:
		arg->type = GTK_TYPE_INVALID;
		break;
	}
}

/* meko_private_window_set
 * Set who we're chatting to
 */
void
meko_private_window_set(MekoPrivateWindow *private_window,
		        Say2Server *server, Say2User *user)
{
	gchar *title;

	g_return_if_fail(private_window != NULL);
	g_return_if_fail(server != NULL);
	g_return_if_fail(user != NULL);

	title = g_strdup_printf(_("Private conversation with %s"),
			        meko_say2_client_label(SAY2_CLIENT(user)));
	gtk_window_set_title(GTK_WINDOW(private_window), title);
	g_free(title);

	meko_chat_interface_set_server(
		MEKO_CHAT_INTERFACE(private_window->interface), server
	);
	meko_chat_interface_set_client(
		MEKO_CHAT_INTERFACE(private_window->interface),
		SAY2_CLIENT(user)
	);
}
