/* meko
 * Copyright (C) 2001 Scott James Remnant <scott@netsplit.com>.
 * All Rights Reserved.
 *
 * say2-channels-commands.c
 *  - Handles commands received from a Say2 channel server (well, ok, from
 *    the registration server right now, but hey)
 * -- 
 * $Id: say2-channel-commands.c,v 1.6 2001/11/03 21:14:09 scott Exp $
 * 
 * This file is distributed according to the GNU General Public
 * License.  For full details, read the top of 'main.c' or the
 * file called COPYING that was distributed with this code.
 */

#include <meko.h>

#include <time.h>

#include <glib.h>
#include <libgnome/gnome-defs.h>
#include <libgnome/gnome-i18n.h>
#include <say2/sock.h>
#include <say2/enaml.h>
#include <say2/enaml-sock.h>
#include <say2/dispatch.h>

#include "say2/say2-user.h"
#include "say2/say2-channel.h"
#include "say2/say2-server.h"

#include "say2-channel-commands.h"


/* Forward prototypes for command functions */

/* Forward prototypes for internal functions */


/* Command dispatch table */
/*static DispatchItem channel_commands[] = {
	{ NULL,				NULL }
};*/


/* CHANNEL_signon
 * Someone has joined the channel
 */
void
CHANNEL_signon(Say2Server *server, Say2Channel *channel,
	       const EnamlBlock *signon)
{
	const gchar *nick, *id, *email;
	time_t	     signon_time;
	Say2User    *user;

	nick = enaml_get_string(signon, "nick", "?!@");
	id = enaml_get_string(signon, "id", "?|@");
	email = enaml_get_string(signon, "email", "?!@");
	signon_time = enaml_get_long(signon, "time", time(NULL));

	/* Get the user and update nick and email information */
	user = say2_server_get_user(server, id);
	gtk_object_set(GTK_OBJECT(user), "nick", nick, "email", email, NULL);

	/* Add user to the channel */
	say2_channel_add_member(channel, user);
	say2_server_message(server, SAY2_CLIENT(channel), SAY2_MSG_CHAN_SIGNON,
			    "time", signon_time,
			    "user", user,
			    NULL);
}

/* CHANNEL_signoff
 * Someone has left the channel
 */
void
CHANNEL_signoff(Say2Server *server, Say2Channel *channel,
		const EnamlBlock *signoff)
{
	const gchar *nick, *id, *user_email;
	time_t	     signoff_time;
	Say2User    *user;

	nick = enaml_get_string(signoff, "nick", "?!@");
	id = enaml_get_string(signoff, "id", "?!@");
	signoff_time = enaml_get_long(signoff, "time", time(NULL));

	/* Get the user and e-mail information */
	user = say2_channel_get_member(channel, id);
	if (user == NULL) {
		gchar *s;

		s = g_strdup_printf("%s: %s", _("Unknown user"), id);
		say2_server_message(server, NULL, SAY2_MSG_BAD_COMMAND,
				    "command", "signoff",
				    "text", s,
				    NULL);
		g_free(s);
		return;
	}
	gtk_object_get(GTK_OBJECT(user), "email", &user_email, NULL);

	/* Remove user from the channel */
	say2_server_message(server, SAY2_CLIENT(channel), SAY2_MSG_CHAN_SIGNOFF,
			    "time", signoff_time,
			    "user", user,
			    NULL);
	say2_channel_remove_member(channel, user);
}

/* CHANNEL_boot_user
 * Someone has been booted off the channel
 */
void
CHANNEL_boot_user(Say2Server *server, Say2Channel *channel,
		  const EnamlBlock *boot_user)
{
	const gchar *nick, *id, *target_nick, *target_id, *text;
	Say2User    *user, *target_user;

	nick = enaml_get_string(boot_user, "nick", "?!@");
	id = enaml_get_string(boot_user, "id", "?!@");
	target_nick = enaml_get_string(boot_user, "target-nick", "?!@");
	target_id = enaml_get_string(boot_user, "target-id", "?!@");
	text = enaml_get_string(boot_user, "text", NULL);

	/* Get who they were */
	user = say2_server_get_user(server, id);
	target_user = say2_channel_get_member(channel, target_id);
	if (target_user == NULL) {
		gchar *s;

		s = g_strdup_printf("%s: %s", _("Unknown user"), id);
		say2_server_message(server, NULL, SAY2_MSG_BAD_COMMAND,
				    "command", "boot-user",
				    "text", s,
				    NULL);
		g_free(s);
		return;
	}

	/* Not sure whether we need to do this? */
	say2_server_message(server, SAY2_CLIENT(target_user),
			    SAY2_MSG_USER_BOOTED,
			    "by", user,
			    "text", text,
			    NULL);
	say2_channel_remove_member(channel, target_user);
}

/* CHANNEL_say
 * Channel message
 */
void
CHANNEL_say(Say2Server *server, Say2Channel *channel, const EnamlBlock *say)
{
	const gchar *id, *nick, *text, *act;
	time_t	     say_time;
	Say2User    *user;

	id = enaml_get_string(say, "id", "?!@");
	nick = enaml_get_string(say, "nick", "?!@");
	say_time = (time_t)enaml_get_long(say, "time", time(NULL));
	text = enaml_get_string(say, "text", NULL);
	act = enaml_get_string(say, "act", NULL);

	user = say2_server_get_user(server, id);

	if (text != NULL) {
		say2_server_message(server, SAY2_CLIENT(channel),
				    SAY2_MSG_CHAN_TEXT,
				    "time", say_time,
				    "user", user,
				    "text", text,
				    NULL);
	}
	if (act != NULL) {
		say2_server_message(server, SAY2_CLIENT(channel),
				    SAY2_MSG_CHAN_ACT,
				    "time", say_time,
				    "user", user,
				    "text", act,
				    NULL);
	}
}
