/* meko
 * Copyright (C) 2001 Scott James Remnant <scott@netsplit.com>.
 * All Rights Reserved.
 *
 * say2-channel.c
 *  - Say2Channel class
 *  - An object to hold information about a channel on the say2 network.
 * -- 
 * $Id: say2-channel.c,v 1.14 2001/09/05 07:43:00 scott Exp $
 * 
 * This file is distributed according to the GNU General Public
 * License.  For full details, read the top of 'main.c' or the
 * file called COPYING that was distributed with this code.
 */

#include <meko.h>

#include <glib.h>
#include <gtk/gtkobject.h>
#include <gtk/gtksignal.h>

#include "say2/say2-client.h"

#include "say2-channel.h"

/* Signals */
enum {
	SIGNON,
	SIGNOFF,
	LAST_SIGNAL
};


/* Forward prototypes for required functions */
static void say2_channel_init(Say2Channel *);
static void say2_channel_class_init(Say2ChannelClass *);

/* Forward prototypes for GtkObject methods */
static void say2_channel_destroy(GtkObject *);

/* Forward prototypes for internal functions */
static gboolean _unref_hash_val(gpointer, gpointer, gpointer);


/* Out parent class */
static Say2ClientClass *parent_class = NULL;

/* Signal array */
static guint channel_signals[LAST_SIGNAL] = { 0 };


/* say2_channel_get_type
 * Returns our GtkType id, allocating it if necessary
 */
guint
say2_channel_get_type(void)
{
	static guint channel_type = 0;

	if (!channel_type) {
		static const GtkTypeInfo channel_info = {
			"Say2Channel",
			sizeof(Say2Channel),
			sizeof(Say2ChannelClass),
			(GtkClassInitFunc)say2_channel_class_init,
			(GtkObjectInitFunc)say2_channel_init,
			/* reserved_1 */ NULL,
			/* reserved_2 */ NULL,
			(GtkClassInitFunc)NULL
		};

		channel_type = gtk_type_unique(SAY2_TYPE_CLIENT, &channel_info);
	}

	return channel_type;
}

/* say2_channel_class_init
 * Initialise the object class, setting our signals and overriding methods
 * of parent object classes
 */
static void
say2_channel_class_init(Say2ChannelClass *klass)
{
	GtkObjectClass *object_class;

	g_return_if_fail(klass != NULL);

	object_class = (GtkObjectClass *)klass;
	parent_class = gtk_type_class(SAY2_TYPE_CLIENT);

	channel_signals[SIGNON] = gtk_signal_new(
		"signon", GTK_RUN_FIRST, object_class->type,
		GTK_SIGNAL_OFFSET(Say2ChannelClass, signon),
		gtk_marshal_NONE__POINTER,
		GTK_TYPE_NONE, 1, GTK_TYPE_POINTER
	);

	channel_signals[SIGNOFF] = gtk_signal_new(
		"signoff", GTK_RUN_FIRST, object_class->type,
		GTK_SIGNAL_OFFSET(Say2ChannelClass, signoff),
		gtk_marshal_NONE__POINTER,
		GTK_TYPE_NONE, 1, GTK_TYPE_POINTER
	);

	gtk_object_class_add_signals(object_class, channel_signals,
				     LAST_SIGNAL);

	object_class->destroy = say2_channel_destroy;

	klass->signon = NULL;
	klass->signoff = NULL;
}

/* say2_channel_init
 * Initialise an object instance, very little to do here other than set some
 * default values
 */
static void
say2_channel_init(Say2Channel *channel)
{
	g_return_if_fail(channel != NULL);

	channel->members = g_hash_table_new(g_str_hash, g_str_equal);
}

/* _unref_hash_val
 * Unreferences data in the members hash table 
 */
static gboolean
_unref_hash_val(gpointer key, gpointer value, gpointer user_data)
{
	g_return_val_if_fail(value != NULL, TRUE);
	g_return_val_if_fail(GTK_IS_OBJECT(value), TRUE);

	gtk_object_unref(GTK_OBJECT(value));

	return TRUE;
}

/* say2_channel_new
 * Create a new object instance with the id given, just a wrapper around
 * gtk_type_new and gtk_object_set really.
 */
Say2Channel *
say2_channel_new(const gchar *id)
{
	Say2Channel *channel;

	channel = SAY2_CHANNEL(gtk_type_new(SAY2_TYPE_CHANNEL));

	if (id)
		gtk_object_set(GTK_OBJECT(channel), "id", id, NULL);

	return channel;
}

/* say2_channel_destroy
 * Destroy an object instance, freeing and unreferencing any data we've
 * been hanging on to
 */
static void
say2_channel_destroy(GtkObject *object)
{
	Say2Channel *channel;

	g_return_if_fail(object != NULL);
	g_return_if_fail(SAY2_IS_CHANNEL(object));

	channel = SAY2_CHANNEL(object);

	/* Clear the members hash (we'll do this on disconnect eventually) */
	g_hash_table_foreach_remove(channel->members, _unref_hash_val, NULL);

	/* Destroy the hash */
	g_hash_table_destroy(channel->members);
	channel->members = NULL;

	/* Chain up */
	if (GTK_OBJECT_CLASS(parent_class)->destroy)
		GTK_OBJECT_CLASS(parent_class)->destroy(object);
}

/* say2_channel_add_member
 * Add a user to the channel's members list
 */
void
say2_channel_add_member(Say2Channel *channel, Say2User *user)
{
	Say2User    *existing_record;
	const gchar *id;

	g_return_if_fail(channel != NULL);
	g_return_if_fail(user != NULL);

	gtk_object_get(GTK_OBJECT(user), "id", &id, NULL);

	/* Make sure they don't already exist */
	existing_record = g_hash_table_lookup(channel->members, id);
	g_return_if_fail(existing_record == NULL);

	/* Add the record */
	g_hash_table_insert(channel->members, (gpointer)id, user);
	gtk_object_ref(GTK_OBJECT(user));
	gtk_object_sink(GTK_OBJECT(user));

	/* Emit a signal */
	gtk_signal_emit(GTK_OBJECT(channel), channel_signals[SIGNON], user);
}

/* say2_channel_remove_member
 * Remove a member from the channel's member list
 */
void
say2_channel_remove_member(Say2Channel *channel, Say2User *user)
{
	Say2User    *existing_record;
	const gchar *id;

	g_return_if_fail(channel != NULL);
	g_return_if_fail(user != NULL);

	/* Check it's in the table first */
	gtk_object_get(GTK_OBJECT(user), "id", &id, NULL);
	existing_record = say2_channel_get_member(channel, id);
	g_return_if_fail(existing_record != NULL);
	g_return_if_fail(existing_record == user);
	
	/* Emit a signal */
	gtk_signal_emit(GTK_OBJECT(channel), channel_signals[SIGNOFF], user);

	/* Actually remove it */
	g_hash_table_remove(channel->members, id);
	gtk_object_unref(GTK_OBJECT(user));
}

/* say2_channel_get_member
 * Retreive a member from the channel's member list
 */
Say2User *
say2_channel_get_member(Say2Channel *channel, const gchar *id)
{
	g_return_val_if_fail(channel != NULL, NULL);

	return g_hash_table_lookup(channel->members, id);
}
