/* meko
 * Copyright (C) 2001 Scott James Remnant <scott@netsplit.com>.
 * All Rights Reserved.
 *
 * say2-client.c
 *  - Say2Client class
 *  - An object to hold information about a client on the say2 network.
 * -- 
 * $Id: say2-client.c,v 1.16 2001/09/10 08:54:10 scott Exp $
 * 
 * This file is distributed according to the GNU General Public
 * License.  For full details, read the top of 'main.c' or the
 * file called COPYING that was distributed with this code.
 */

#include <meko.h>

#include <string.h>

#include <glib.h>
#include <gtk/gtkobject.h>
#include <gtk/gtksignal.h>

#include "say2-client.h"

/* Arguments
 * gchar *	id		Unique id on the server
 * gchar *	email		E-Mail address
 * gchar *	tagline		Tagline
 * gchar *	client		Client software version
 * gchar *	host		Host they are connected from
 * time_t	login_at	When they (last?) logged in
 * time_t	logout_at	When they (last?) logged out
 */
enum {
	ARG_0,
	ARG_ID,
	ARG_EMAIL,
	ARG_TAGLINE,
	ARG_CLIENT,
	ARG_HOST,
	ARG_LOGIN_AT,
	ARG_LOGOUT_AT
};

/* Signals */
enum {
	CHANGED,
	INVALID,
	LAST_SIGNAL
};


/* Forward prototypes for required functions */
static void say2_client_init(Say2Client *);
static void say2_client_class_init(Say2ClientClass *);

/* Forward prototypes for GtkObject methods */
static void say2_client_set_arg(GtkObject *, GtkArg *, guint);
static void say2_client_get_arg(GtkObject *, GtkArg *, guint);
static void say2_client_destroy(GtkObject *);

/* Forward prototypes for internal functions */
static gboolean _free_hash(gpointer, gpointer, gpointer);


/* Our parent class */
static GtkObjectClass *parent_class = NULL;

/* Signal array */
static guint client_signals[LAST_SIGNAL] = { 0 };


/* say2_client_get_type
 * Returns our GtkType id, allocating it if necessary
 */
guint
say2_client_get_type(void)
{
	static guint client_type = 0;

	if (!client_type) {
		static const GtkTypeInfo client_info = {
			"Say2Client",
			sizeof(Say2Client),
			sizeof(Say2ClientClass),
			(GtkClassInitFunc)say2_client_class_init,
			(GtkObjectInitFunc)say2_client_init,
			/* reserved_1 */ NULL,
			/* reserved_2 */ NULL,
			(GtkClassInitFunc)NULL
		};

		client_type = gtk_type_unique(GTK_TYPE_OBJECT, &client_info);
	}

	return client_type;
}

/* say2_client_class_init
 * Initialise the object class, setting our arguments, signals and overriding
 * methods of parent object classes
 */
static void
say2_client_class_init(Say2ClientClass *klass)
{
	GtkObjectClass *object_class;

	g_return_if_fail(klass != NULL);

	object_class = (GtkObjectClass *)klass;
	parent_class = gtk_type_class(GTK_TYPE_OBJECT);

	gtk_object_add_arg_type("Say2Client::id", GTK_TYPE_STRING,
				GTK_ARG_READWRITE, ARG_ID);
	gtk_object_add_arg_type("Say2Client::email", GTK_TYPE_STRING,
				GTK_ARG_READWRITE, ARG_EMAIL);
	gtk_object_add_arg_type("Say2Client::tagline", GTK_TYPE_STRING,
				GTK_ARG_READWRITE, ARG_TAGLINE);
	gtk_object_add_arg_type("Say2Client::client", GTK_TYPE_STRING,
				GTK_ARG_READWRITE, ARG_CLIENT);
	gtk_object_add_arg_type("Say2Client::host", GTK_TYPE_STRING,
				GTK_ARG_READWRITE, ARG_HOST);
	gtk_object_add_arg_type("Say2Client::login_at", GTK_TYPE_ULONG,
				GTK_ARG_READWRITE, ARG_LOGIN_AT);
	gtk_object_add_arg_type("Say2Client::logout_at", GTK_TYPE_ULONG,
				GTK_ARG_READWRITE, ARG_LOGOUT_AT);

	client_signals[CHANGED] = gtk_signal_new(
		"changed", GTK_RUN_FIRST | GTK_RUN_ACTION, object_class->type,
		GTK_SIGNAL_OFFSET(Say2ClientClass, changed),
		gtk_marshal_NONE__NONE,
		GTK_TYPE_NONE, 0
	);

	client_signals[INVALID] = gtk_signal_new(
		"invalid", GTK_RUN_FIRST, object_class->type,
		GTK_SIGNAL_OFFSET(Say2ClientClass, invalid),
		gtk_marshal_NONE__NONE,
		GTK_TYPE_NONE, 0
	);

	gtk_object_class_add_signals(object_class, client_signals, LAST_SIGNAL);

	object_class->set_arg = say2_client_set_arg;
	object_class->get_arg = say2_client_get_arg;
	object_class->destroy = say2_client_destroy;

	klass->changed = NULL;
	klass->invalid = NULL;
}

/* say2_client_init
 * Initialise an object instance, very little to do here other than set default
 * values
 */
static void
say2_client_init(Say2Client *client)
{
	g_return_if_fail(client != NULL);

	client->id = NULL;
	client->email = NULL;
	client->tagline = NULL;
	client->client = NULL;
	client->host = NULL;
	client->login_at = 0;
	client->logout_at = 0;

	client->attr = g_hash_table_new(g_str_hash, g_str_equal);

	client->icon = 0;
	client->icon_sz = 0;
}

/* _free_hash
 * Frees the values given to it and returns TRUE
 */
static gboolean
_free_hash(gpointer key, gpointer value, gpointer user_data)
{
	g_free(key);
	g_free(value);

	return TRUE;
}

/* say2_client_destroy
 * Destroy an object instance, freeing and unreferencing any data we've
 * been hanging on to
 */
static void
say2_client_destroy(GtkObject *object)
{
	Say2Client *client;

	g_return_if_fail(object != NULL);
	g_return_if_fail(SAY2_IS_CLIENT(object));

	client = SAY2_CLIENT(object);

	g_free(client->id);
	g_free(client->email);
	g_free(client->tagline);
	g_free(client->client);
	g_free(client->host);

	g_free(client->icon);
	client->icon_sz = 0;

	g_hash_table_foreach_remove(client->attr, _free_hash, NULL);
	g_hash_table_destroy(client->attr);
	client->attr = NULL;

	/* Chain up */
	if (GTK_OBJECT_CLASS(parent_class)->destroy)
		GTK_OBJECT_CLASS(parent_class)->destroy(object);
}

/* say2_client_set_arg
 * Set object arguments
 */
static void
say2_client_set_arg(GtkObject *object, GtkArg *arg, guint arg_id)
{
	Say2Client *client;

	g_return_if_fail(object != NULL);
	g_return_if_fail(SAY2_IS_CLIENT(object));

	client = SAY2_CLIENT(object);
	switch (arg_id) {
	case ARG_ID:
		g_free(client->id);
		client->id = g_strdup(GTK_VALUE_STRING(*arg));
		break;
	case ARG_EMAIL:
		g_free(client->email);
		client->email = g_strdup(GTK_VALUE_STRING(*arg));
		break;
	case ARG_TAGLINE:
		g_free(client->tagline);
		client->tagline = g_strdup(GTK_VALUE_STRING(*arg));
		break;
	case ARG_CLIENT:
		g_free(client->client);
		client->client = g_strdup(GTK_VALUE_STRING(*arg));
		break;
	case ARG_HOST:
		g_free(client->host);
		client->host = g_strdup(GTK_VALUE_STRING(*arg));
		break;
	case ARG_LOGIN_AT:
		client->login_at = GTK_VALUE_ULONG(*arg);
		break;
	case ARG_LOGOUT_AT:
		client->login_at = GTK_VALUE_ULONG(*arg);
		break;
	}

	say2_client_changed(client);
}

/* say2_client_get_arg
 * Get object arguments
 */
static void
say2_client_get_arg(GtkObject *object, GtkArg *arg, guint arg_id)
{
	Say2Client *client;

	g_return_if_fail(object != NULL);
	g_return_if_fail(SAY2_IS_CLIENT(object));

	client = SAY2_CLIENT(object);
	switch (arg_id) {
	case ARG_ID:
		GTK_VALUE_STRING(*arg) = client->id;
		break;
	case ARG_EMAIL:
		GTK_VALUE_STRING(*arg) = client->email;
		break;
	case ARG_TAGLINE:
		GTK_VALUE_STRING(*arg) = client->tagline;
		break;
	case ARG_CLIENT:
		GTK_VALUE_STRING(*arg) = client->client;
		break;
	case ARG_HOST:
		GTK_VALUE_STRING(*arg) = client->host;
		break;
	case ARG_LOGIN_AT:
		GTK_VALUE_ULONG(*arg) = client->login_at;
		break;
	case ARG_LOGOUT_AT:
		GTK_VALUE_ULONG(*arg) = client->logout_at;
		break;
	default:
		arg->type = GTK_TYPE_INVALID;
		break;
	}
}

/* say2_client_changed
 * Emit the changed signal
 */
void
say2_client_changed(Say2Client *client)
{
	g_return_if_fail(client != NULL);

	gtk_signal_emit(GTK_OBJECT(client), client_signals[CHANGED]);
}

/* say2_client_invalid
 * Emit the invalid signal
 */
void
say2_client_invalid(Say2Client *client)
{
	g_return_if_fail(client != NULL);

	gtk_signal_emit(GTK_OBJECT(client), client_signals[INVALID]);
}

/* say2_client_set_attr
 * Set an attribute of the client, one of the optional string values.
 */
void
say2_client_set_attr(Say2Client *client, const gchar *key, const gchar *val)
{
	gchar *lwrkey, *oldkey, *oldval;

	g_return_if_fail(client != NULL);
	g_return_if_fail(SAY2_IS_CLIENT(client));

	/* All attributes are lowercase */
	lwrkey = g_strdup(key);
	g_strdown(lwrkey);

	/* Free any old value */
	if (g_hash_table_lookup_extended(client->attr, lwrkey,
					 (gpointer *)&oldkey,
					 (gpointer *)&oldval)) {
		g_hash_table_remove(client->attr, lwrkey);
		g_free(oldkey);
		g_free(oldval);
	}

	/* Don't insert a NULL record, just for efficiency */
	if (!val) {
		g_free(lwrkey);
		say2_client_changed(client);
		return;
	}

	/* Insert the new value (eats lwrkey) */
	g_hash_table_insert(client->attr, lwrkey, g_strdup(val));
	say2_client_changed(client);
}

/* say2_client_get_attr
 * Get an attribute of the client, returns the string value, if you need to
 * do anything particular with it, it'll need to be copied
 */
const gchar *
say2_client_get_attr(Say2Client *client, const gchar *key)
{
	const gchar *val;
	gchar	    *lwrkey;

	g_return_val_if_fail(client != NULL, NULL);
	g_return_val_if_fail(SAY2_IS_CLIENT(client), NULL);

	/* All attributes are lowercase */
	lwrkey = g_strdup(key);
	g_strdown(lwrkey);

	/* Lookup and return */
	val = g_hash_table_lookup(client->attr, lwrkey);
	g_free(lwrkey);

	return val;
}

/* say2_client_foreach_attr
 * Wrapper around g_hash_table_foreach for the attributes list
 */
void
say2_client_foreach_attr(Say2Client *client, GHFunc func, gpointer user_data)
{
	g_return_if_fail(client != NULL);

	g_hash_table_foreach(client->attr, func, user_data);
}

/* say2_client_set_icon
 * Set the client's icon data and size, acts directly on the object as icons
 * aren't stored as attributes
 */
void
say2_client_set_icon(Say2Client *client, const char *icon, int icon_sz)
{
	g_return_if_fail(client != NULL);

	g_free(client->icon);

	client->icon = g_new0(char, icon_sz);
	memcpy(client->icon, icon, icon_sz);

	client->icon_sz = icon_sz;
	
	say2_client_changed(client);
}

/* say2_client_get_icon
 * Get the client's icon data and size directly from the object, as icons aren't
 * stored as attributes.
 */
int
say2_client_get_icon(Say2Client *client, char **icon)
{
	g_return_val_if_fail(client != NULL, 0);

	if (icon)
		*icon = client->icon;

	return client->icon_sz;
}
