/* meko
 * Copyright (C) 2001 Scott James Remnant <scott@netsplit.com>.
 * All Rights Reserved.
 *
 * say2-server-commands.c
 *  - Handles commands received from a Say2 registration server
 * -- 
 * $Id: say2-server-commands.c,v 1.13.4.2 2001/11/23 22:34:44 scott Exp $
 * 
 * This file is distributed according to the GNU General Public
 * License.  For full details, read the top of 'main.c' or the
 * file called COPYING that was distributed with this code.
 */

#include <meko.h>

#include <string.h>
#include <time.h>

#include <glib.h>
#include <libgnome/gnome-defs.h>
#include <libgnome/gnome-i18n.h>
#include <say2/sock.h>
#include <say2/enaml.h>
#include <say2/enaml-sock.h>
#include <say2/dispatch.h>

#include "say2/say2-user.h"
#include "say2/say2-server.h"
#include "say2/say2-channel.h"
#include "say2/say2-channel-commands.h"

#include "say2-server-commands.h"


/* Forward prototypes for command functions */
static void _error(Say2Server *, const EnamlBlock *);
static void _hello(Say2Server *, const EnamlBlock *);
static void _server_info(Say2Server *, const EnamlBlock *);
static void _motd(Say2Server *, const EnamlBlock *);
static void _welcome(Say2Server *, const EnamlBlock *);
static void _bad_nick(Say2Server *, const EnamlBlock *);
static void _bad_address(Say2Server *, const EnamlBlock *);
static void _logout(Say2Server *, const EnamlBlock *);
static void _replicant_terminated(Say2Server *, const EnamlBlock *);
static void _ping(Say2Server *, const EnamlBlock *);
static void _signon(Say2Server *, const EnamlBlock *);
static void _signoff(Say2Server *, const EnamlBlock *);
static void _prefs(Say2Server *, const EnamlBlock *);
static void _no_user_record(Say2Server *, const EnamlBlock *);
static void _status(Say2Server *, const EnamlBlock *);
static void _change_nick(Say2Server *, const EnamlBlock *);
static void _members(Say2Server *, const EnamlBlock *);
static void _begin_response(Say2Server *, const EnamlBlock *);
static void _end_response(Say2Server *, const EnamlBlock *);
static void _client_info(Say2Server *, const EnamlBlock *);
static void _not_present(Say2Server *, const EnamlBlock *);
static void _no_such_client(Say2Server *, const EnamlBlock *);
static void _boot_user(Say2Server *, const EnamlBlock *);
static void _client_attributes(Say2Server *, const EnamlBlock *);
static void _set_attributes(Say2Server *, const EnamlBlock *);
static void _say(Say2Server *, const EnamlBlock *);
static void _message(Say2Server *, const EnamlBlock *);
static void _wall(Say2Server *, const EnamlBlock *);
static void _start_tls(Say2Server *server, const EnamlBlock *start_tls);

/* Forward prototypes for internal functions */


/* Command dispatch table */
static DispatchItem server_commands[] = {
	{ "error",			_error },
	{ "hello",			_hello },
	{ "server-info",		_server_info },
	{ "motd",			_motd },
	{ "welcome",			_welcome },
	{ "bad-nick",			_bad_nick },
	{ "bad-address",		_bad_address },
	{ "logout",			_logout },
	{ "replicant-terminated",	_replicant_terminated },
	{ "ping",			_ping },
	{ "signon",			_signon },
	{ "signoff",			_signoff },
	{ "prefs",			_prefs },
	{ "no-user-record",		_no_user_record },
	{ "status",			_status },
	{ "change-nick",		_change_nick },
	{ "members",			_members },
	{ "begin-response",		_begin_response },
	{ "end-response",		_end_response },
	{ "client-info",		_client_info },
	{ "not-present",		_not_present },
	{ "no-such-client",		_no_such_client },
	{ "boot-user",			_boot_user },
	{ "client-attributes",		_client_attributes },
	{ "set-attributes",		_set_attributes },
	{ "say",			_say },
	{ "message",			_message },
	{ "wall",			_wall },
        { "start-tls",			_start_tls },
	{ NULL,				NULL }
};


/* say2_server_command
 * Handle a command and associated block received from the say2 server by
 * dispatching it to a function in this source file
 */
gboolean
say2_server_command(Say2Server *server, const gchar *command,
		    const EnamlBlock *block)
{
	static void *server_dispatch = NULL;

	g_return_val_if_fail(server != NULL, FALSE);
	g_return_val_if_fail(server->state >= SAY2_SERVER_CONNECTED, FALSE);

	/* Only a couple of commands can be sent as flags, make sure that
	 * the block exists here except for those too commands
	 */
	if (g_strcasecmp(command, "hello") && g_strcasecmp(command, "ping")
	    && g_strcasecmp(command, "no-user-record")
            && g_strcasecmp(command, "start-tls")
	    && (block == NULL)) {
		say2_server_message(server, NULL, SAY2_MSG_BAD_COMMAND,
				    "command", command,
				    "text", _("Command not sent as a block"),
				    NULL);
		return TRUE;
	}
	
	if (server_dispatch == NULL)
		server_dispatch = dispatch_create(server_commands);

	return dispatch_go(server_dispatch, command, server,
			   (void *)block, NULL);
}

/* _error
 * Generic error reponse
 */
static void
_error(Say2Server *server, const EnamlBlock *error)
{
	const gchar *command, *text;

	command = enaml_get_string(error, "command", NULL);
	text = enaml_get_string(error, "text", "?!@");

	/* Actually generates the error as the command not as an error command
	 * this means if one's squelching "prefs" and a prefs error comes back
	 * that error will be squelched, which is probably what we want.
	 */
	say2_server_message(server, NULL, SAY2_MSG_ERROR,
			    "command", command,
			    "text", text,
			    NULL);
}

/* _hello
 * Initial greeting that confirms we're actually talking to a say2 server
 */
static void
_hello(Say2Server *server, const EnamlBlock *hello)
{
        GtkArg arg;

	if (say2_server_get_state(server) != SAY2_SERVER_CONNECTED) {
		say2_server_message(server, NULL, SAY2_MSG_BAD_COMMAND,
				    "command", "hello",
				    "text", _("Unexpected at this time"),
				    NULL);
		return;
	}

	if (hello != NULL) {
		EnamlList *protocol;
		glong	   major, minor;

		/* Protocol is a list of two ulongs */
		protocol = enaml_get_list(hello, "protocol");
		if (protocol && (enaml_list_length(protocol) >= 2)) {
			major = enaml_list_get_long(protocol, 0, 0);
			minor = enaml_list_get_long(protocol, 1, 0);
		} else {
			major = 0;
			minor = 0;
		}

		/* Set flags for things we know about this protocol
		 * version (FIXME: none, yet)
		 */
	}

	/* It's a say2 server, yay! */
	say2_server_verified(server);

        /* try ssl? */
        arg.name = "Say2Server::ssl";
        gtk_object_arg_get(GTK_OBJECT(server), &arg, NULL);
        if (GTK_VALUE_BOOL(arg)) {
                say2_server_send(server, "start-tls", NULL);
                /* do login after SSL is established */
                return;
        }

	/* Try to login, this'll return FALSE if there's not enough info
	 * so we can print a command
	 */
	if (!say2_server_login(server))
		say2_server_message(server, NULL, SAY2_MSG_NEED_LOGIN, NULL);
}

/* _start_tls */
static void
_start_tls(Say2Server *server, const EnamlBlock *start_tls)
{
        if (! say2_server_begin_ssl(server)) {
                say2_server_message(server, NULL, SAY2_MSG_SSL_FAILED, NULL);
        } else {
                say2_server_message(server, NULL, SAY2_MSG_SSL_SUCCESS, NULL);
        }

        /* now do the delayed login */
        if (! say2_server_login(server))
                say2_server_message(server, NULL, SAY2_MSG_NEED_LOGIN, NULL);
}

/* _server_info
 * Server is telling us all about itself
 */
static void
_server_info(Say2Server *server, const EnamlBlock *server_info)
{
	const gchar *software, *version, *text;
	time_t	     uptime;
	
	software = enaml_get_string(server_info, "server", "?!@");
	version = enaml_get_string(server_info, "version", "?!@");
	text = enaml_get_string(server_info, "text", NULL);
	uptime = enaml_get_long(server_info, "uptime", 0);

	say2_server_message(server, NULL, SAY2_MSG_SERVER_INFO,
			    "software", software,
			    "version", version,
			    "uptime", uptime,
			    "text", text,
			    NULL);
}

/* _motd
 * Message of the day
 */
static void
_motd(Say2Server *server, const EnamlBlock *motd)
{
	const gchar *text;

	text = enaml_get_string(motd, "text", "");

	say2_server_message(server, NULL, SAY2_MSG_MOTD,
			    "text", text,
			    NULL);
}

/* _welcome
 * Our login was successful, so we're welcomed to the server
 */
static void
_welcome(Say2Server *server, const EnamlBlock *welcome)
{
	const gchar *nick, *id, *host;
	gboolean     attached, replicant;
	Say2User    *me;

	nick = enaml_get_string(welcome, "nick", "?!@");
	id = enaml_get_string(welcome, "id", NULL);
	host = enaml_get_string(welcome, "host", "?!@");
	attached = (enaml_type(welcome, "attached") != ENAML_KEY_UNDEF);
	replicant = (enaml_type(welcome, "replicant") != ENAML_KEY_UNDEF);

	/* Just check that we're not using screwy-server */
	if (id == NULL) {
		say2_server_message(server, NULL, SAY2_MSG_BAD_COMMAND,
				    "command", "welcome",
				    "text", _("Required paramter missing: id"),
				    NULL);
		return;
	}

	/* Welcome is sent to indicate we're logged in, but also to indicate
	 * a new replicant has logged in.
	 */
	if (say2_server_get_state(server) == SAY2_SERVER_VERIFIED) {
		/* Update the embedded user with the new information */
		me = say2_server_get_me(server);
		gtk_object_set(GTK_OBJECT(me), "id", id, "nick", nick, NULL);

		say2_server_message(server, NULL, SAY2_MSG_WELCOME,
				    "attached", attached,
				    NULL);

		/* Update the server status, this will also add the embedded
		 * user to the list of users and request more information
		 */
		say2_server_authenticated(server);
	}

	/* Announce new replicant */
	if (replicant) {
		say2_server_message(server, NULL, SAY2_MSG_REPLICANT_NEW,
				    "host", host,
				    NULL);
	}
}

/* _bad_nick
 * Our nickname wasn't accepted for some reason
 */
static void
_bad_nick(Say2Server *server, const EnamlBlock *bad_nick)
{
	const gchar *nick, *text;

	nick = enaml_get_string(bad_nick, "nick", "?!@");
	text = enaml_get_string(bad_nick, "text", "?!@");

	/* Send two messages, one indicating that we've got a bad nickname,
	 * the other indicating that we need to login again
	 */
	say2_server_message(server, NULL, SAY2_MSG_BAD_NICKNAME,
			    "nick", nick,
			    "text", text,
			    NULL);
	say2_server_message(server, NULL, SAY2_MSG_NEED_LOGIN, NULL);
}

/* _bad_address
 * Our address wasn't accepted for some reason
 */
static void
_bad_address(Say2Server *server, const EnamlBlock *bad_address)
{
	const gchar *address;

	address = enaml_get_string(bad_address, "address", "?!@");

	/* Send two messages, one indicating that our address was bad,
	 * the other indicating that we need to login again
	 */
	say2_server_message(server, NULL, SAY2_MSG_BAD_ADDRESS,
			    "address", address,
			    NULL);
	say2_server_message(server, NULL, SAY2_MSG_NEED_LOGIN, NULL);
}

/* _logout
 * Server signing off.  We'll just print it's message and sloppily wait
 * for it to close our connection.
 */
static void
_logout(Say2Server *server, const EnamlBlock *logout)
{
	const gchar *text, *detached;

	text = enaml_get_string(logout, "text", NULL);
	detached = enaml_get_string(logout, "detached", NULL);

	say2_server_message(server, NULL, SAY2_MSG_LOGOUT,
			    "detached", detached,
			    "text", text,
			    NULL);
}

/* _replicant_terminated
 * One of our replicants has been retired
 */
static void
_replicant_terminated(Say2Server *server,
		      const EnamlBlock *replicant_terminated)
{
	const gchar *host;

	host = enaml_get_string(replicant_terminated, "host", "?!@");

	say2_server_message(server, NULL, SAY2_MSG_REPLICANT_GONE,
			    "host", host,
			    NULL);
}

/* _ping
 * Server is checking we're not stoned, man
 */
static void
_ping(Say2Server *server, const EnamlBlock *ping)
{
	/* Although it's not required by the protocol, we send back anything
	 * in the ping block, just in case it was wanted.  Otherwise we sent
	 * it back as a flag.
	 */
	if (ping) {
		say2_server_send(server, NULL, "ping:%!B", ping);
	} else {
		say2_server_send(server, NULL, "ping");
	}
}

/* _signon
 * Someone we want notification for has signed on
 */
static void
_signon(Say2Server *server, const EnamlBlock *signon)
{
	const gchar *nick, *id, *email;
	time_t	     signon_time;

	nick = enaml_get_string(signon, "nick", "?!@");
	id = enaml_get_string(signon, "id", "?|@");
	email = enaml_get_string(signon, "email", "?!@");
	signon_time = enaml_get_long(signon, "time", time(NULL));

	/* This could be a signon of a user in our buddy list */
	/* FIXME */
	
	/* Treat it as a channel signon as well */
	if (say2_server_get_state(server) == SAY2_SERVER_ACTIVE)
		CHANNEL_signon(server, server->party_line, signon);
}

/* _signoff
 * Someone we want notification for has signed off
 */
static void
_signoff(Say2Server *server, const EnamlBlock *signoff)
{
	const gchar *nick, *id;
	time_t	     signoff_time;

	nick = enaml_get_string(signoff, "nick", "?!@");
	id = enaml_get_string(signoff, "id", "?!@");
	signoff_time = enaml_get_long(signoff, "time", time(NULL));

	/* This could be a signoff of a user in our buddy list */
	/* FIXME */

	/* Treat it as a channel signoff as well */
	if (say2_server_get_state(server) == SAY2_SERVER_ACTIVE)
		CHANNEL_signoff(server, server->party_line, signoff);
}

/* _prefs
 * Preferences have been changed
 */
static void
_prefs(Say2Server *server, const EnamlBlock *prefs)
{
	const gchar *tagline, *notify;
	gboolean     password, icon;

	tagline = enaml_get_string(prefs, "tagline", NULL);
	notify = enaml_get_string(prefs, "notify", NULL);
	password = (enaml_type(prefs, "password-changed") != ENAML_KEY_UNDEF);
	icon = (enaml_type(prefs, "icon") != ENAML_KEY_UNDEF);

	/* Save the new tagline */
	if (tagline) {
		gtk_object_set(GTK_OBJECT(say2_server_get_me(server)),
			       "tagline", tagline, NULL);
	}

	/* All goes out as one message */
	say2_server_message(server, NULL, SAY2_MSG_PREFS,
			    "tagline", tagline,
			    "notify", notify,
			    "icon", icon,
			    "password", password,
			    NULL);
}

/* _no_user_record
 * We sent a prefs or attributes message when we didn't have a user record
 */
static void
_no_user_record(Say2Server *server, const EnamlBlock *no_user_record)
{
	say2_server_message(server, NULL, SAY2_MSG_ERROR,
			    "text", _("You don't have a user record"),
			    NULL);
}

/* _status
 * The status of someone we want notification for has changed
 */
static void
_status(Say2Server *server, const EnamlBlock *status)
{
	const gchar *nick, *id, *mode, *text, *status_messages;
	time_t	     status_time;
	Say2User    *user;
	gboolean     messages;

	nick = enaml_get_string(status, "nick", "?!@");
	id = enaml_get_string(status, "id", "?!@");
	status_messages = enaml_get_string(status, "messages", "on");
	status_time = enaml_get_long(status, "time", time(NULL));
	mode = enaml_get_string(status, "mode", "offline");
	text = enaml_get_string(status, "text", NULL);

	if (!g_strcasecmp(status_messages, "on")) {
		messages = TRUE;
	} else {
		messages = FALSE;
	}

	user = say2_server_get_user(server, id);
	say2_user_set_status(user, mode, text, messages, status_time);

	say2_server_message(server, SAY2_CLIENT(user), SAY2_MSG_USER_STATUS,
			    "time", status_time,
			    "mode", mode,
			    "text", text,
			    "messages", messages,
			    NULL);
}

/* _change_nick
 * Someone we want notification for has changed their nickname
 */
static void
_change_nick(Say2Server *server, const EnamlBlock *change_nick)
{
	const gchar *id, *old_nick, *new_nick;
	Say2User    *user;

	id = enaml_get_string(change_nick, "id", "?!@");
	old_nick = enaml_get_string(change_nick, "old-nick", "?!@");
	new_nick = enaml_get_string(change_nick, "nick", "?!@");

	user = say2_server_get_user(server, id);
	gtk_object_set(GTK_OBJECT(user), "nick", new_nick, NULL);

	say2_server_message(server, SAY2_CLIENT(user), SAY2_MSG_USER_NICK,
			    "old-nick", old_nick,
			    NULL);
}

/* _members
 * Short list of online user or channel ids on the registry server
 */
static void
_members(Say2Server *server, const EnamlBlock *members)
{
	const gchar *type;
	EnamlList   *list;
	GString	    *buff;
	gint	     count, i;
	gboolean     is_user;
	
	type = enaml_get_string(members, "type", "all");
	list = enaml_get_list(members, "list");

	is_user = (!g_strcasecmp(type, "user") ? TRUE : FALSE);

	/* Build up the list */
	buff = g_string_new("");
	count = list ? enaml_list_length(list) : 0;
	for (i = 0; i < count; i++) {
		const gchar *id;
		Say2User    *user;

		/* Get the id of the member, we know its an id cause that's
		 * all we let them do
		 */
		id = enaml_list_get_string(list, i, "?!@");
		g_string_sprintfa(buff, " %s", id);

		if (!is_user)
			continue;

		/* If it's a user id, make sure it's in the users table and
		 * channel members list
		 */
		user = say2_server_get_user(server, id);
		if (!say2_channel_get_member(server->party_line, id))
			say2_channel_add_member(server->party_line, user);
	}

	if (!g_strcasecmp(type, "user")) {
		say2_server_message(server, SAY2_CLIENT(server->party_line),
				    SAY2_MSG_CHAN_MEMBERS,
				    "list", buff->str,
				    NULL);
	} else {
		say2_server_message(server, NULL, SAY2_MSG_MEMBERS,
				    "type", type,
				    "list", buff->str,
				    NULL);
	}

	g_string_free(buff, TRUE);
}

/* _begin_response
 * Start of a multiple-command response
 */
static void
_begin_response(Say2Server *server, const EnamlBlock *begin_response)
{
	const gchar *response;
	glong	     count, total;

	count = enaml_get_long(begin_response, "count", 0);
	total = enaml_get_long(begin_response, "total", 0);
	response = enaml_get_string(begin_response, "response", "?!@");
}

/* _end_response
 * End of a multiple-command response
 */
static void
_end_response(Say2Server *server, const EnamlBlock *end_response)
{
	const gchar *response;

	response = enaml_get_string(end_response, "response", "?!@");
}

/* _client_info
 * Information about a client on the system, use to update our records
 */
static void
_client_info(Say2Server *server, const EnamlBlock *client_info)
{
	const gchar *nick, *id, *type, *email, *tagline, *mode, *status;
	const gchar *software, *address, *client_messages;
	time_t	     login, idle, status_time, last_login, last_logout;
	gboolean     guest, offline;
	const char  *icon;
	int	     icon_sz;
	Say2Client  *client;

	/* Lots of possible fields, get them all and sort them out later */
	nick = enaml_get_string(client_info, "nick", "?!@");
	id = enaml_get_string(client_info, "id", "?!@");
	type = enaml_get_string(client_info, "type", "?!@");
	email = enaml_get_string(client_info, "email", "?!@");
	tagline = enaml_get_string(client_info, "tagline", NULL);
	mode = enaml_get_string(client_info, "mode", "?!@");
	status = enaml_get_string(client_info, "status", NULL);
	software = enaml_get_string(client_info, "client", NULL);
	address = enaml_get_string(client_info, "address", "?!@");
	login = enaml_get_long(client_info, "login", time(NULL));
	idle = enaml_get_long(client_info, "idle", 0);
	status_time = enaml_get_long(client_info, "status-time", 0);
	last_login = enaml_get_long(client_info, "last-login", time(NULL));
	last_logout = enaml_get_long(client_info, "last-logout", time(NULL));
	guest = (enaml_type(client_info, "guest") != ENAML_KEY_UNDEF);
	offline = (enaml_type(client_info, "offline") != ENAML_KEY_UNDEF);
	icon_sz = enaml_get_blob(client_info, "icon", &icon);
	client_messages = enaml_get_string(client_info, "messages", "on");

	/* We understand two types of client; users and channels */
	if (!g_strcasecmp(type, "user")) {
		gboolean messages;

		if (!g_strcasecmp(client_messages, "on")) {
			messages = TRUE;
		} else {
			messages = FALSE;
		}

		client = SAY2_CLIENT(say2_server_get_user(server, id));
	
		gtk_object_set(GTK_OBJECT(client), "nick", nick, NULL);
		
		/* Set user specific information, depends a bit whether the
		 * client is offline or not
		 */
		if (offline) {
			gtk_object_set(GTK_OBJECT(client),
				       "status", SAY2_USER_OFFLINE,
				       "login_at", last_login,
				       "logout_at", last_logout,
				       "messages", FALSE,
				       NULL);
		} else {
			say2_user_set_status(SAY2_USER(client), mode, status, 
					     messages,
					     time(NULL) - status_time);

			gtk_object_set(GTK_OBJECT(client),
				       "login_at", login, "guest", guest, NULL);
		}

	} else if (!g_strcasecmp(type, "channel")) {
		/* Can't handle channels yet */
		say2_server_message(server, NULL, SAY2_MSG_BAD_COMMAND,
				    "command", "client-info",
				    "text", _("Can't handle channels yet"),
				    NULL);
		return;

	} else {
		say2_server_message(server, NULL, SAY2_MSG_BAD_COMMAND,
				    "command", "client-info",
				    "text", _("Unknown client type"),
				    NULL);
		return;
	}

	/* Set the shared stuff */
	gtk_object_set(GTK_OBJECT(client),
		       "email", email,
		       "tagline", tagline,
		       "client", software,
		       NULL);
	if (icon_sz)
		say2_client_set_icon(client, icon, icon_sz);

	/* FIXME archive in list, perhaps */
	say2_server_message(server, client, SAY2_MSG_USER_INFO, NULL);
}

/* _not_present
 * A client that was whois'd isn't present
 */
static void
_not_present(Say2Server *server, const EnamlBlock *not_present)
{
	const gchar *nick, *id, *type;

	nick = enaml_get_string(not_present, "nick", NULL);
	id = enaml_get_string(not_present, "id", NULL);
	type = enaml_get_string(not_present, "type", "?!@");

	if (g_strcasecmp(type, "user")) {
		say2_server_message(server, NULL, SAY2_MSG_BAD_COMMAND,
				    "command", "not-present",
				    "text", _("Received for non-user"),
				    NULL);
		return;
	}

	if ((nick == NULL) && (id == NULL)) {
		say2_server_message(server, NULL, SAY2_MSG_BAD_COMMAND,
				    "command", "not-present",
				    "text", _("Required parameters missing"),
				    NULL);
		return;
	}

	/* Remove the user from the users list if they aren't
	 * present and it was something by id that errored (usually
	 * this happens just for guest users).  Also emit the invalid signal
	 * on that user.
	 *
	 * We generate different message events here because they
	 * sementically mean different things from our point of view
	 */
	if (id != NULL) {
		Say2User *user;

		user = say2_server_get_user(server, id);
		say2_server_message(server, SAY2_CLIENT(user),
				    SAY2_MSG_USER_INVALID, NULL);
		say2_client_invalid(SAY2_CLIENT(user));
		say2_server_remove_user(server, user);
	} else {
		say2_server_message(server, NULL, SAY2_MSG_NICK_NOT_PRESENT,
				    "nick", nick,
				    NULL);
	}
}

/* _no_such_client
 * This means a client that was whois'd doesn't exist
 */
static void
_no_such_client(Say2Server *server, const EnamlBlock *no_such_client)
{
	const gchar *id, *type;

	id = enaml_get_string(no_such_client, "id", "?!@");
	type = enaml_get_string(no_such_client, "type", "?!@");

	if (!g_strcasecmp(type, "user")) {
		Say2User *user;

		/* Remove the user from the users list, they don't exist.
		 * This usually happens when get_user() is called on a
		 * non-existant id.  Also emit the invalid signal on the user
		 * so certain things can drop it.
		 */

		user = say2_server_get_user(server, id);
		say2_server_message(server, SAY2_CLIENT(user),
				    SAY2_MSG_USER_INVALID, NULL);
		say2_client_invalid(SAY2_CLIENT(user));
		say2_server_remove_user(server, user);

	} else {
		say2_server_message(server, NULL, SAY2_MSG_CLIENT_INVALID,
				    "type", type,
				    "id", id,
				    NULL);
	}
}

/* _boot_user
 * Someone's just booted someone off the party line
 */
static void
_boot_user(Say2Server *server, const EnamlBlock *boot_user)
{
	/* Handled entirely in channel right now, because it's current
	 * semantic mean that it's an alternate signoff
	 */
	CHANNEL_boot_user(server, server->party_line, boot_user);
}

/* _client_attributes
 * We're received client attributes for a user from the server
 */
static void
_client_attributes(Say2Server *server, const EnamlBlock *client_attributes)
{
	const gchar *id, *type, *attribute;
	Say2Client  *client;
	EnamlBlock  *attributes;
	EnamlIter    iter;

	id = enaml_get_string(client_attributes, "id", "?!@");
	type = enaml_get_string(client_attributes, "type", "?!@");
	attributes = enaml_get_block(client_attributes, "attributes");

	if (!g_strcasecmp(type, "user")) {
		client = SAY2_CLIENT(say2_server_get_user(server, id));

	} else if (!g_strcasecmp(type, "channel")) {
		/* Can't handle channels yet */
		say2_server_message(server, NULL, SAY2_MSG_BAD_COMMAND,
				    "command", "client-attributes",
				    "text", _("Can't handle channels yet"),
				    NULL);
		return;

	} else {
		say2_server_message(server, NULL, SAY2_MSG_BAD_COMMAND,
				    "command", "client-attributes",
				    "text", _("Unknown client type"),
				    NULL);
		return;
	}

	/* Politely bomb out here if no attributes */
	if (attributes == NULL) {
		say2_server_message(server, client,
				    SAY2_MSG_USER_ATTRIBUTES, NULL);
		return;
	}

	for (attribute = enaml_iter_first(attributes, &iter); attribute != NULL;
	     attribute = enaml_iter_next(&iter)) {
		const gchar *value;
		
		/* We only care about string/number values */
		if (enaml_type(attributes, attribute) != ENAML_KEY_VALUE)
			continue;

		value = enaml_get_string(attributes, attribute, "");
		say2_client_set_attr(client, attribute, value);
	}
	
	say2_server_message(server, client, SAY2_MSG_USER_ATTRIBUTES, NULL);
}

/* _set_attributes
 * Indicates that attributes of our own have been set
 */
static void
_set_attributes(Say2Server *server, const EnamlBlock *set_attributes)
{
	EnamlList *attributes;
	GString   *buff;
	gint	   count, i;

	attributes = enaml_get_list(set_attributes, "attributes");

	/* Politely bomb out here if none set */
	if (attributes == NULL) {
		say2_server_message(server, NULL, SAY2_MSG_ATTRIBUTES_SET,
				    NULL);
		return;
	}

	/* Build up the list */
	buff = g_string_new("");
	count = enaml_list_length(attributes);
	for (i = 0; i < count; i++) {
		const gchar *attribute;

		attribute = enaml_list_get_string(attributes, i, "?!@");
		g_string_sprintfa(buff, " %s", attribute);
	}

	say2_server_message(server, NULL, SAY2_MSG_ATTRIBUTES_SET,
			    "list", buff->str,
			    NULL);
	g_string_free(buff, TRUE);
}

/* _say
 * Channel message
 */
static void
_say(Say2Server *server, const EnamlBlock *say)
{
        const gchar *id, *nick, *text, *act;
        time_t       say_time;

        id = enaml_get_string(say, "id", "?!@");
        nick = enaml_get_string(say, "nick", "?!@");
        say_time = (time_t)enaml_get_long(say, "time", time(NULL));
        text = enaml_get_string(say, "text", NULL);
        act = enaml_get_string(say, "act", NULL);

	/* Treat it as a channel say */
	if (say2_server_get_state(server) == SAY2_SERVER_ACTIVE)
		CHANNEL_say(server, server->party_line, say);
}

/* _message
 * Private message
 */
static void
_message(Say2Server *server, const EnamlBlock *message)
{
        const gchar *id, *nick, *to_id, *to_nick, *text, *act;
        time_t       message_time;
	gboolean     stored;
	Say2User    *user, *to_user, *me;

        id = enaml_get_string(message, "id", "?!@");
        nick = enaml_get_string(message, "nick", "?!@");
	to_id = enaml_get_string(message, "to-id", "?!@");
	to_nick = enaml_get_string(message, "to-nick", "?!@");
        message_time = (time_t)enaml_get_long(message, "time", time(NULL));
	stored = (enaml_type(message, "stored") != ENAML_KEY_UNDEF);
        text = enaml_get_string(message, "text", NULL);
        act = enaml_get_string(message, "act", NULL);

	user = say2_server_get_user(server, id);
	to_user = say2_server_get_user(server, to_id);
	me = say2_server_get_me(server);

	/* The server only ever sends us one message command, but we need to
	 * use it to generate the outgoing and incoming server messages.
	 *
	 * This is different from "say" because we generally don't want to
	 * distinguish between what we send and not.
	 */

	/* From me */
	if (user == me) {
		if (text != NULL) {
			say2_server_message(server, SAY2_CLIENT(to_user),
					    SAY2_MSG_USER_TEXT_TO,
					    "time", message_time,
					    "text", text,
					    "user", to_user,
					    "stored", stored,
					    NULL);
		}
		if (act != NULL) {
			say2_server_message(server, SAY2_CLIENT(to_user),
					    SAY2_MSG_USER_ACT_TO,
					    "time", message_time,
					    "text", act,
					    "user", to_user,
					    "stored", stored,
					    NULL);
		}
	}

	/* To me */
	if (to_user == me) {
		if (text != NULL) {
			say2_server_message(server, SAY2_CLIENT(user),
					    SAY2_MSG_USER_TEXT,
					    "time", message_time,
					    "text", text,
					    NULL);
		}
		if (act != NULL) {
			say2_server_message(server, SAY2_CLIENT(user),
					    SAY2_MSG_USER_ACT,
					    "time", message_time,
					    "text", act,
					    NULL);
		}
	}
}

/* _wall
 * Global message
 */
static void
_wall(Say2Server *server, const EnamlBlock *wall)
{
        const gchar *id, *nick, *text;
        time_t       wall_time;
	Say2User    *user;

        id = enaml_get_string(wall, "id", "?!@");
        nick = enaml_get_string(wall, "nick", "?!@");
        wall_time = (time_t)enaml_get_long(wall, "time", time(NULL));
        text = enaml_get_string(wall, "text", "");

	user = say2_server_get_user(server, id);

	say2_server_message(server, SAY2_CLIENT(user), SAY2_MSG_WALL,
			    "time", wall_time,
			    "user", user,
			    "text", text,
			    NULL);
}
