/* meko
 * Copyright (C) 2001 Scott James Remnant <scott@netsplit.com>.
 * All Rights Reserved.
 *
 * say2-server.c
 *  - Say2Server class
 *  - An object to hold information about a connection to a registration
 *    server on the say2 network
 * -- 
 * $Id: say2-server.c,v 1.26.4.3 2001/11/24 01:25:51 scott Exp $
 * 
 * This file is distributed according to the GNU General Public
 * License.  For full details, read the top of 'main.c' or the
 * file called COPYING that was distributed with this code.
 */

#include <meko.h>

#include <stdio.h> /* only needed for a hack */
#include <stdarg.h>
#include <stdlib.h>
#include <netdb.h>

#include <glib.h>
#include <gtk/gtkobject.h>
#include <gtk/gtksignal.h>
#include <libgnome/gnome-defs.h>
#include <libgnome/gnome-i18n.h>
#include <libgnome/gnome-util.h> /* only needed for a hack */
#include <say2/say2.h>

#include "say2/say2-client.h"
#include "say2/say2-user.h"
#include "say2/say2-channel.h"
#include "say2/say2-message.h"
#include "say2/say2-server-commands.h"

#include "say2-server.h"

/* Arguments
 * gchar *	hostname	Hostname to connect to
 * gint		port		Port to connect to
 * gchar *	port_str	(as above, but converts)
 * gchar *	password	Password to use
 * gboolean	replicant	Don't kick off existing connections
 * gboolean     ssl             Use SSL when connecting
 */
enum {
	ARG_0,
	ARG_HOSTNAME,
	ARG_PORT,
	ARG_PORT_STR,
	ARG_PASSWORD,
	ARG_REPLICANT,
        ARG_SSL,
};

/* Signals */
enum {
	CONNECTED,
	VERIFIED,
	AUTHENTICATED,
	DISCONNECTED,
	SIGNON,
	SIGNOFF,
	MESSAGE,
	LAST_SIGNAL
};


/* Squelch information */
typedef struct _squelch      Squelch;
typedef struct _squelch_info SquelchInfo;
struct _squelch {
	/* List of SquelchInfo structures */
	GList *info;

	/* Callback function and user data.  Same prototype and semantics as
	 * the "message" signal function (deliberately :)
	 */
	SquelchFunc func;
	gpointer    user_data;
};
struct _squelch_info {
	/* Squelch on specific client and/or message type.
	 * NULL or SAY2_MSG_ANY means it'll match anything.
	 */
	Say2Client      *client;
	Say2MessageType  message_type;
};


/* Forward prototypes for required functions */
static void say2_server_init(Say2Server *);
static void say2_server_class_init(Say2ServerClass *);

/* Forward prototypes for GtkObject methods */
static void say2_server_set_arg(GtkObject *, GtkArg *, guint);
static void say2_server_get_arg(GtkObject *, GtkArg *, guint);
static void say2_server_destroy(GtkObject *);

/* Forward prototypes for Say2Server methods */
static void say2_real_server_verified(Say2Server *);
static void say2_real_server_authenticated(Say2Server *);
static void say2_real_server_disconnected(Say2Server *);

/* Forward prototypes for internal functions */
static void	_init_party_line(Say2Server *);
static void	_add_attr(const gchar *, const gchar *, EnamlBlock *);
static void	_add_user(Say2Server *, Say2User *);
static void     _find_str(const gchar *, Say2User *, gpointer[]);
static gboolean _unref_hash_val(gpointer, gpointer, gpointer);
static void	_sock_ready_cb(Socket *, gint);
static void	_sock_eof_cb(Socket *);
static void	_sock_read_cb(Socket *);


/* Our parent class */
static GtkObjectClass *parent_class = NULL;

/* Signal array */
static guint server_signals[LAST_SIGNAL] = { 0 };


/* say2_server_get_type
 * Returns our GtkType id, allocating it if necessary
 */
guint
say2_server_get_type(void)
{
	static guint server_type = 0;

	if (!server_type) {
		static const GtkTypeInfo server_info = {
			"Say2Server",
			sizeof(Say2Server),
			sizeof(Say2ServerClass),
			(GtkClassInitFunc)say2_server_class_init,
			(GtkObjectInitFunc)say2_server_init,
			/* reserved_1 */ NULL,
			/* reserved_2 */ NULL,
			(GtkClassInitFunc)NULL
		};

		server_type = gtk_type_unique(GTK_TYPE_OBJECT, &server_info);
	}

	return server_type;
}

/* say2_server_class_init
 * Initialise the object class, setting our arguments, signals and overriding
 * methods of parent object classes
 */
static void
say2_server_class_init(Say2ServerClass *klass)
{
	GtkObjectClass *object_class;

	g_return_if_fail(klass != NULL);

	object_class = (GtkObjectClass *)klass;
	parent_class = gtk_type_class(GTK_TYPE_OBJECT);

	gtk_object_add_arg_type("Say2Server::hostname", GTK_TYPE_STRING,
				GTK_ARG_READWRITE, ARG_HOSTNAME);
	gtk_object_add_arg_type("Say2Server::port", GTK_TYPE_INT,
				GTK_ARG_READWRITE, ARG_PORT);
	gtk_object_add_arg_type("Say2Server::port_str", GTK_TYPE_STRING,
				GTK_ARG_WRITABLE, ARG_PORT_STR);
	gtk_object_add_arg_type("Say2Server::password", GTK_TYPE_STRING,
				GTK_ARG_WRITABLE, ARG_PASSWORD);
	gtk_object_add_arg_type("Say2Server::replicant", GTK_TYPE_BOOL,
				GTK_ARG_READWRITE, ARG_REPLICANT);
        gtk_object_add_arg_type("Say2Server::ssl", GTK_TYPE_BOOL,
                                GTK_ARG_READWRITE, ARG_SSL);

	server_signals[CONNECTED] = gtk_signal_new(
		"connected", GTK_RUN_FIRST, object_class->type,
		GTK_SIGNAL_OFFSET(Say2ServerClass, connected),
		gtk_marshal_NONE__NONE,
		GTK_TYPE_NONE, 0
	);

	server_signals[VERIFIED] = gtk_signal_new(
		"verified", GTK_RUN_FIRST, object_class->type,
		GTK_SIGNAL_OFFSET(Say2ServerClass, verified),
		gtk_marshal_NONE__NONE,
		GTK_TYPE_NONE, 0
	);


	server_signals[AUTHENTICATED] = gtk_signal_new(
		"authenticated", GTK_RUN_FIRST, object_class->type,
		GTK_SIGNAL_OFFSET(Say2ServerClass, authenticated),
		gtk_marshal_NONE__NONE,
		GTK_TYPE_NONE, 0
	);

	server_signals[DISCONNECTED] = gtk_signal_new(
		"disconnected", GTK_RUN_LAST, object_class->type,
		GTK_SIGNAL_OFFSET(Say2ServerClass, disconnected),
		gtk_marshal_NONE__NONE,
		GTK_TYPE_NONE, 0
	);

	server_signals[SIGNON] = gtk_signal_new(
		"signon", GTK_RUN_FIRST, object_class->type,
		GTK_SIGNAL_OFFSET(Say2ServerClass, signon),
		gtk_marshal_NONE__POINTER,
		GTK_TYPE_NONE, 1, GTK_TYPE_POINTER
	);

	server_signals[SIGNOFF] = gtk_signal_new(
		"signoff", GTK_RUN_FIRST, object_class->type,
		GTK_SIGNAL_OFFSET(Say2ServerClass, signoff),
		gtk_marshal_NONE__POINTER,
		GTK_TYPE_NONE, 1, GTK_TYPE_POINTER
	);

	server_signals[MESSAGE] = gtk_signal_new(
		"message", GTK_RUN_FIRST, object_class->type,
		GTK_SIGNAL_OFFSET(Say2ServerClass, message),
		gtk_marshal_NONE__POINTER_INT_POINTER, GTK_TYPE_NONE,
		3, GTK_TYPE_POINTER, GTK_TYPE_INT, GTK_TYPE_POINTER
	);

	gtk_object_class_add_signals(object_class, server_signals, LAST_SIGNAL);

	object_class->set_arg = say2_server_set_arg;
	object_class->get_arg = say2_server_get_arg;
	object_class->destroy = say2_server_destroy;

	klass->connected = NULL;
	klass->verified = say2_real_server_verified;
	klass->authenticated = say2_real_server_authenticated;
	klass->disconnected = say2_real_server_disconnected;
	klass->signon = NULL;
	klass->signoff = NULL;
	klass->message = NULL;
}

/* say2_server_init
 * Initialise an object instance, very little to do here other than set default
 * values
 */
static void
say2_server_init(Say2Server *server)
{
	g_return_if_fail(server != NULL);

	server->hostname = NULL;
	server->port = 0;
	server->password = NULL;
	server->replicant = 0;
        server->ssl = 0;

	server->flags = 0;
	server->state = SAY2_SERVER_INACTIVE;
	server->sock = NULL;
	server->party_line = NULL;
	server->squelches = NULL;

	/* Create a user record for ourselves, and create a hash table.
	 * We'll worry about actually putting ourselves into it once we're
	 * sure of our ID.
	 */
	server->me = say2_user_new(NULL);
	gtk_object_ref(GTK_OBJECT(server->me));
	gtk_object_sink(GTK_OBJECT(server->me));
	server->users = g_hash_table_new(g_str_hash, g_str_equal);
}

/* say2_server_new
 * Create a new object instance, just a wrapper around gtk_type_new really
 */
Say2Server *
say2_server_new(void)
{
	Say2Server *server;

	server = SAY2_SERVER(gtk_type_new(SAY2_TYPE_SERVER));

	return server;
}

/* say2_server_destroy
 * Destroy an object instance, freeing and unreferencing any data we've
 * been hanging on to
 */
static void
say2_server_destroy(GtkObject *object)
{
	Say2Server *server;

	g_return_if_fail(object != NULL);
	g_return_if_fail(SAY2_IS_SERVER(object));

	server = SAY2_SERVER(object);

	if (server->state > SAY2_SERVER_INACTIVE)
		say2_server_disconnect(server);

	g_free(server->hostname);
	g_free(server->password);

	/* Finish with our own user record */
	gtk_object_unref(GTK_OBJECT(server->me));
	server->me = NULL;

	/* Destroy the hash, it's cleared by disconnect() */
	g_hash_table_destroy(server->users);
	server->users = NULL;

	/* Chain up */
	if (GTK_OBJECT_CLASS(parent_class)->destroy)
		GTK_OBJECT_CLASS(parent_class)->destroy(object);
}

/* say2_server_set_arg
 * Set object arguments
 */
static void
say2_server_set_arg(GtkObject *object, GtkArg *arg, guint arg_id)
{
	Say2Server *server;

	g_return_if_fail(object != NULL);
	g_return_if_fail(SAY2_IS_SERVER(object));

	server = SAY2_SERVER(object);
	switch (arg_id) {
	case ARG_HOSTNAME:
		g_free(server->hostname);
		server->hostname = g_strdup(GTK_VALUE_STRING(*arg));
		break;
	case ARG_PORT:
		server->port = GTK_VALUE_INT(*arg);
		break;
	case ARG_PORT_STR:
		if (GTK_VALUE_STRING(*arg)) {
			struct servent *servent;

			/* Look up the well known port name */
			servent = getservbyname(GTK_VALUE_STRING(*arg), "tcp");
			if (servent) {
				server->port = servent->s_port;
			} else {
				server->port = atoi(GTK_VALUE_STRING(*arg));
			}
		} else {
			server->port = 0;
		}
	case ARG_PASSWORD:
		g_free(server->password);
		server->password = g_strdup(GTK_VALUE_STRING(*arg));
		break;
	case ARG_REPLICANT:
		server->replicant = GTK_VALUE_BOOL(*arg);
		break;
        case ARG_SSL:
                server->ssl = GTK_VALUE_BOOL(*arg);
                break;
	}
}

/* say2_server_get_arg
 * Get object arguments
 */
static void
say2_server_get_arg(GtkObject *object, GtkArg *arg, guint arg_id)
{
	Say2Server *server;

	g_return_if_fail(object != NULL);
	g_return_if_fail(SAY2_IS_SERVER(object));

	server = SAY2_SERVER(object);
	switch (arg_id) {
	case ARG_HOSTNAME:
		GTK_VALUE_STRING(*arg) = server->hostname;
		break;
	case ARG_PORT:
		GTK_VALUE_INT(*arg) = server->port;
		break;
	case ARG_REPLICANT:
		GTK_VALUE_BOOL(*arg) = server->replicant;
		break;
        case ARG_SSL:
                GTK_VALUE_BOOL(*arg) = server->ssl;
                break;
	default:
		arg->type = GTK_TYPE_INVALID;
		break;
	}
}

/* say2_real_server_verified
 * Update our state
 */
static void
say2_real_server_verified(Say2Server *server)
{
	g_return_if_fail(server != NULL);

	server->state = SAY2_SERVER_VERIFIED;
}

/* _init_party_line
 * HACK until we get real channel support, create the party line
 */
static void
_init_party_line(Say2Server *server)
{
	FILE  *icon_fd;
	gchar *icon_filename;
	char   icon[4096];
	int    icon_sz;

	g_return_if_fail(server != NULL);

	/* Load the party line icon */
	icon_filename = g_concat_dir_and_file(MEKO_IMAGESDIR, "party-line.png");
	icon_fd = fopen(icon_filename, "r");
	icon_sz = fread(icon, 1, sizeof(icon), icon_fd);
	fclose(icon_fd);
	g_free(icon_filename);

	/* Create the implicit party line channel and get the user list */
	server->party_line = SAY2_CHANNEL(gtk_object_new(
		SAY2_TYPE_CHANNEL,
		"id", "party-line",
		"email", "say2-dev@lag.net",
		"tagline", "We put the KER into KERwhee!",
		NULL
	));
	gtk_object_ref(GTK_OBJECT(server->party_line));
	gtk_object_sink(GTK_OBJECT(server->party_line));
	say2_client_set_icon(SAY2_CLIENT(server->party_line), icon, icon_sz);
	say2_server_send(server, "list-clients", "type:%s by-id", "user");
	say2_server_add_squelch(server, NULL, NULL, 1,
				NULL, SAY2_MSG_CHAN_MEMBERS);
}

/* _add_attr
 * Add the attribute given to the EnamlBlock
 */
static void
_add_attr(const gchar *key, const gchar *value, EnamlBlock *block)
{
	enaml_add_string(block, key, value, ENAML_COPY);
}

/* say2_real_server_authenticated
 * Ok, authenticated, lets play some Music with Rocks in!
 * Cool, as long as it's not Savage Garden again!
 */
static void
say2_real_server_authenticated(Say2Server *server)
{
	const gchar *id, *tagline, *icon;
	int	     icon_sz;
	gboolean     guest;

	g_return_if_fail(server != NULL);

	/* Change state to active and emit signal */
	server->state = SAY2_SERVER_ACTIVE;

	/* Grab some basic stuff about our user record */
	icon_sz = say2_client_get_icon(SAY2_CLIENT(server->me), (char **)&icon);
	gtk_object_get(GTK_OBJECT(server->me),
		       "id", &id, "tagline", &tagline, "guest", &guest, NULL);

	/* Update our user record on the server with any new information
	 * given apon login. 
	 */
	if (!guest) {
		EnamlBlock *set_attributes;

		/* Set preferences */
		say2_server_send(server, "prefs", "notify tagline:%s icon:%x",
				 tagline, (icon_sz ? icon : NULL), icon_sz);
		say2_server_add_squelch(server, NULL, NULL, 1,
					NULL, SAY2_MSG_PREFS);

		/* Set attributes */
		set_attributes = enaml_new();
		say2_client_foreach_attr(SAY2_CLIENT(server->me),
					 (GHFunc)_add_attr, set_attributes);
		say2_server_send(server, NULL,
				 "set-attributes:%B", set_attributes);
		say2_server_add_squelch(server, NULL, NULL, 1,
					NULL, SAY2_MSG_ATTRIBUTES_SET);
	}

	/* Try and get our attributes as well */
	say2_server_send(server, "get-attributes", "id:%s type:%s", id, "user");
	say2_server_add_squelch(server, NULL, NULL, 2,
				SAY2_CLIENT(server->me),
					SAY2_MSG_USER_ATTRIBUTES,
				SAY2_CLIENT(server->me), SAY2_MSG_USER_INVALID);

	/* Add our own record to the list (we know we've got a valid id). */
	_add_user(server, server->me);

	/* Hack the party line for now */
	_init_party_line(server);
}

/* _unref_hash_val
 * Unreferences data in the users hash table
 */
static gboolean
_unref_hash_val(gpointer key, gpointer value, gpointer user_data)
{
	g_return_val_if_fail(value != NULL, TRUE);
	g_return_val_if_fail(SAY2_IS_CLIENT(value), TRUE);

	gtk_object_unref(GTK_OBJECT(value));
	
	return TRUE;
}

/* say2_real_server_disconnected
 * Disconnect from the registration server
 */
static void
say2_real_server_disconnected(Say2Server *server)
{
	g_return_if_fail(server != NULL);

	if (server->sock) {
		/* Don't handle eof and the like here just in case */
		socket_set_eof_fn((Socket *)server->sock, NULL);
		socket_set_read_fn((Socket *)server->sock, NULL);
		
		/* Graceful disconnect */
		say2_server_send(server, "logout", "");
		
		/* Close the socket */
		socket_close((Socket *)server->sock);
		server->sock = NULL;
	}

	/* Don't need the channel anymore */
	if (server->party_line) {
		gtk_object_unref(GTK_OBJECT(server->party_line));
		server->party_line = NULL;
	}
	
	/* Clear the users hash */
	g_hash_table_foreach_remove(server->users, _unref_hash_val, NULL);

	/* Destroy the squelches list */
	while (server->squelches) {
		Squelch *squelch;

		squelch = (Squelch *)server->squelches->data;
		while (squelch->info) {
			SquelchInfo *squelch_info;

			squelch_info = (SquelchInfo *)squelch->info->data;
			squelch->info = g_list_remove(squelch->info,
						      squelch_info);
			g_free(squelch_info);
		}
		
		server->squelches = g_list_remove(server->squelches, squelch);
		g_free(squelch);
	}

	/* Finish up */
	server->state = SAY2_SERVER_INACTIVE;
}

/* say2_server_get_state
 * Return the state of the server
 */
Say2ServerState
say2_server_get_state(Say2Server *server)
{
	g_return_val_if_fail(server != NULL, SAY2_SERVER_INACTIVE);

	return server->state;
}

/* say2_server_get_me
 * Return information about ourselves
 */
Say2User *
say2_server_get_me(Say2Server *server)
{
	g_return_val_if_fail(server != NULL, NULL);

	return server->me;
}

/* _add_user
 * Add a user we already know about to the users list
 */
static void
_add_user(Say2Server *server, Say2User *user)
{
	Say2User    *existing_record;
	const gchar *id;

	g_return_if_fail(server != NULL);
	g_return_if_fail(user != NULL);

	gtk_object_get(GTK_OBJECT(user), "id", &id, NULL);

	/* Make sure they don't already exist */
	existing_record = g_hash_table_lookup(server->users, id);
	g_return_if_fail(existing_record == NULL);

	/* Add the record */
	g_hash_table_insert(server->users, (gpointer)id, user);
	gtk_object_ref(GTK_OBJECT(user));
	gtk_object_sink(GTK_OBJECT(user));

	/* Find out all about them */
	say2_server_send(server, "whois", "id:%s type:%s icon", id, "user");
	say2_server_add_squelch(server, NULL, NULL, 2,
				SAY2_CLIENT(user), SAY2_MSG_USER_INFO,
				SAY2_CLIENT(user), SAY2_MSG_USER_INVALID);
}

/* say2_server_get_user
 * Get the record for a user id off the users list.  If they're not on there
 * we assume it's a valid id, create the record, put it on the list and ask
 * the server to give us the information asynchronously.  If the record is
 * important and needs to stay around some time, the "changed" signal lets
 * you know when information arrives and the "invalid" signal lets you know
 * if the server doesn't know anything about it.
 */
Say2User *
say2_server_get_user(Say2Server *server, const gchar *id)
{
	Say2User *user;

	g_return_val_if_fail(server != NULL, NULL);
	g_return_val_if_fail(id != NULL, NULL);

	user = g_hash_table_lookup(server->users, id);
	if (user == NULL) {
		/* Create and add */
		user = say2_user_new(id);
		_add_user(server, user);
	}

	return user;
}

/* say2_server_remove_user
 * Remove a user from the users table.  This should be used sparingly, there's
 * rarely a need to do this.
 */
void
say2_server_remove_user(Say2Server *server, Say2User *user)
{
	const gchar *id;

	g_return_if_fail(server != NULL);
	g_return_if_fail(user != NULL);

	gtk_object_get(GTK_OBJECT(user), "id", &id, NULL);
	g_hash_table_remove(server->users, id);

	gtk_object_unref(GTK_OBJECT(user));
}

/* _find_str
 * Check the string given in the parameters, and if it matches this user, set
 * the pointer too
 */
static void
_find_str(const gchar *id, Say2User *user, gpointer params[])
{
	const gchar *nick;

	g_return_if_fail(id != NULL);
	g_return_if_fail(user != NULL);
	g_return_if_fail(params[0] != NULL);

	if (params[1] != NULL)
		return;

	gtk_object_get(GTK_OBJECT(user), "nick", &nick, NULL);

	if (!g_strcasecmp(nick, params[0]))
		params[1] = user;

	return;
}

/* say2_server_get_user_by_name
 * Return a user from the users table who's id or nickname matches
 * the one given
 */
Say2User *
say2_server_get_user_by_name(Say2Server *server, const gchar *str)
{
	gpointer params[2];

	g_return_val_if_fail(server != NULL, NULL);
	g_return_val_if_fail(str != NULL, NULL);

	params[0] = (gpointer)str;
	params[1] = NULL;

	g_hash_table_foreach(server->users, (GHFunc)_find_str, params);

	if (params[1] == NULL) {
		gchar *id;

		id = g_strdup(str);
		g_strdown(id);
		params[1] = say2_server_get_user(server, id);
		g_free(id);
	}
	return params[1];
}

/* say2_server_connect
 * Connect to a Say2 registration server
 */
void
say2_server_connect(Say2Server *server)
{
	g_return_if_fail(server != NULL);
	g_return_if_fail(server->state == SAY2_SERVER_INACTIVE);
	g_return_if_fail(server->hostname != NULL);
	g_return_if_fail(server->port != 0);

	say2_server_message(server, NULL, SAY2_MSG_CONNECTING, NULL);

	/* Create the buffered socket */
	server->sock = socket_new(NULL, 0);
	if (server->sock == NULL) {
		say2_server_message(server, NULL, SAY2_MSG_CONNECTION_FAILED,
				    "text", _("Failed to make socket"),
				    NULL);
		say2_server_disconnect(server);
		return;
	}
	socket_set_data((Socket *)server->sock, server);

	/* Attempt the connection */
	if (!socket_connect((Socket *)server->sock, server->hostname,
			    server->port, _sock_ready_cb)) {
		say2_server_message(server, NULL, SAY2_MSG_CONNECTION_FAILED,
				    "text", _("Failed to establish connection"),
				    NULL);

		socket_close((Socket *)server->sock);
		server->sock = NULL;
		
		say2_server_disconnect(server);
		return;
	}
	
	server->state = SAY2_SERVER_CONNECTING;
}

gboolean
say2_server_begin_ssl(Say2Server *server)
{
#ifdef HAVE_OPENSSL
        return (socket_begin_ssl((Socket *)server->sock, SOCKET_SSL_CLIENT) ? TRUE : FALSE);
#else
	return FALSE;
#endif
}

/* say2_server_disconnect
 * Emit the disconnected signal
 */
void
say2_server_disconnect(Say2Server *server)
{
	g_return_if_fail(server != NULL);
	g_return_if_fail(server->state > SAY2_SERVER_INACTIVE);

	say2_server_message(server, NULL, SAY2_MSG_DISCONNECTED, NULL);
	gtk_signal_emit(GTK_OBJECT(server), server_signals[DISCONNECTED]);
}

/* say2_server_login
 * Login to a registration server.  Returns FALSE if there isn't enough
 * information available to do so.
 */
gboolean
say2_server_login(Say2Server *server)
{
	const gchar *id, *nick, *email, *tagline, *password;
	gboolean     guest, replicant;

	g_return_val_if_fail(server != NULL, FALSE);
	g_return_val_if_fail(server->state == SAY2_SERVER_VERIFIED, FALSE);

	/* Get login information from the embedded user information */
	gtk_object_get(GTK_OBJECT(server->me),
		       "guest", &guest,
		       "id", &id,
		       "nick", &nick,
		       "email", &email,
		       "tagline", &tagline,
		       NULL);

	/* Sanity check for required information, don't login if we don't. */
	if (guest) {
		if ((nick == NULL) || (email == NULL))
			return FALSE;

		/* Make sure we don't send an id or password when logging
		 * in as guest
		 */
		id = password = NULL;
		replicant = FALSE;
	} else {
		if ((id == NULL) || (server->password == NULL))
			return FALSE;

		/* Send the right password */
		password = server->password;
		replicant = server->replicant;
	}

	/* Send a login block */
	say2_server_send(server, "login",
			 "id:%s password:%s nick:%s email:%s tagline:%s "
			 "client:%s replicant:%F",
			 id, password, nick, email, tagline,
			 APPNAME " " VERSION, replicant);

	return TRUE;
}

/* say2_server_verified
 * Emit the verified signal
 */
void
say2_server_verified(Say2Server *server)
{
	g_return_if_fail(server != NULL);
	g_return_if_fail(server->state == SAY2_SERVER_CONNECTED);

	gtk_signal_emit(GTK_OBJECT(server), server_signals[VERIFIED]);
}

/* say2_server_authenticated
 * Emit the authenticated signal
 */
void
say2_server_authenticated(Say2Server *server)
{
	g_return_if_fail(server != NULL);
	g_return_if_fail(server->state == SAY2_SERVER_VERIFIED);

	gtk_signal_emit(GTK_OBJECT(server), server_signals[AUTHENTICATED]);
}

/* say2_server_send
 * Send a message by building the enaml block and sending it along our
 * socket
 */
void
say2_server_send(Say2Server *server, const gchar *tag, const gchar *fmt, ...)
{
	EnamlBlock *out;
	va_list     args;

	va_start(args, fmt);
	out = enaml_vbuildf(tag, fmt, args);
	va_end(args);

	send_enaml_block((Socket *)server->sock, out);
	enaml_free(out);
}

/* say2_server_message
 * Collect arguments up into a hash and dispatch the message.  After the
 * intial three arguments, arguments come in pairs GTK-style, first a string
 * key then a pointer value (which musn't be volatile)
 */
void
say2_server_message(Say2Server *server, Say2Client *client,
		    Say2MessageType message_type, ...)
{
	va_list      args;
	const gchar *key;
	GHashTable  *hash;
	GList	    *list;
	gboolean     send_msg;

	g_return_if_fail(server != NULL);
	g_return_if_fail(message_type > SAY2_MSG_ANY);
	g_return_if_fail(message_type < SAY2_MSG_LAST);

	va_start(args, message_type);
	hash = g_hash_table_new(g_str_hash, g_str_equal);

	/* Collect the argument pairs into a hash, we only use this for the
	 * duration of the dispatch call, so as long as their not volatile
	 * (which you deliberately ask not to get) we don't have to dup()
	 * anything
	 */
	while ((key = va_arg(args, const gchar *)) != NULL) {
		gpointer val;

		val = va_arg(args, gpointer);
		g_hash_table_insert(hash, (gpointer)key, val);
	}

	/* Look through the squelches list */
	send_msg = TRUE;
	for (list = g_list_first(server->squelches); list;
	     list = g_list_next(list)) {
		Squelch *squelch;
		GList   *info_list;

		squelch = list->data;
		for (info_list = g_list_first(squelch->info); info_list;
		     info_list = g_list_next(info_list)) {
			SquelchInfo *squelch_info;

			squelch_info = info_list->data;

			/* See whether it matches */
			if ((squelch_info->client != NULL)
			    && (squelch_info->client != client))
				continue;
			if ((squelch_info->message_type != SAY2_MSG_ANY)
			    && (squelch_info->message_type != message_type))
				continue;

			/* It does, mark and break */
			send_msg = FALSE;
			break;
		}

		/* If send_msg is still true after that, move on to the next */
		if (send_msg)
			continue;
		
		/* Free up this item and remove it from the list */
		while (squelch->info) {
			SquelchInfo *squelch_info;

			squelch_info = (SquelchInfo *)squelch->info->data;
			squelch->info = g_list_remove(squelch->info,
						      squelch_info);
			g_free(squelch_info);
		}
		server->squelches = g_list_remove(server->squelches, squelch);

		/* Dispatch the function */
		if (squelch->func) {
			squelch->func(server, client, message_type, hash,
				      squelch->user_data);
		}

		/* Free up and break */
		g_free(squelch);
		break;
	}

	/* Emit the message signal, unless it was squelched */
	if (send_msg) {
		gtk_signal_emit(GTK_OBJECT(server), server_signals[MESSAGE],
				client, message_type, hash);
	}

	/* Done */
	g_hash_table_destroy(hash);
	va_end(args);
}

/* say2_server_add_squelch
 * Prevent one or more types of message signal from being emitted in the usual
 * way.  An optional callback function can be given to actually do some
 * work on receipt of the function.  The arguments must be the number of pairs
 * of Say2Client * and Say2MessageType as specified by num_pairs.
 */
void
say2_server_add_squelch(Say2Server *server,
			SquelchFunc func, gpointer user_data,
			gint num_pairs, ...)
{
	Squelch *squelch;
	va_list  args;
	gint     i;

	g_return_if_fail(server != NULL);

	va_start(args, num_pairs);

	squelch = g_new0(Squelch, 1);
	squelch->info = NULL;
	squelch->func = func;
	squelch->user_data = user_data;

	for (i = 0; i < num_pairs; i++) {
		SquelchInfo *squelch_info;

		squelch_info = g_new0(SquelchInfo, 1);
		squelch_info->client = va_arg(args, Say2Client *);
		squelch_info->message_type = va_arg(args, Say2MessageType);

		squelch->info = g_list_prepend(squelch->info, squelch_info);
	}
	squelch->info = g_list_reverse(squelch->info);
	server->squelches = g_list_append(server->squelches, squelch);

	va_end(args);
}

/* _sock_ready_cb
 * Connection to the remote server has completed, whether successfully or
 * not
 */
static void
_sock_ready_cb(Socket *sock, gint success)
{
	Say2Server *server;

	g_return_if_fail(sock != NULL);

	server = socket_get_data(sock);
	g_return_if_fail(server != NULL);
	g_return_if_fail((Socket *)server->sock == sock);

	/* Connection failed */
	if (!success) {
		say2_server_message(server, NULL, SAY2_MSG_CONNECTION_FAILED,
				    "text", _("Unable to connect"),
				    NULL);

		socket_close(server->sock);
		server->sock = NULL;

		say2_server_disconnect(server);
		return;
	}

	/* Connection succeeeded
	 * Change the state, set callbacks and emit the signal.
	 */
	say2_server_message(server, NULL, SAY2_MSG_CONNECTED, NULL);
	server->state = SAY2_SERVER_CONNECTED;
	socket_set_eof_fn(server->sock, _sock_eof_cb);
	socket_set_read_fn(server->sock, _sock_read_cb);
	gtk_signal_emit(GTK_OBJECT(server), server_signals[CONNECTED]);
}

/* _sock_eof_cb
 * Connection to the remote server was dropped
 */
static void
_sock_eof_cb(Socket *sock)
{
	Say2Server *server;

	g_return_if_fail(sock != NULL);

	server = socket_get_data(sock);
	g_return_if_fail(server != NULL);
	g_return_if_fail((Socket *)server->sock == sock);

	socket_close(server->sock);
	server->sock = NULL;

	say2_server_disconnect(server);
}

/* _sock_read_cb
 * Handle data received from remote servers
 */
static void
_sock_read_cb(Socket *sock)
{
	Say2Server *server;

	g_return_if_fail(sock != NULL);

	server = socket_get_data(sock);
	g_return_if_fail(server != NULL);
	g_return_if_fail((Socket *)server->sock == sock);

	/* Eat all complete lines in one cycle */
	for (;;) {
		EnamlBlock  *block;
		EnamlIter    iter;
		gchar	    *line;
		const gchar *key;

		line = socket_getline(server->sock);
		if (!line)
			break;

		/* Parse the enaml in the line */
		block = enaml_parse(line);
		if (!block) {
			say2_server_message(server, NULL, SAY2_MSG_BAD_ENAML,
					    "enaml", line,
					    NULL);
			goto loop;
		}

		/* Parse all the keys and blocks that were on that line */
		for (key = enaml_iter_first(block, &iter); key != NULL;
		     key = enaml_iter_next(&iter)) {
			const EnamlBlock *subblock;

			subblock = enaml_get_block(block, key);
			if (say2_server_command(server, key, subblock))
				continue;

			/* Dispatch of command failed */
			say2_server_message(server, NULL, SAY2_MSG_BAD_COMMAND,
					    "command", key,
					    "text", _("Unknown command"),
					    NULL);
		}

loop:
		/* Free up stuff */
		enaml_free(block);
		g_free(line);
	}
}
