/* meko
 * Copyright (C) 2001 Scott James Remnant <scott@netsplit.com>.
 * All Rights Reserved.
 *
 * say2-server.h
 * -- 
 * $Id: say2-server.h,v 1.14.4.1 2001/11/23 22:34:44 scott Exp $
 * 
 * This file is distributed according to the GNU General Public
 * License.  For full details, read the top of 'main.c' or the
 * file called COPYING that was distributed with this code.
 */

#ifndef _MEKO_SAY2_SERVER_H
#define _MEKO_SAY2_SERVER_H

#include <glib.h>
#include <gtk/gtkobject.h>

#include "say2/say2-client.h"
#include "say2/say2-user.h"
#include "say2/say2-channel.h"
#include "say2/say2-message.h"

/* Convenience defines */
#define SAY2_TYPE_SERVER \
	(say2_server_get_type())
#define SAY2_SERVER(obj) \
	(GTK_CHECK_CAST((obj), SAY2_TYPE_SERVER, Say2Server))
#define SAY2_SERVER_CLASS(klass) \
	(GTK_CHECK_CLASS_CAST((klass), SAY2_TYPE_SERVER, Say2ServerClass))
#define SAY2_IS_SERVER(obj) \
	(GTK_CHECK_TYPE((obj), SAY2_TYPE_SERVER))
#define SAY2_IS_SERVER_CLASS(klass) \
	(GTK_CHECK_CLASS_TYPE((klass), SAY2_TYPE_SERVER))

/* Defines related to flags */
#define SAY2_SERVER_SET_FLAGS(server,flag)	((server)->flags |= (flag))
#define SAY2_SERVER_UNSET_FLAGS(server,flag)	((server)->flags &= ~(flag))
#define SAY2_SERVER_CHECK_FLAG(server,flag)	((server)->flags & (flag))


/* Linear states the server structure can be in */
typedef enum {
	SAY2_SERVER_INACTIVE,
	SAY2_SERVER_CONNECTING,
	SAY2_SERVER_CONNECTED,
	SAY2_SERVER_VERIFIED,
	SAY2_SERVER_ACTIVE
} Say2ServerState;

/* Things that different protocol versions may or may not support */
typedef enum {
	SAY2_HAS_REPLICANTS	= 1 << 0
} Say2ServerFlags;


/* Typedefs */
typedef struct _say2_server	  Say2Server;
typedef struct _say2_server_class Say2ServerClass;

/* Instance data */
struct _say2_server {
	GtkObject object;

	/* Arguments */
	gchar *hostname;
	gint   port;
	gchar *password;
	guint  replicant : 1;
        guint  ssl : 1;

	/* Members */

	/* Flags */
	Say2ServerFlags flags;

	/* State of the connection */
	Say2ServerState state;

	/* Socket (actually of a type known only to libsay2) */
	void *sock;

	/* Pointer to user's own record, and list of users */
	Say2User   *me;
	GHashTable *users;

	/* Party line (single channel until we get real channel servers) */
	Say2Channel *party_line;

	/* List of squelches */
	GList *squelches;
};

/* Object class data */
struct _say2_server_class {
	GtkObjectClass parent_class;

	void (*connected)(Say2Server *);
	void (*verified)(Say2Server *);
	void (*authenticated)(Say2Server *);
	void (*disconnected)(Say2Server *);
	void (*signon)(Say2Server *, Say2User *);
	void (*signoff)(Say2Server *, Say2User *);
	void (*message)(Say2Server *, Say2Client *, Say2MessageType,
			GHashTable *);
};


/* Callback function for squelches */
typedef void (*SquelchFunc)(Say2Server *, Say2Client *, Say2MessageType,
			    GHashTable *, gpointer);


/* Required functions */
Say2Server *say2_server_new(void);
GtkType	    say2_server_get_type(void);

/* Get the embedded information */
Say2ServerState say2_server_get_state(Say2Server *);
Say2User *	say2_server_get_me(Say2Server *);

/* Find a user on the users table, creating it if necessary */
Say2User *say2_server_get_user(Say2Server *, const gchar *);

/* Remove a user from the users table */
void say2_server_remove_user(Say2Server *, Say2User *);

/* Find a user on the users table by its id or nickname */
Say2User *say2_server_get_user_by_name(Say2Server *, const gchar *);

/* Connect to and disconnect from a server */
void say2_server_connect(Say2Server *);
void say2_server_disconnect(Say2Server *);

/* Attempt to start up SSL on a server */
gboolean say2_server_begin_ssl(Say2Server *);

/* Login to a say2 server, returns FALSE if not enough info to do so */
gboolean say2_server_login(Say2Server *);

/* Change the state of the server and emit signals */
void say2_server_verified(Say2Server *);
void say2_server_authenticated(Say2Server *);

/* Send a message to the server */
void say2_server_send(Say2Server *, const gchar *tag, const gchar *fmt, ...);

/* Message dispatching */
void say2_server_message(Say2Server *, Say2Client *, Say2MessageType, ...);

/* Add a squelch */
void say2_server_add_squelch(Say2Server *, SquelchFunc, gpointer, gint, ...);

#endif /* !_MEKO_SAY2_SERVER_H */
