/* meko
 * Copyright (C) 2001 Scott James Remnant <scott@netsplit.com>.
 * All Rights Reserved.
 *
 * say2-user.c
 *  - Say2User class
 *  - An object to hold information about a user on the say2 network.
 * -- 
 * $Id: say2-user.c,v 1.15.4.1 2001/11/23 22:19:55 scott Exp $
 * 
 * This file is distributed according to the GNU General Public
 * License.  For full details, read the top of 'main.c' or the
 * file called COPYING that was distributed with this code.
 */

#include <meko.h>

#include <time.h>

#include <glib.h>
#include <gtk/gtkobject.h>

#include "say2/say2-client.h"

#include "say2-user.h"

/* Arguments
 * gboolean		guest		User doesn't have a registered id
 * gchar *		nick		Nickname for this session
 * Say2UserStatus	status		User's current status
 * gchar *		status_msg	Exactly what the user is doing
 * time_t		status_time	When the user's status last changed
 * gboolean		messages	User is receiving private messages
 */
enum {
	ARG_0,
	ARG_GUEST,
	ARG_NICK,
	ARG_STATUS,
	ARG_STATUS_MSG,
	ARG_STATUS_TIME,
	ARG_MESSAGES
};


/* Forward prototypes for required functions */
static void say2_user_init(Say2User *);
static void say2_user_class_init(Say2UserClass *);

/* Forward prototypes for GtkObject methods */
static void say2_user_set_arg(GtkObject *, GtkArg *, guint);
static void say2_user_get_arg(GtkObject *, GtkArg *, guint);
static void say2_user_destroy(GtkObject *);


/* Our parent class */
static Say2ClientClass *parent_class = NULL;


/* say2_user_get_type
 * Returns our GtkType id, allocating it if necessary
 */
guint
say2_user_get_type(void)
{
	static guint user_type = 0;

	if (!user_type) {
		static const GtkTypeInfo user_info = {
			"Say2User",
			sizeof(Say2User),
			sizeof(Say2UserClass),
			(GtkClassInitFunc)say2_user_class_init,
			(GtkObjectInitFunc)say2_user_init,
			/* reserved_1 */ NULL,
			/* reserved_2 */ NULL,
			(GtkClassInitFunc)NULL
		};

		user_type = gtk_type_unique(SAY2_TYPE_CLIENT, &user_info);
	}

	return user_type;
}

/* say2_user_class_init
 * Initialise the object class, setting our arguments and overriding methods
 * of parent object classes
 */
static void
say2_user_class_init(Say2UserClass *klass)
{
	GtkObjectClass *object_class;

	g_return_if_fail(klass != NULL);

	object_class = (GtkObjectClass *)klass;
	parent_class = gtk_type_class(SAY2_TYPE_CLIENT);

	gtk_object_add_arg_type("Say2User::guest", GTK_TYPE_BOOL,
				GTK_ARG_READWRITE, ARG_GUEST);
	gtk_object_add_arg_type("Say2User::nick", GTK_TYPE_STRING,
				GTK_ARG_READWRITE, ARG_NICK);
	gtk_object_add_arg_type("Say2User::status", GTK_TYPE_INT,
				GTK_ARG_READWRITE, ARG_STATUS);
	gtk_object_add_arg_type("Say2User::status_msg", GTK_TYPE_STRING,
				GTK_ARG_READWRITE, ARG_STATUS_MSG);
	gtk_object_add_arg_type("Say2User::status_time", GTK_TYPE_ULONG,
				GTK_ARG_READWRITE, ARG_STATUS_TIME);
	gtk_object_add_arg_type("Say2User::messages", GTK_TYPE_BOOL,
				GTK_ARG_READWRITE, ARG_MESSAGES);
	
	object_class->set_arg = say2_user_set_arg;
	object_class->get_arg = say2_user_get_arg;
	object_class->destroy = say2_user_destroy;
}

/* say2_user_init
 * Initialise an object instance, very little to do here other than set default
 * values
 */
static void
say2_user_init(Say2User *user)
{
	g_return_if_fail(user != NULL);

	user->guest = FALSE;
	user->nick = NULL;
	user->status = SAY2_USER_OFFLINE;
	user->status_msg = NULL;
	user->status_time = 0;
	user->messages = TRUE;
}

/* say2_user_new
 * Create a new object instance with the id given, just a wrapper around
 * gtk_type_new and gtk_object_set really.
 */
Say2User *
say2_user_new(const gchar *id)
{
	Say2User *user;

	user = SAY2_USER(gtk_type_new(SAY2_TYPE_USER));

	if (id)
		gtk_object_set(GTK_OBJECT(user), "id", id, NULL);

	return user;
}

/* say2_user_destroy
 * Destroy an object instance, freeing and unreferencing any data we've
 * been hanging on to
 */
static void
say2_user_destroy(GtkObject *object)
{
	Say2User *user;

	g_return_if_fail(object != NULL);
	g_return_if_fail(SAY2_IS_USER(object));

	user = SAY2_USER(object);

	g_free(user->nick);
	g_free(user->status_msg);

	/* Chain up */
	if (GTK_OBJECT_CLASS(parent_class)->destroy)
		GTK_OBJECT_CLASS(parent_class)->destroy(object);
}

/* say2_client_set_arg
 * Set object arguments
 */
static void
say2_user_set_arg(GtkObject *object, GtkArg *arg, guint arg_id)
{
	Say2User *user;

	g_return_if_fail(object != NULL);
	g_return_if_fail(SAY2_IS_USER(object));

	user = SAY2_USER(object);
	switch (arg_id) {
	case ARG_GUEST:
		user->guest = GTK_VALUE_BOOL(*arg);
		break;
	case ARG_NICK:
		g_free(user->nick);
		user->nick = g_strdup(GTK_VALUE_STRING(*arg));
		break;
	case ARG_STATUS:
		user->status = GTK_VALUE_INT(*arg);
		g_free(user->status_msg);
		user->status_msg = NULL;
		user->status_time = time(NULL);
		break;
	case ARG_STATUS_MSG:
		g_free(user->status_msg);
		user->status_msg = g_strdup(GTK_VALUE_STRING(*arg));
		user->status_time = time(NULL);
		break;
	case ARG_STATUS_TIME:
		user->status_time = GTK_VALUE_ULONG(*arg);
		break;
	case ARG_MESSAGES:
		user->messages = GTK_VALUE_BOOL(*arg);
		break;
	}

	say2_client_changed(SAY2_CLIENT(user));
}

/* say2_client_get_arg
 * Get object arguments
 */
static void
say2_user_get_arg(GtkObject *object, GtkArg *arg, guint arg_id)
{
	Say2User *user;

	g_return_if_fail(object != NULL);
	g_return_if_fail(SAY2_IS_USER(object));

	user = SAY2_USER(object);
	switch (arg_id) {
	case ARG_GUEST:
		GTK_VALUE_BOOL(*arg) = user->guest;
		break;
	case ARG_NICK:
		GTK_VALUE_STRING(*arg) = user->nick;
		break;
	case ARG_STATUS:
		GTK_VALUE_INT(*arg) = user->status;
		break;
	case ARG_STATUS_MSG:
		GTK_VALUE_STRING(*arg) = user->status_msg;
		break;
	case ARG_STATUS_TIME:
		GTK_VALUE_ULONG(*arg) = user->status_time;
		break;
	case ARG_MESSAGES:
		GTK_VALUE_BOOL(*arg) = user->messages;
		break;
	default:
		arg->type = GTK_TYPE_INVALID;
		break;
	}
}

/* say2_user_set_status
 * Convenient function to set all of the status attributes at once, converting
 * from a string version of the status to a Say2UserStatus
 */
void
say2_user_set_status(Say2User *user, const gchar *status,
		     const gchar *status_msg, gboolean messages,
		     time_t status_time)
{
	Say2UserStatus user_status;

	g_return_if_fail(user != NULL);

	user_status = say2_user_status_from_str(status);
	gtk_object_set(GTK_OBJECT(user),
		       "status", user_status, "status_msg", status_msg,
		       "status_time", (status_time ? status_time : time(NULL)),
		       "messages", messages,
		       NULL);
}

/* say2_user_get_status
 * Convenient function to get all of the status attributes at once, convering
 * the status into a string suitable for a say2 server
 */
const gchar *
say2_user_get_status(Say2User *user, gchar **status_msg, time_t *status_time)
{
	Say2UserStatus status;
	
	g_return_val_if_fail(user != NULL, SAY2_USER_UNKNOWN);

	gtk_object_get(GTK_OBJECT(user), "status", &status,
		       "status_msg", status_msg,
		       "status_time", status_time,
		       NULL);

	return say2_user_status_to_str(status);
}

/* say2_user_status_from_str
 * Returns a Say2UserStatus when given it's string representation
 */
Say2UserStatus
say2_user_status_from_str(const gchar *status)
{
	Say2UserStatus user_status;

	if (status == NULL) {
		user_status = SAY2_USER_UNKNOWN;
	} else if (!g_strcasecmp(status, "active")) {
		user_status = SAY2_USER_ACTIVE;
	} else if (!g_strcasecmp(status, "busy")) {
		user_status = SAY2_USER_BUSY;
	} else if (!g_strcasecmp(status, "away")) {
		user_status = SAY2_USER_AWAY;
	} else if (!g_strcasecmp(status, "idle")) {
		user_status = SAY2_USER_IDLE;
	} else if (!g_strcasecmp(status, "offline")) {
		user_status = SAY2_USER_OFFLINE;
	} else {
		user_status = SAY2_USER_UNKNOWN;
	}

	return user_status;
}

/* say2_user_status_to_str
 * Returns a say2 status string for a given user status
 */
const gchar *
say2_user_status_to_str(Say2UserStatus user_status)
{
	const gchar *status;
		
	switch (user_status) {
	case SAY2_USER_ACTIVE:
		status = "active";
		break;
	case SAY2_USER_BUSY:
		status = "busy";
		break;
	case SAY2_USER_AWAY:
		status = "away";
		break;
	case SAY2_USER_IDLE:
		status = "idle";
		break;
	case SAY2_USER_OFFLINE:
		status = "offline";
		break;
	default:
		status = "unknown";
		break;
	}

	return status;
}
