/* meko
 * Copyright (C) 2001 Scott James Remnant <scott@netsplit.com>.
 * All Rights Reserved.
 *
 * message-text.c
 *  - Takes the parameters of a message signal from a Say2Server and returns
 *    a string that can be output anywhere
 * -- 
 * $Id: message-text.c,v 1.3.4.3 2001/11/24 01:39:45 scott Exp $
 * 
 * This file is distributed according to the GNU General Public
 * License.  For full details, read the top of 'main.c' or the
 * file called COPYING that was distributed with this code.
 */

#include <meko.h>

#include <string.h>
#include <time.h>

#include <glib.h>
#include <gtk/gtkobject.h>
#include <libgnome/gnome-defs.h>
#include <libgnome/gnome-i18n.h>

#include "say2/say2-server.h"
#include "say2/say2-client.h"
#include "say2/say2-user.h"
#include "say2/say2-message.h"
#include "util/say2/client-label.h"

#include "message-text.h"


/* meko_message_text
 * Convert a message signal from a server into a string that can be outputted.
 * The result is NULL if the string can't be created, otherwise it's a string
 * that must be freed after use.
 */
gchar *
meko_message_text(Say2Server *server, Say2Client *client,
		  Say2MessageType message_type, GHashTable *hash)
{
	gchar *msg;

	g_return_val_if_fail(hash != NULL, NULL);
	g_return_val_if_fail(server != NULL, NULL);
	g_return_val_if_fail(message_type >= SAY2_MSG_ANY, NULL);
	g_return_val_if_fail(message_type < SAY2_MSG_LAST, NULL);

	/* We want the message, but its a bit encoded right now (FIXME) so
	 * decode it into something useful
	 */
	if (message_type == SAY2_MSG_CONNECTING) {
		const gchar *hostname;
		gint	     port;

		gtk_object_get(GTK_OBJECT(server), "hostname", &hostname,
			       "port", &port, NULL);
		msg = g_strdup_printf("*** Connecting to %s port %d\n",
				      hostname, port);

	} else if (message_type == SAY2_MSG_CONNECTION_FAILED) {
		msg = g_strdup_printf("*** Connection failed! :(\n");

	} else if (message_type == SAY2_MSG_CONNECTED) {
		msg = g_strdup_printf("*** Connected!\n");

	} else if (message_type == SAY2_MSG_NEED_LOGIN) {
		msg = g_strdup_printf("*** Please /login or /guest\n");

	} else if (message_type == SAY2_MSG_DISCONNECTED) {
		msg = g_strdup_printf("*** Disconnected\n");

        } else if (message_type == SAY2_MSG_SSL_SUCCESS) {
                msg = g_strdup_printf("*** SSL secure link established!\n");

        } else if (message_type == SAY2_MSG_SSL_FAILED) {
                msg = g_strdup_printf("*** SSL attempt failed. :(\n");

	} else if (message_type == SAY2_MSG_ERROR) {
		const gchar *command, *text;

		command = g_hash_table_lookup(hash, "command");
		text = g_hash_table_lookup(hash, "text");

		if (command) {
			msg = g_strdup_printf("::: %s Error: %s\n",
					      command, text);
		} else {
			msg = g_strdup_printf("::: Error: %s\n", text);
		}
	
	} else if (message_type == SAY2_MSG_BAD_ENAML) {
		const gchar *enaml;

		enaml = g_hash_table_lookup(hash, "enaml");

		msg = g_strdup_printf("::: Bad Enaml: %s\n", enaml);

	} else if (message_type == SAY2_MSG_BAD_COMMAND) {
		const gchar *command, *text;

		command = g_hash_table_lookup(hash, "command");
		text = g_hash_table_lookup(hash, "text");

		msg = g_strdup_printf("*** Bad Command Format: %s: %s\n",
				      command, text);

	} else if (message_type == SAY2_MSG_BAD_NICKNAME) {
		const gchar *nick, *text;

		nick = g_hash_table_lookup(hash, "nick");
		text = g_hash_table_lookup(hash, "text");

		msg = g_strdup_printf("::: Bad Nickname: %s: %s\n", nick, text);

	} else if (message_type == SAY2_MSG_BAD_ADDRESS) {
		const gchar *address;

		address = g_hash_table_lookup(hash, "address");

		msg = g_strdup_printf("::: Bad Address: %s\n", address);

	} else if (message_type == SAY2_MSG_SERVER_INFO) {
		const gchar *software, *version, *text;
		time_t	     uptime;
		GString     *buff;

		software = g_hash_table_lookup(hash, "software");
		version = g_hash_table_lookup(hash, "version");
		text = g_hash_table_lookup(hash, "text");
		uptime = (time_t)g_hash_table_lookup(hash, "uptime");

		buff = g_string_new("");
		g_string_sprintfa(buff, "::: %s/%s up since %ld\n",
				  software, version, uptime);
		if (text)
			g_string_sprintfa(buff, "::: %s\n", text);
		msg = buff->str;
		g_string_free(buff, FALSE);

	} else if (message_type == SAY2_MSG_MOTD) {
		const gchar *text;

		text = g_hash_table_lookup(hash, "text");

		msg = g_strdup_printf("::: [MOTD] %s\n", text);

	} else if (message_type == SAY2_MSG_WELCOME) {
		const gchar *id, *label;
		Say2User    *me;
		gboolean     attached;
		GString     *buff;

		attached = (gboolean)g_hash_table_lookup(hash, "attached");

		me = say2_server_get_me(server);
		gtk_object_get(GTK_OBJECT(me), "id", &id, NULL);
		label = meko_say2_client_label(SAY2_CLIENT(me));

		buff = g_string_new("");
		g_string_sprintfa(buff, "::: Welcome, %s!\n", label);
		g_string_sprintfa(buff, "::: Your ID here is %s\n", id);
		if (attached) {
			g_string_sprintfa(buff, "::: Attached to previous "
					  "session\n");
		}
		msg = buff->str;
		g_string_free(buff, FALSE);

	} else if (message_type == SAY2_MSG_LOGOUT) {
		const gchar *detached, *text;
		GString     *buff;

		detached = g_hash_table_lookup(hash, "detached");
		text = g_hash_table_lookup(hash, "text");

		buff = g_string_new("");
		if (detached) {
			g_string_sprintfa(buff, "::: Detached by host "
					        "from %s\n", detached);
		}
		if (text) {
			g_string_sprintfa(buff, "::: %s\n", text);
		} else {
			g_string_sprintfa(buff, "::: Goodbye!\n");
		}
		msg = buff->str;
		g_string_free(buff, FALSE);

	} else if (message_type == SAY2_MSG_REPLICANT_NEW) {
		const gchar *host;

		host = g_hash_table_lookup(hash, "host");

		msg = g_strdup_printf("::: New replicant from %s\n", host);

	} else if (message_type == SAY2_MSG_REPLICANT_GONE) {
		const gchar *host;

		host = g_hash_table_lookup(hash, "host");

		msg = g_strdup_printf("::: Replicant from %s retired\n", host);

	} else if (message_type == SAY2_MSG_ATTRIBUTES_SET) {
		const gchar *list;

		list = g_hash_table_lookup(hash, "list");

		if (list) {
			msg = g_strdup_printf("::: Attributes set: %s\n", list);
		} else {
			msg = g_strdup_printf("::: Attributes set\n");
		}

	} else if (message_type == SAY2_MSG_PREFS) {
		const gchar *tagline, *notify;
		gboolean     icon, password;
		GString     *buff;

		tagline = g_hash_table_lookup(hash, "tagline");
		notify = g_hash_table_lookup(hash, "notify");
		icon = (gboolean)g_hash_table_lookup(hash, "icon");
		password = (gboolean)g_hash_table_lookup(hash, "password");

		buff = g_string_new("");
		g_string_sprintfa(buff, "::: Preferences set:\n");
		if (tagline)
			g_string_sprintfa(buff, ":::   Tagline: %s\n", tagline);
		if (notify)
			g_string_sprintfa(buff, ":::   Notify: %s\n", notify);
		if (icon)
			g_string_sprintfa(buff, "::: Icon changed\n");
		if (password)
			g_string_sprintfa(buff, "::: Password changed\n");
		msg = buff->str;
		g_string_free(buff, FALSE);

	} else if (message_type == SAY2_MSG_NICK_NOT_PRESENT) {
		const gchar *nick;

		nick = g_hash_table_lookup(hash, "nick");

		msg = g_strdup_printf("::: Nickname not present: %s\n", nick);

	} else if (message_type == SAY2_MSG_CLIENT_INVALID) {
		const gchar *type, *id;

		type = g_hash_table_lookup(hash, "type");
		id = g_hash_table_lookup(hash, "id");

		msg = g_strdup_printf("::: Invalid %s: %s\n", type, id);

	} else if (message_type == SAY2_MSG_MEMBERS) {
		const gchar *type, *list;

		type = g_hash_table_lookup(hash, "type");
		list = g_hash_table_lookup(hash, "list");

		msg = g_strdup_printf("::: %s list: %s\n", type, list);

	} else if (message_type == SAY2_MSG_WALL) {
		const gchar *label, *text;
		gchar	     time_str[12];
		time_t	     text_time;
		Say2User    *user;

		g_return_val_if_fail(SAY2_IS_USER(client), NULL);

		text = g_hash_table_lookup(hash, "text");
		text_time = (time_t)g_hash_table_lookup(hash, "time");
		user = g_hash_table_lookup(hash, "user");
		g_return_val_if_fail(SAY2_IS_USER(user), NULL);

		strftime(time_str, sizeof(time_str), " [%H:%M:%S] ",
			 localtime(&text_time));

		label = meko_say2_client_label(SAY2_CLIENT(user));
		msg = g_strdup_printf(":::%s WALL from %s: %s\n",
				      time_str, label, text);

	} else if (message_type == SAY2_MSG_USER_INVALID) {
		const gchar *label;

		g_return_val_if_fail(SAY2_IS_USER(client), NULL);

		label = meko_say2_client_label(client);
		msg = g_strdup_printf("::: Invalid user: %s\n", label);
	
	} else if (message_type == SAY2_MSG_USER_INFO) {
		/* FIXME, whois output here */
		return NULL;

	} else if (message_type == SAY2_MSG_USER_ATTRIBUTES) {
		/* FIXME, attributes here */
		return NULL;

	} else if (message_type == SAY2_MSG_USER_STATUS) {
		const gchar *label, *mode, *text;
		gchar	     time_str[12];
		time_t	     status_time;
		gboolean     messages;

		g_return_val_if_fail(SAY2_IS_USER(client), NULL);

		mode = g_hash_table_lookup(hash, "mode");
		text = g_hash_table_lookup(hash, "text");
		status_time = (time_t)g_hash_table_lookup(hash, "time");
		messages = (gboolean)g_hash_table_lookup(hash, "messages");

		strftime(time_str, sizeof(time_str), " [%H:%M:%S]",
			 localtime(&status_time));

		label = meko_say2_client_label(client);
		if (text) {
			msg = g_strdup_printf(":::%s %s is now %s%s: %s\n",
					      time_str, label, mode,
					      (messages
					       ? "" : " (messages stored)"),
					      text);
		} else if (messages) {
			msg = g_strdup_printf(":::%s %s is now %s\n",
					      time_str, label, mode);
		} else {
			msg = g_strdup_printf(":::%s %s is now %s "
					      "(messages stored)\n",
					      time_str, label, mode);
		}

	} else if (message_type == SAY2_MSG_USER_NICK) {
		const gchar *label, *old_nick, *new_nick;

		g_return_val_if_fail(SAY2_IS_USER(client), NULL);

		old_nick = g_hash_table_lookup(hash, "old-nick");

		label = meko_say2_client_label(client);
		gtk_object_get(GTK_OBJECT(client), "nick", &new_nick, NULL);
		if (strcmp(label, new_nick)) {
			msg = g_strdup_printf("::: %s is now known as %s "
					      "(will displayed as %s)\n",
					      old_nick, new_nick, label);
		} else {
			msg = g_strdup_printf("::: %s is now known as %s\n",
					      old_nick, new_nick);
		}

	} else if (message_type == SAY2_MSG_USER_TEXT) {
		const gchar *label, *text;
		gchar	     time_str[12];
		time_t	     text_time;

		g_return_val_if_fail(SAY2_IS_USER(client), NULL);

		text = g_hash_table_lookup(hash, "text");
		text_time = (time_t)g_hash_table_lookup(hash, "time");

		strftime(time_str, sizeof(time_str), "[%H:%M:%S] ",
			 localtime(&text_time));

		label = meko_say2_client_label(client);
		msg = g_strdup_printf("%s[%s] %s\n", time_str, label, text);

	} else if (message_type == SAY2_MSG_USER_ACT) {
		const gchar *label, *text;
		gchar	     time_str[12];
		time_t	     act_time;

		g_return_val_if_fail(SAY2_IS_USER(client), NULL);

		text = g_hash_table_lookup(hash, "text");
		act_time = (time_t)g_hash_table_lookup(hash, "time");

		strftime(time_str, sizeof(time_str), "[%H:%M:%S] ",
			 localtime(&act_time));

		label = meko_say2_client_label(client);
		msg = g_strdup_printf("%s[* %s] %s\n", time_str, label, text);

	} else if (message_type == SAY2_MSG_USER_TEXT_TO) {
		const gchar *label, *text;
		gchar	     time_str[12];
		gboolean     stored;
		time_t	     text_time;
		Say2User    *user, *me;

		g_return_val_if_fail(SAY2_IS_USER(client), NULL);

		text = g_hash_table_lookup(hash, "text");
		text_time = (time_t)g_hash_table_lookup(hash, "time");
		user = g_hash_table_lookup(hash, "user");
		g_return_val_if_fail(SAY2_IS_USER(user), NULL);
		stored = (gboolean)g_hash_table_lookup(hash, "stored");

		strftime(time_str, sizeof(time_str), "[%H:%M:%S] ",
			 localtime(&text_time));

		me = say2_server_get_me(server);
		label = meko_say2_client_label(SAY2_CLIENT(me));
		msg = g_strdup_printf("%s[%s%s] %s\n", time_str, label,
				      (stored ? " (stored)" : ""), text);

	} else if (message_type == SAY2_MSG_USER_ACT_TO) {
		const gchar *label, *text;
		gchar	     time_str[12];
		gboolean     stored;
		time_t	     act_time;
		Say2User    *user, *me;

		g_return_val_if_fail(SAY2_IS_USER(client), NULL);

		text = g_hash_table_lookup(hash, "text");
		act_time = (time_t)g_hash_table_lookup(hash, "time");
		user = g_hash_table_lookup(hash, "user");
		g_return_val_if_fail(SAY2_IS_USER(user), NULL);
		stored = (gboolean)g_hash_table_lookup(hash, "stored");

		strftime(time_str, sizeof(time_str), "[%H:%M:%S] ",
			 localtime(&act_time));

		me = say2_server_get_me(server);
		label = meko_say2_client_label(SAY2_CLIENT(me));
		msg = g_strdup_printf("%s[* %s%s] %s\n", time_str, label,
				      (stored ? " (stored)" : ""), text);

	} else if (message_type == SAY2_MSG_CHAN_INVALID) {
		const gchar *label;

		g_return_val_if_fail(SAY2_IS_CHANNEL(client), NULL);

		label = meko_say2_client_label(client);
		msg = g_strdup_printf("::: Invalid channel: %s\n", label);

	} else if (message_type == SAY2_MSG_CHAN_MEMBERS) {
		const gchar *label, *list;

		g_return_val_if_fail(SAY2_IS_CHANNEL(client), NULL);

		list = g_hash_table_lookup(hash, "list");

		label = meko_say2_client_label(client);
		msg = g_strdup_printf("--- %s members: %s\n", label, list);

	} else if (message_type == SAY2_MSG_CHAN_SIGNON) {
		const gchar *user_label, *chan_label, *email;
		gchar        time_str[12];
		time_t       signon_time;
		Say2User    *user;

		g_return_val_if_fail(SAY2_IS_CHANNEL(client), NULL);

		signon_time = (time_t)g_hash_table_lookup(hash, "time");
		user = g_hash_table_lookup(hash, "user");
		g_return_val_if_fail(SAY2_IS_USER(user), NULL);

		strftime(time_str, sizeof(time_str), " [%H:%M:%S]",
			 localtime(&signon_time));

		gtk_object_get(GTK_OBJECT(user), "email", &email, NULL);

		user_label = meko_say2_client_label(SAY2_CLIENT(user));
		chan_label = meko_say2_client_label(client);
		msg = g_strdup_printf("-->%s Signon: %s <%s> joined %s\n",
				      time_str, user_label, email,
				      chan_label);

	} else if (message_type == SAY2_MSG_CHAN_SIGNOFF) {
		const gchar *user_label, *chan_label, *email;
		gchar        time_str[12];
		time_t       signoff_time;
		Say2User    *user;

		g_return_val_if_fail(SAY2_IS_CHANNEL(client), NULL);

		signoff_time = (time_t)g_hash_table_lookup(hash, "time");
		user = g_hash_table_lookup(hash, "user");
		g_return_val_if_fail(SAY2_IS_USER(user), NULL);

		strftime(time_str, sizeof(time_str), " [%H:%M:%S]",
			 localtime(&signoff_time));

		gtk_object_get(GTK_OBJECT(user), "email", &email, NULL);

		user_label = meko_say2_client_label(SAY2_CLIENT(user));
		chan_label = meko_say2_client_label(client);
		msg = g_strdup_printf("<--%s Signoff: %s <%s> left %s\n",
				      time_str, user_label, email,
				      chan_label);

	} else if (message_type == SAY2_MSG_CHAN_TEXT) {
		const gchar *label, *text;
		gchar	     time_str[12];
		time_t	     text_time;
		Say2User    *user;

		g_return_val_if_fail(SAY2_IS_CHANNEL(client), NULL);

		text = g_hash_table_lookup(hash, "text");
		text_time = (time_t)g_hash_table_lookup(hash, "time");
		user = g_hash_table_lookup(hash, "user");
		g_return_val_if_fail(SAY2_IS_USER(user), NULL);

		strftime(time_str, sizeof(time_str), "[%H:%M:%S] ",
			 localtime(&text_time));

		label = meko_say2_client_label(SAY2_CLIENT(user));
		msg = g_strdup_printf("%s<%s> %s\n", time_str, label, text);

	} else if (message_type == SAY2_MSG_CHAN_ACT) {
		const gchar *label, *text;
		gchar	     time_str[12];
		time_t	     act_time;
		Say2User    *user;

		g_return_val_if_fail(SAY2_IS_CHANNEL(client), NULL);

		text = g_hash_table_lookup(hash, "text");
		act_time = (time_t)g_hash_table_lookup(hash, "time");
		user = g_hash_table_lookup(hash, "user");
		g_return_val_if_fail(SAY2_IS_USER(user), NULL);

		strftime(time_str, sizeof(time_str), "[%H:%M:%S] ",
			 localtime(&act_time));

		label = meko_say2_client_label(SAY2_CLIENT(user));
		msg = g_strdup_printf("%s* %s %s\n", time_str, label, text);

	} else {
		return NULL;
	}

	return msg;
}
