/* meko
 * Copyright (C) 2001 Scott James Remnant <scott@netsplit.com>.
 * All Rights Reserved.
 *
 * user-commands.c
 *  - Commands users can run in the traditional IRC / style
 * -- 
 * $Id: user-commands.c,v 1.1 2001/10/02 22:54:18 scott Exp $
 * 
 * This file is distributed according to the GNU General Public
 * License.  For full details, read the top of 'main.c' or the
 * file called COPYING that was distributed with this code.
 */

#include <meko.h>

#include <string.h>

#include <glib.h>
#include <gtk/gtkobject.h>
#include <libgnome/gnome-defs.h>
#include <libgnome/gnome-i18n.h>
#include <say2/dispatch.h>

#include "say2/say2-server.h"
#include "say2/say2-client.h"
#include "say2/say2-channel.h"

#include "user-commands.h"


/* Forward prototypes for command functions */
static void _say(gpointer [], const gchar *, gchar **);
static void _me(gpointer [], const gchar *, gchar **);
static void _msg(gpointer [], const gchar *, gchar **);
static void _action(gpointer [], const gchar *, gchar **);
static void _wall(gpointer [], const gchar *, gchar **);

/* Forward prototypes for internal functions */


/* Command dispatch table for when server is active */
static DispatchItem active_commands[] = {
	{ "say",			_say },
	{ "me",				_me },
	{ "msg",			_msg },
	{ "m",				_msg },
	{ "action",			_action },
	{ "act",			_action },
	{ "wall",			_wall },
	{ NULL,				NULL }
};

/* Command dispatch table for when server isn't active */
static DispatchItem inactive_commands[] = {
	{ NULL,				NULL }
};

/* Command prefix */
static const gchar command_prefix[] = "/";


/* meko_user_command
 * Take text typed by the user, and send it to the say2 server in the format
 * of a valid command.  A string may be returned which should be output and
 * freed.
 */
gchar *
meko_user_command(Say2Server *server, Say2Client *client, const gchar *text)
{
	static void *active_dispatch = NULL;
	static void *inactive_dispatch = NULL;
	void        *dispatch;

	g_return_val_if_fail(server != NULL, NULL);
	g_return_val_if_fail(text != NULL, NULL);

	if (!strlen(text))
		return NULL;

	/* Certain commands can only be used for an active server */
	if (say2_server_get_state(server) == SAY2_SERVER_ACTIVE) {
		if (active_dispatch == NULL)
			active_dispatch = dispatch_create(active_commands);
		dispatch = active_dispatch;
	} else {
		if (inactive_dispatch == NULL)
			inactive_dispatch = dispatch_create(inactive_commands);
		dispatch = inactive_dispatch;
	}

	/* If it begins with a command prefix, treat it as a command, otherwise
	 * assume it's text to send to our client
	 */
	if (!strncmp(text, command_prefix, strlen(command_prefix))) {
		gchar  *cmd;
		size_t  cmdlen;
		
		/* Skip the command prefix and copy the command */
		text += strlen(command_prefix);
		cmdlen = strcspn(text, " ");
		cmd = g_strndup(text, cmdlen);

		/* Skip the command and space */
		text += cmdlen;
		if (*text)
			text++;

		/* Only dispatch if there was a command, otherwise drop out
		 * and carry on as if there wasn't a command
		 */
		if (cmdlen > 0) {
			gpointer  params[2];
			gchar    *ret_msg;

			ret_msg = NULL;

			params[0] = server;
			params[1] = client;
			
			if (!dispatch_go(dispatch, cmd, params,
					 (gpointer)text, &ret_msg)) {
				return g_strdup_printf("%s: %s",
						       _("Unknown Command"),
						       cmd);
			}

			g_free(cmd);
			return ret_msg;
		} else {
			g_free(cmd);
		}
	}
	
	/* Assume it wasn't a command, so send as a say/message */
	if ((client != NULL)
	    && (say2_server_get_state(server) == SAY2_SERVER_ACTIVE)) {
		meko_user_message(server, client, "text", text);
		return NULL;

	} else {
		return g_strdup_printf(_("Can't send text"));
	}
}

/* meko_user_message
 * Send a message as a say, message or wall (depding on the value of client)
 * to the server, with the specified message type
 */
void
meko_user_message(Say2Server *server, Say2Client *client, const gchar *type,
		  const gchar *text)
{
	g_return_if_fail(server != NULL);
	g_return_if_fail(type != NULL);
	g_return_if_fail(text != NULL);

	if (client == NULL) {
		/* Send message as a "wall" (ie to everyone) */
		say2_server_send(server, "wall", "text:%s", text);

	} else if (SAY2_IS_CHANNEL(client)) {
		/* Send message as a "say" (ie to a channel) */
		say2_server_send(server, "say", "%s:%s", type, text);

	} else {
		gchar *id;

		gtk_object_get(GTK_OBJECT(client), "id", &id, NULL);

		/* Send message as a "message" (ie to a specific someone) */
		say2_server_send(server, "message", "id:%s %s:%s", id,
				 type, text);
	}
}

/* _say
 * Send a message to the client as if it was typed
 */
static void
_say(gpointer params[], const gchar *text, gchar **ret_msg)
{
	meko_user_message((Say2Server *)params[0], (Say2Client *)params[1],
			  "text", text);
}

static void
_me(gpointer params[], const gchar *text, gchar **ret_msg)
{
	meko_user_message((Say2Server *)params[0], (Say2Client *)params[1],
			  "act", text);
}

static void
_msg(gpointer params[], const gchar *text, gchar **ret_msg)
{
	Say2User *user;
	gchar     **arg;

	arg = g_strsplit(text, " ", 1);
	user = say2_server_get_user_by_name((Say2Server *)params[0], arg[0]);
	meko_user_message((Say2Server *)params[0], SAY2_CLIENT(user),
			  "text", arg[1]);

	g_strfreev(arg);
}

static void
_action(gpointer params[], const gchar *text, gchar **ret_msg)
{
	Say2User *user;
	gchar     **arg;

	arg = g_strsplit(text, " ", 1);
	user = say2_server_get_user_by_name((Say2Server *)params[0], arg[0]);
	meko_user_message((Say2Server *)params[0], SAY2_CLIENT(user),
			  "act", arg[1]);

	g_strfreev(arg);
}

static void
_wall(gpointer params[], const gchar *text, gchar **ret_msg)
{
	meko_user_message((Say2Server *)params[0], NULL, "text", text);
}
