/* meko
 * Copyright (C) 2001 Scott James Remnant <scott@netsplit.com>.
 * All Rights Reserved.
 *
 * meko-chat-interface.c
 *  - MekoChatInterface class
 *  - A widget, derived from GtkVBox that allows the user to talk to people
 *    and issue say2 commands.  (Yes, this is meat!)
 * -- 
 * $Id: meko-chat-interface.c,v 1.19.4.1 2001/11/23 22:19:55 scott Exp $
 * 
 * This file is distributed according to the GNU General Public
 * License.  For full details, read the top of 'main.c' or the
 * file called COPYING that was distributed with this code.
 */

#include <meko.h>

#include <string.h>
#include <stdarg.h>

#include <glib.h>
#include <gdk/gdk.h>
#include <gtk/gtkobject.h>
#include <gtk/gtksignal.h>
#include <gtk/gtkwidget.h>
#include <gtk/gtkcontainer.h>
#include <gtk/gtkframe.h>
#include <gtk/gtkpaned.h>
#include <gtk/gtkhpaned.h>
#include <gtk/gtkbox.h>
#include <gtk/gtkhbox.h>
#include <gtk/gtkvbox.h>
#include <gtk/gtktext.h>
#include <gtk/gtkentry.h>
#include <gtk/gtklist.h>
#include <gtk/gtkvscrollbar.h>
#include <gtk/gtkscrolledwindow.h>
#include <libgnome/gnome-defs.h>
#include <libgnome/gnome-i18n.h>
#include <libgnomeui/gnome-dialog.h>

#include "say2/say2-server.h"
#include "say2/say2-client.h"
#include "say2/say2-user.h"
#include "say2/say2-channel.h"
#include "util/gui/image-widget.h"
#include "util/say2/client-label.h"
#include "util/ui/message-text.h"
#include "util/ui/user-commands.h"
#include "widgets/client-info/meko-client-info.h"
#include "widgets/client-list/meko-client-list.h"
#include "widgets/status-menu/meko-status-menu.h"
#include "widgets/status-menu/meko-status-menu-dialog.h"

#include "meko-chat-interface.h"

/* Arguments
 * Say2Server	*server		registration server to send commands to
 * Say2Client	*client		client (user or channel) to send messages to
 */
enum {
	ARG_0,
	ARG_SERVER,
	ARG_CLIENT
};

/* Signals */
enum {
	REMODEL,
	LAST_SIGNAL
};

/* Fonts */
#define TEXT_FONT "-*-lucidatypewriter-medium-r-normal-*-12-*-*-*-m-*-iso8859-1"


/* Forward prototypes for required functions */
static void meko_chat_interface_init(MekoChatInterface *);
static void meko_chat_interface_class_init(MekoChatInterfaceClass *);

/* Forward prototypes for GtkObject methods */
static void meko_chat_interface_set_arg(GtkObject *, GtkArg *, guint);
static void meko_chat_interface_get_arg(GtkObject *, GtkArg *, guint);
static void meko_chat_interface_destroy(GtkObject *);

/* Forward prototypes for MekoChatInterface methods */
static void meko_real_chat_interface_remodel(MekoChatInterface *);

/* Forward prototypes for MekoChatInterface method wrappers */
static void meko_chat_interface_remodel(MekoChatInterface *);

/* Forward prototypes for internal functions */
static void _change_font(GtkWidget *);
static void _status_changed_cb(MekoStatusMenu *, MekoChatInterface *);
static void _status_dialog_clicked_cb(MekoStatusMenuDialog *, gint,
				      const gchar *);
static void _entry_activate_cb(GtkEntry *, MekoChatInterface *);
static void _channel_signon_cb(Say2Channel *, Say2User *, MekoChatInterface *);
static void _channel_signoff_cb(Say2Channel *, Say2User *, MekoChatInterface *);
static void _server_disconnected_cb(Say2Server *, MekoChatInterface *);
static void _server_message_cb(Say2Server *, Say2Client *, Say2MessageType,
			       GHashTable *, MekoChatInterface *);

/* Our parent class */
static GtkVBoxClass *parent_class = NULL;

/* Signal array */
static guint chat_interface_signals[LAST_SIGNAL] = { 0 };


/* meko_chat_interface_get_type
 * Returns our GtkType id, allocating it if necessary.
 */
guint
meko_chat_interface_get_type(void)
{
	static guint chat_interface_type = 0;

	if (!chat_interface_type) {
		static const GtkTypeInfo chat_interface_info = {
			"MekoChatInterface",
			sizeof(MekoChatInterface),
			sizeof(MekoChatInterfaceClass),
			(GtkClassInitFunc)meko_chat_interface_class_init,
			(GtkObjectInitFunc)meko_chat_interface_init,
			/* reserved_1 */ NULL,
			/* reserved_2 */ NULL,
			(GtkClassInitFunc)NULL
		};

		chat_interface_type = gtk_type_unique(GTK_TYPE_VBOX,
						      &chat_interface_info);
	}

	return chat_interface_type;
}

/* meko_chat_interface_class_init
 * Initialise the object class, setting our arguments and overriding methods
 * of parent object classes
 */
static void
meko_chat_interface_class_init(MekoChatInterfaceClass *klass)
{
	GtkObjectClass *object_class;

	g_return_if_fail(klass != NULL);

	object_class = (GtkObjectClass *)klass;
	parent_class = gtk_type_class(GTK_TYPE_BUTTON);

	gtk_object_add_arg_type("MekoChatInterface::server", GTK_TYPE_POINTER,
				GTK_ARG_READWRITE, ARG_SERVER);
	gtk_object_add_arg_type("MekoChatInterface::client", GTK_TYPE_POINTER,
				GTK_ARG_READWRITE, ARG_CLIENT);

	chat_interface_signals[REMODEL] = gtk_signal_new(
		"remodel", GTK_RUN_FIRST, object_class->type,
		GTK_SIGNAL_OFFSET(MekoChatInterfaceClass, remodel),
		gtk_marshal_NONE__NONE,
		GTK_TYPE_NONE, 0
	);

	gtk_object_class_add_signals(object_class, chat_interface_signals,
				     LAST_SIGNAL);

	object_class->set_arg = meko_chat_interface_set_arg;
	object_class->get_arg = meko_chat_interface_get_arg;
	object_class->destroy = meko_chat_interface_destroy;

	klass->remodel = meko_real_chat_interface_remodel;
}

/* _change_font
 * Change the font of a widget to a monospaced one
 */
static void
_change_font(GtkWidget *widget)
{
	GtkStyle *style;
	
	g_return_if_fail(widget != NULL);

	gtk_widget_set_rc_style(widget);
	style = gtk_style_copy(gtk_widget_get_style(widget));

	style->font = gdk_font_load(TEXT_FONT);
	if (style->font == NULL)
		style->font = gdk_font_load("fixed");
	if (style->font == NULL)
		style->font = style->font;
	
	gtk_widget_set_style(widget, style);
	gtk_style_unref(style);
}

/* meko_chat_interface_init
 * Initialise an object instance, setting our default values, build our static
 * contents
 */
static void
meko_chat_interface_init(MekoChatInterface *chat_interface)
{
	GtkWidget *scroll;

	g_return_if_fail(chat_interface != NULL);

	gtk_box_set_spacing(GTK_BOX(chat_interface), 2);

	chat_interface->server = NULL;
	chat_interface->client = NULL;
	memset(chat_interface->recv_msg, sizeof(chat_interface->recv_msg), 0);
	chat_interface->server_disconnected_sigh = 0;
	chat_interface->server_message_sigh = 0;

	/* Construct the contents */

	/* These will be added as needed by remodel */
	chat_interface->info_box = NULL;
	chat_interface->info = NULL;
	chat_interface->pane = NULL;
	chat_interface->list = NULL;

	/* Horizontal box for the status menu and entry box */
	chat_interface->btm_box = gtk_hbox_new(FALSE, 2);
	gtk_box_pack_end(GTK_BOX(chat_interface), chat_interface->btm_box,
			 FALSE, FALSE, 0);
	gtk_widget_show(chat_interface->btm_box);

	/* Status menu */
	chat_interface->status = meko_status_menu_new(NULL);
	meko_status_menu_set_value(MEKO_STATUS_MENU(chat_interface->status),
				   SAY2_USER_OFFLINE);
	gtk_box_pack_start(GTK_BOX(chat_interface->btm_box),
			   chat_interface->status, FALSE, FALSE, 0);
	gtk_signal_connect(GTK_OBJECT(chat_interface->status), "changed",
			   GTK_SIGNAL_FUNC(_status_changed_cb), chat_interface);
	gtk_widget_show(chat_interface->status);

	/* Entry box */
	chat_interface->entry = gtk_entry_new();
	_change_font(chat_interface->entry);
	gtk_box_pack_start(GTK_BOX(chat_interface->btm_box),
			   chat_interface->entry, TRUE, TRUE, 0);
	gtk_signal_connect(GTK_OBJECT(chat_interface->entry), "activate",
			   GTK_SIGNAL_FUNC(_entry_activate_cb), chat_interface);
	gtk_widget_show(chat_interface->entry);

	/* Horizontal box for the text box and it's scroll bar. */
	chat_interface->text_box = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(chat_interface), chat_interface->text_box,
			   TRUE, TRUE, 0);
	gtk_widget_show(chat_interface->text_box);

	/* Text box */
	chat_interface->text = gtk_text_new(NULL, NULL);
	_change_font(chat_interface->text);
	gtk_text_set_editable(GTK_TEXT(chat_interface->text), FALSE);
	gtk_text_set_line_wrap(GTK_TEXT(chat_interface->text), TRUE);
	gtk_text_set_word_wrap(GTK_TEXT(chat_interface->text), TRUE);
	gtk_box_pack_start(GTK_BOX(chat_interface->text_box),
			   chat_interface->text, TRUE, TRUE, 0);
	gtk_widget_show(chat_interface->text);

	/* Scrollbar for the text box */
	scroll = gtk_vscrollbar_new(GTK_TEXT(chat_interface->text)->vadj);
	gtk_box_pack_start(GTK_BOX(chat_interface->text_box), scroll,
			   FALSE, FALSE, 0);
	gtk_widget_show(scroll);
}

/* meko_chat_interface_new
 * Create a new object instance, just a wrapper around gtk_type_new and
 * gtk_object_set really
 */
GtkWidget *
meko_chat_interface_new(Say2Server *server, Say2Client *client)
{
	GtkWidget *chat_interface;

	chat_interface = GTK_WIDGET(gtk_type_new(MEKO_TYPE_CHAT_INTERFACE));

	if (server) {
		gtk_object_set(GTK_OBJECT(chat_interface),
			       "server", server, NULL);
	}
	if (client) {
		gtk_object_set(GTK_OBJECT(chat_interface),
			       "client", client, NULL);
	}

	return chat_interface;
}

/* meko_chat_interface_destroy
 * Destroy an object instance, freeing and unreferencing any data we've
 * been hanging on to
 */
static void
meko_chat_interface_destroy(GtkObject *object)
{
	MekoChatInterface *chat_interface;

	g_return_if_fail(object != NULL);
	g_return_if_fail(MEKO_IS_CHAT_INTERFACE(object));
	
	chat_interface = MEKO_CHAT_INTERFACE(object);

	gtk_object_set(object, "server", NULL, "client", NULL, NULL);

	/* Chain up */
	if (GTK_OBJECT_CLASS(parent_class)->destroy)
		GTK_OBJECT_CLASS(parent_class)->destroy(object);
}

/* meko_chat_interface_set_arg
 * Set object arguments, just call the real function
 */
static void
meko_chat_interface_set_arg(GtkObject *object, GtkArg *arg, guint arg_id)
{
	MekoChatInterface *chat_interface;

	g_return_if_fail(object != NULL);
	g_return_if_fail(MEKO_IS_CHAT_INTERFACE(object));

	chat_interface = MEKO_CHAT_INTERFACE(object);

	switch (arg_id) {
	case ARG_SERVER:
		meko_chat_interface_set_server(chat_interface,
					       GTK_VALUE_POINTER(*arg));
		break;
	case ARG_CLIENT:
		meko_chat_interface_set_client(chat_interface,
					       GTK_VALUE_POINTER(*arg));
		break;
	}
}

/* meko_chat_interface_get_arg
 * Get object arguments, just call the real function
 */
static void
meko_chat_interface_get_arg(GtkObject *object, GtkArg *arg, guint arg_id)
{
	MekoChatInterface *chat_interface;

	g_return_if_fail(object != NULL);
	g_return_if_fail(MEKO_IS_CHAT_INTERFACE(object));

	chat_interface = MEKO_CHAT_INTERFACE(object);

	switch (arg_id) {
	case ARG_SERVER:
		GTK_VALUE_POINTER(*arg) = chat_interface->server;
		break;
	case ARG_CLIENT:
		GTK_VALUE_POINTER(*arg) = chat_interface->client;
		break;
	default:
		arg->type = GTK_TYPE_INVALID;
		break;
	}
}

/* meko_real_chat_interface_remodel
 * Remodel the widget based on the values of it's arguments
 */
static void
meko_real_chat_interface_remodel(MekoChatInterface *chat_interface)
{
	gint order;

	g_return_if_fail(chat_interface != NULL);

	/* Throw away the info box */
	if ((chat_interface->client == NULL)
	    && (chat_interface->info_box != NULL)) {
		gtk_container_remove(
			GTK_CONTAINER(chat_interface->info_box->parent),
			chat_interface->info_box
		);
		chat_interface->info_box = chat_interface->info = NULL;
	}

	/* Throw away the pane */
	if (((chat_interface->client == NULL)
             || !SAY2_IS_CHANNEL(chat_interface->client))
	    && (chat_interface->pane != NULL)) {
		/* Keep the text box, putting it back in the interface */
		gtk_widget_ref(chat_interface->text_box);
		gtk_container_remove(
			GTK_CONTAINER(chat_interface->text_box->parent),
			chat_interface->text_box
		);
		gtk_box_pack_start(GTK_BOX(chat_interface),
				   chat_interface->text_box, TRUE, TRUE, 0);
		gtk_widget_unref(chat_interface->text_box);

		/* Throw away everything else */
		gtk_container_remove(
			GTK_CONTAINER(chat_interface->pane->parent),
			chat_interface->pane
		);
		chat_interface->pane = chat_interface->list = NULL;
	}

	/* Need an info box now */
	if ((chat_interface->client != NULL)
	    && (chat_interface->info_box == NULL)) {
		/* Frame */
		chat_interface->info_box = gtk_frame_new(NULL);
		gtk_frame_set_shadow_type(GTK_FRAME(chat_interface->info_box),
					  GTK_SHADOW_NONE);
		gtk_box_pack_start(GTK_BOX(chat_interface),
				   chat_interface->info_box, FALSE, FALSE, 0);
		gtk_widget_show(chat_interface->info_box);

		/* Info goes in the frame */
		chat_interface->info = meko_client_info_new(NULL);
		meko_client_info_set_with_status(
			MEKO_CLIENT_INFO(chat_interface->info), TRUE
		);
		meko_client_info_set_show_online(
			MEKO_CLIENT_INFO(chat_interface->info), FALSE
		);
		gtk_container_add(GTK_CONTAINER(chat_interface->info_box),
				  chat_interface->info);
		gtk_widget_show(chat_interface->info);
	}

	/* Need a pane now */
	if ((chat_interface->client != NULL)
	    && SAY2_IS_CHANNEL(chat_interface->client)
	    && (chat_interface->pane == NULL)) {
		GtkWidget *scrollwin;

		/* Get the text box out of the box */
		gtk_widget_ref(chat_interface->text_box);
		gtk_container_remove(
			GTK_CONTAINER(chat_interface->text_box->parent),
			chat_interface->text_box
		);

		/* Create the pane */
		chat_interface->pane = gtk_hpaned_new();
		gtk_box_pack_start(GTK_BOX(chat_interface),
				   chat_interface->pane, TRUE, TRUE, 0);
		gtk_widget_show(chat_interface->pane);

		/* Create a scrolling window for the list */
		scrollwin = gtk_scrolled_window_new(NULL, NULL);
		gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrollwin),
					       GTK_POLICY_NEVER,
					       GTK_POLICY_AUTOMATIC);
		gtk_paned_pack2(GTK_PANED(chat_interface->pane), scrollwin,
				FALSE, TRUE);
		gtk_widget_set_usize(scrollwin, 165, -1);
		gtk_widget_show(scrollwin);

		/* Create the list */
		chat_interface->list = meko_client_list_new();
		gtk_list_set_selection_mode(GTK_LIST(chat_interface->list),
					    GTK_SELECTION_MULTIPLE);
		gtk_scrolled_window_add_with_viewport(
			GTK_SCROLLED_WINDOW(scrollwin), chat_interface->list
		);
		gtk_widget_show(chat_interface->list);

		/* Replace the text box here */
		gtk_paned_pack1(GTK_PANED(chat_interface->pane),
				chat_interface->text_box, TRUE, TRUE);
		gtk_widget_unref(chat_interface->text_box);
	}

	/* Point the info box at the client */
	if (chat_interface->client != NULL) {
		meko_client_info_set_client(
			MEKO_CLIENT_INFO(chat_interface->info),
			chat_interface->client
		);
	}

	/* Make sure everything's in the right order */
	order = 0;
	if (chat_interface->info_box) {
		gtk_box_reorder_child(GTK_BOX(chat_interface),
				      chat_interface->info_box, order++);
	}
	if (chat_interface->text_box) {
		gtk_box_reorder_child(GTK_BOX(chat_interface),
				      chat_interface->text_box, order++);
	}
	if (chat_interface->btm_box) {
		gtk_box_reorder_child(GTK_BOX(chat_interface),
				      chat_interface->btm_box, order++);
	}
}

/* meko_chat_interface_remodel
 * Emit the remodel signal
 */
static void
meko_chat_interface_remodel(MekoChatInterface *chat_interface)
{
	g_return_if_fail(chat_interface != NULL);

	gtk_signal_emit(GTK_OBJECT(chat_interface),
			chat_interface_signals[REMODEL]);
}

/* meko_chat_interface_set_server
 * Set what registration server we send commands to, we also watch in case
 * this gets disconnected
 */
void
meko_chat_interface_set_server(MekoChatInterface *chat_interface,
			       Say2Server *server)
{
	Say2User *me;
	
	g_return_if_fail(chat_interface != NULL);

	if (server == chat_interface->server)
		return;

	/* Disconnect the existing signal */
	if (chat_interface->server) {
		gtk_signal_disconnect(GTK_OBJECT(chat_interface->server),
				      chat_interface->server_disconnected_sigh);
		gtk_signal_disconnect(GTK_OBJECT(chat_interface->server),
				      chat_interface->server_message_sigh);
		gtk_object_unref(GTK_OBJECT(chat_interface->server));
	}

	chat_interface->server = server;

	/* Connect a signal */
	if (chat_interface->server) {
		gtk_object_ref(GTK_OBJECT(chat_interface->server));
		gtk_object_sink(GTK_OBJECT(chat_interface->server));
		chat_interface->server_disconnected_sigh = gtk_signal_connect(
			GTK_OBJECT(chat_interface->server), "disconnected",
			GTK_SIGNAL_FUNC(_server_disconnected_cb), chat_interface
		);
		chat_interface->server_message_sigh = gtk_signal_connect(
			GTK_OBJECT(chat_interface->server), "message",
			GTK_SIGNAL_FUNC(_server_message_cb), chat_interface
		);

		/* Update the status menu so it is using the
		 * server's "me" user
		 */
		me = say2_server_get_me(chat_interface->server);
		meko_status_menu_set_user(
			MEKO_STATUS_MENU(chat_interface->status), me
		);
	} else {
		/* Just set it to NULL, it'll be disabled by the remodel */
		meko_status_menu_set_user(
			MEKO_STATUS_MENU(chat_interface->status), NULL
		);

		/* Get rid of client too, they're related to the server */
		meko_chat_interface_set_client(chat_interface, NULL);
	}

	/* Remodel the interface */
	meko_chat_interface_remodel(chat_interface);
}

/* meko_chat_interface_set_client
 * Set what client, either a channel or a user, that we send text to.  We also
 * watch in case this leaves
 */
void
meko_chat_interface_set_client(MekoChatInterface *chat_interface,
			       Say2Client *client)
{
	g_return_if_fail(chat_interface != NULL);

	if (client == chat_interface->client)
		return;

	/* If changing from one channel to another, clear the list */
	if ((client != NULL) && SAY2_IS_CHANNEL(client)
	    && (chat_interface->client != NULL)
	    && SAY2_IS_CHANNEL(chat_interface->client)) {
		gtk_list_clear_items(GTK_LIST(chat_interface->list), 0, -1);
	}

	/* Disconnect the existing signal */
	if (chat_interface->client) {
		if (SAY2_IS_CHANNEL(chat_interface->client)) {
			gtk_signal_disconnect(
				GTK_OBJECT(chat_interface->client),
				chat_interface->channel_signon_sigh
			);
			gtk_signal_disconnect(
				GTK_OBJECT(chat_interface->client),
				chat_interface->channel_signoff_sigh
			);
		}

		gtk_object_unref(GTK_OBJECT(chat_interface->client));
	}

	chat_interface->client = client;

	/* Connect a signal */
	if (chat_interface->client) {
		gtk_object_ref(GTK_OBJECT(chat_interface->client));
		gtk_object_sink(GTK_OBJECT(chat_interface->client));

		if (SAY2_IS_CHANNEL(chat_interface->client)) {
			chat_interface->channel_signon_sigh
				= gtk_signal_connect(
					GTK_OBJECT(chat_interface->client),
					"signon",
					GTK_SIGNAL_FUNC(_channel_signon_cb),
					chat_interface
				);
			chat_interface->channel_signoff_sigh
				= gtk_signal_connect(
					GTK_OBJECT(chat_interface->client),
					"signoff",
					GTK_SIGNAL_FUNC(_channel_signoff_cb),
					chat_interface
				);
		}
	}

	/* Remodel the interface */
	meko_chat_interface_remodel(chat_interface);
}

/* meko_chat_interface_get_server
 * Return what registration server we send commands to
 */
Say2Server *
meko_chat_interface_get_server(MekoChatInterface *chat_interface)
{
	Say2Server *server;
	
	g_return_val_if_fail(chat_interface != NULL, NULL);

	gtk_object_get(GTK_OBJECT(chat_interface), "server", &server, NULL);

	return server;
}

/* meko_chat_interface_get_client
 * Return what client we send messages to
 */
Say2Client *
meko_chat_interface_get_client(MekoChatInterface *chat_interface)
{
	Say2Client *client;
	
	g_return_val_if_fail(chat_interface != NULL, NULL);

	gtk_object_get(GTK_OBJECT(chat_interface), "client", &client, NULL);

	return client;
}

/* meko_chat_interface_get_client_list
 * Return our client list if we have one
 */
GtkWidget *
meko_chat_interface_get_client_list(MekoChatInterface *chat_interface)
{
	g_return_val_if_fail(chat_interface != NULL, NULL);

	return chat_interface->list;
}

/* meko_chat_interface_recv_msg_set
 * Receive all the messages listed in the variable arguments, list is
 * terminated with SAY2_MSG_LAST
 */
void
meko_chat_interface_recv_msg_set(MekoChatInterface *chat_interface, ...)
{
	Say2MessageType message_type;
	va_list		args;

	g_return_if_fail(chat_interface != NULL);

	va_start(args, chat_interface);
	while ((message_type = va_arg(args, Say2MessageType)) != SAY2_MSG_LAST)
	{
		g_return_if_fail(message_type >= SAY2_MSG_ANY);
		g_return_if_fail(message_type < SAY2_MSG_LAST);
		chat_interface->recv_msg[message_type] = TRUE;
	}
	va_end(args);
}

/* meko_chat_interface_recv_msg_unset
 * Don't receive all the messages listed in the variable arguments, list is
 * terminated with SAY2_MSG_LAST
 */
void
meko_chat_interface_recv_msg_unset(MekoChatInterface *chat_interface, ...)
{
	Say2MessageType message_type;
	va_list		args;

	g_return_if_fail(chat_interface != NULL);

	va_start(args, chat_interface);
	while ((message_type = va_arg(args, Say2MessageType)) != SAY2_MSG_LAST)
	{
		g_return_if_fail(message_type >= SAY2_MSG_ANY);
		g_return_if_fail(message_type < SAY2_MSG_LAST);
		chat_interface->recv_msg[message_type] = FALSE;
	}
	va_end(args);
}

/* meko_chat_interface_get_recv_msg
 * Return whether we receive the message given
 */
gboolean
meko_chat_interface_get_recv_msg(MekoChatInterface *chat_interface,
				 Say2MessageType message_type)
{
	g_return_val_if_fail(chat_interface != NULL, FALSE);
	g_return_val_if_fail(message_type >= SAY2_MSG_ANY, FALSE);
	g_return_val_if_fail(message_type < SAY2_MSG_LAST, FALSE);

	return chat_interface->recv_msg[message_type];
}

/* _status_changed_cb
 * User has chosen a new status from the menu, so provided it's different
 * from their status on the server, pop up a dialog to get the rest of the
 * info
 */
static void
_status_changed_cb(MekoStatusMenu *status_menu,
		   MekoChatInterface *chat_interface)
{
	Say2User       *user;
	Say2UserStatus  user_status, old_status;
	GtkWidget      *status_dialog;
	
	g_return_if_fail(status_menu != NULL);
	g_return_if_fail(chat_interface != NULL);

	/* Make sure the server is suitably connected */
	if ((chat_interface->server == NULL)
	    || (say2_server_get_state(chat_interface->server)
		< SAY2_SERVER_ACTIVE)) {
		/* Change it to offline */
		meko_status_menu_set_value(status_menu, SAY2_USER_OFFLINE);
		return;
	}

	/* Find out what the old status was */
	user = say2_server_get_me(chat_interface->server);
	gtk_object_get(GTK_OBJECT(user), "status", &old_status, NULL);
	
	/* No point actually doing anything */
	user_status = meko_status_menu_get_value(status_menu);
	if (user_status == old_status)
		return;

	/* Open a change status dialog */
	status_dialog = meko_status_menu_dialog_new(user_status);
	gtk_object_set_data(GTK_OBJECT(status_dialog), "chat-interface",
			    chat_interface);
	gnome_dialog_set_close(GNOME_DIALOG(status_dialog), TRUE);
	gnome_dialog_close_hides(GNOME_DIALOG(status_dialog), FALSE);
	gtk_signal_connect(GTK_OBJECT(status_dialog), "clicked",
			   GTK_SIGNAL_FUNC(_status_dialog_clicked_cb),
			   (gpointer)say2_user_status_to_str(user_status));
	gtk_widget_show(status_dialog);
}

/* _status_dialog_clicked_cb
 * User has clicked ok in the change status dialog, so set their status and
 * any other information with the server
 */
static void
_status_dialog_clicked_cb(MekoStatusMenuDialog *status_dialog,
			  gint button, const gchar *status)
{
	MekoChatInterface *chat_interface;
	const gchar	  *status_msg;
	gboolean	   receive;

	g_return_if_fail(status_dialog != NULL);
	g_return_if_fail(status != NULL);

	/* Extract the chat interface pointer */
	chat_interface = gtk_object_get_data(GTK_OBJECT(status_dialog),
					     "chat-interface");
	g_return_if_fail(chat_interface != NULL);

	/* User clicked cancel, don't do anything */
	if (button != 0)
		return;

	/* Grab the info out of the dialog */
	status_msg = meko_status_menu_dialog_get_message(status_dialog);
	receive = meko_status_menu_dialog_get_receive(status_dialog);

	/* Send the change to the server */
	say2_server_send(chat_interface->server, "status",
			 "mode:%s text:%s messages:%s",
			 status, status_msg, (receive ? "on" : "off")); 
}

/* _entry_activate_cb
 * User has typed text into the box and pressed enter, send it as a command
 * to the server
 */
static void
_entry_activate_cb(GtkEntry *entry, MekoChatInterface *chat_interface)
{
	const gchar *text;
	gchar       *msg;
	
	g_return_if_fail(entry != NULL);
	g_return_if_fail(chat_interface != NULL);

	text = gtk_entry_get_text(entry);
	msg = meko_user_command(chat_interface->server, chat_interface->client,
				text);
	gtk_entry_set_text(entry, "");

	if (msg != NULL) {
		gchar *full_msg;

		full_msg = g_strdup_printf("*** %s\n", msg);
		gtk_text_insert(GTK_TEXT(chat_interface->text), NULL, NULL,
				NULL, full_msg, strlen(full_msg));
		g_free(full_msg);
	}
}

/* _channel_signon_cb
 * Someone's joined the channel, so add them to the list
 */
static void
_channel_signon_cb(Say2Channel *channel, Say2User *user,
		   MekoChatInterface *chat_interface)
{
	g_return_if_fail(channel != NULL);
	g_return_if_fail(chat_interface != NULL);
	g_return_if_fail(chat_interface->client == SAY2_CLIENT(channel));
	g_return_if_fail(user != NULL);

	meko_client_list_add(MEKO_CLIENT_LIST(chat_interface->list),
			     SAY2_CLIENT(user));
}

/* _channel_signoff_cb
 * Someone's left the channel, so remove them from the list
 */
static void
_channel_signoff_cb(Say2Channel *channel, Say2User *user,
		    MekoChatInterface *chat_interface)
{
	g_return_if_fail(channel != NULL);
	g_return_if_fail(chat_interface != NULL);
	g_return_if_fail(chat_interface->client == SAY2_CLIENT(channel));
	g_return_if_fail(user != NULL);

	meko_client_list_remove(MEKO_CLIENT_LIST(chat_interface->list),
			        SAY2_CLIENT(user));
}

/* _server_disconnected_cb
 * Server has disconnected, so completely disown it
 */
static void
_server_disconnected_cb(Say2Server *server, MekoChatInterface *chat_interface)
{
	g_return_if_fail(server != NULL);
	g_return_if_fail(chat_interface != NULL);
	g_return_if_fail(chat_interface->server == server);

	/* Just disown the server, everything will get done in the right way */
	meko_chat_interface_set_server(chat_interface, NULL);
}

/* _server_message_cb
 * Received a message from the server, check if we want it and if we do
 * add it to our text box
 */
static void
_server_message_cb(Say2Server *server, Say2Client *client,
		   Say2MessageType message_type, GHashTable *hash,
		   MekoChatInterface *chat_interface)
{
	gchar *msg;

	g_return_if_fail(hash != NULL);
	g_return_if_fail(server != NULL);
	g_return_if_fail(chat_interface != NULL);
	g_return_if_fail(chat_interface->server == server);
	g_return_if_fail(message_type >= SAY2_MSG_ANY);
	g_return_if_fail(message_type < SAY2_MSG_LAST);

	/* Capture messages if the message client is NULL, or we're not bound
	 * to a specific client, or if the client is the one we're bound to
	 */
	if ((client != NULL) && (chat_interface->client != NULL)
	    && (client != chat_interface->client)) {
		/* If we're bound to a channel, we check whether the message
		 * client is a member, and if so, capture the message
		 */
		if (SAY2_IS_CHANNEL(chat_interface->client)
		    && SAY2_IS_USER(client)) {
			const gchar *id;
			Say2Channel *chan;
			Say2User    *user;

			chan = SAY2_CHANNEL(chat_interface->client);
			user = SAY2_USER(client);
		
			gtk_object_get(GTK_OBJECT(client), "id", &id, NULL);
			if (say2_channel_get_member(chan, id) != user)
				return;

			/* Check that the user doesn't already have a private
			 * window (FIXME this is more information than I want
			 * this widget to have)
			 */
			if (gtk_object_get_data(GTK_OBJECT(client),
						"private_window") != NULL)
				return;
		} else {
			return;
		}
	}

	/* Message (or SAY2_MSG_ANY) must be TRUE in our array */
	if ((chat_interface->recv_msg[SAY2_MSG_ANY] == FALSE)
	    && (chat_interface->recv_msg[message_type] == FALSE))
		return;

	/* Get the text for the message */
	msg = meko_message_text(server, client, message_type, hash);
	if (msg != NULL) {
		/* Insert the text into the text box */
		gtk_text_insert(GTK_TEXT(chat_interface->text), NULL, NULL,
				NULL, msg, strlen(msg));
	
		/* Free up */
		g_free(msg);
	}
}
