/* meko
 * Copyright (C) 2001 Scott James Remnant <scott@netsplit.com>.
 * All Rights Reserved.
 *
 * meko-client-info.c
 *  - MekoClientInfo class
 *  - A widget, derived from GnomeCanvas, displaying a user's basic information
 * -- 
 * $Id: meko-client-info.c,v 1.27.2.2 2001/11/24 01:41:05 scott Exp $
 * 
 * This file is distributed according to the GNU General Public
 * License.  For full details, read the top of 'main.c' or the
 * file called COPYING that was distributed with this code.
 */

#include <meko.h>

#include <glib.h>
#include <gdk/gdk.h>
#include <gtk/gtkobject.h>
#include <gtk/gtksignal.h>
#include <gtk/gtkwidget.h>
#include <gtk/gtklistitem.h>
#include <gtk/gtktooltips.h>
#include <gtk/gtkselection.h>
#include <gtk/gtkdnd.h>
#include <libgnome/gnome-defs.h>
#include <libgnome/gnome-i18n.h>
#include <libgnome/gnome-util.h>
#include <libgnomeui/gnome-canvas.h>
#include <libgnomeui/gnome-canvas-widget.h>
#include <libgnomeui/gnome-canvas-rect-ellipse.h>
#include <libgnomeui/gnome-canvas-text.h>
#include <gdk-pixbuf/gdk-pixbuf.h>
#include <gdk-pixbuf/gdk-pixbuf-loader.h>
#include <gdk-pixbuf/gnome-canvas-pixbuf.h>

#include "say2/say2-client.h"
#include "say2/say2-user.h"
#include "util/gui/status-image-widget.h"
#include "util/say2/client-label.h"

#include "meko-client-info.h"

/* Arguments
 * MekoSay2Client *client	client to display, either user or channel
 * gboolean	   with_status	include status or tagline information
 * gboolean	   show_online	show the online status glyph, don't hide it
 */
enum {
	ARG_0,
	ARG_CLIENT,
	ARG_WITH_STATUS,
	ARG_SHOW_ONLINE
};

/* Signals */
enum {
	REMODEL,
	RECOLOR,
	LAST_SIGNAL
};

/* Drag types */
enum {
	DRAG_CLIENT_ID_LIST,
	LAST_DRAG
};

/* Fonts */
#define NICK_FONT "-*-helvetica-bold-r-normal-*-14-*-*-*-p-*-iso8859-1"
#define TEXT_FONT "-*-helvetica-medium-r-normal-*-10-*-*-*-p-*-iso8859-1"


/* Forward prototypes for required functions */
static void meko_client_info_init(MekoClientInfo *);
static void meko_client_info_class_init(MekoClientInfoClass *);

/* Forward prototypes for GtkObject methods */
static void meko_client_info_set_arg(GtkObject *, GtkArg *, guint);
static void meko_client_info_get_arg(GtkObject *, GtkArg *, guint);
static void meko_client_info_destroy(GtkObject *);

/* Forward prototypes for GtkWidget methods */
static void meko_client_info_realize(GtkWidget *);
static void meko_client_info_size_request(GtkWidget *, GtkRequisition *);
static void meko_client_info_size_allocate(GtkWidget *, GtkAllocation *);
static void meko_client_info_state_changed(GtkWidget *, GtkStateType);
static void meko_client_info_drag_data_get(GtkWidget *, GdkDragContext *,
					   GtkSelectionData *, guint, guint);

/* Forward prototypes for MekoClientInfo methods */
static void meko_real_client_info_remodel(MekoClientInfo *);
static void meko_real_client_info_recolor(MekoClientInfo *);

/* Forward prototypes for MekoClientInfo method wrappers */
static void meko_client_info_remodel(MekoClientInfo *);
static void meko_client_info_recolor(MekoClientInfo *);

/* Forward prototypes for internal functions */
static void _client_changed_cb(Say2Client *, MekoClientInfo *);


/* Our parent class */
static GnomeCanvasClass *parent_class = NULL;

/* Signal array */
static guint client_info_signals[LAST_SIGNAL] = { 0 };


/* meko_client_info_get_type
 * Returns our GtkType id, allocating it if necessary.
 */
guint
meko_client_info_get_type(void)
{
	static guint client_info_type = 0;

	if (!client_info_type) {
		static const GtkTypeInfo client_info_info = {
			"MekoClientInfo",
			sizeof(MekoClientInfo),
			sizeof(MekoClientInfoClass),
			(GtkClassInitFunc)meko_client_info_class_init,
			(GtkObjectInitFunc)meko_client_info_init,
			/* reserved_1 */ NULL,
			/* reserved_2 */ NULL,
			(GtkClassInitFunc)NULL
		};

		client_info_type = gtk_type_unique(GNOME_TYPE_CANVAS,
						   &client_info_info);
	}

	return client_info_type;
}

/* meko_client_info_class_init
 * Initialise the object class, setting our arguments and overriding methods
 * of parent object classes
 */
static void
meko_client_info_class_init(MekoClientInfoClass *klass)
{
	GtkObjectClass *object_class;
	GtkWidgetClass *widget_class;

	g_return_if_fail(klass != NULL);

	object_class = (GtkObjectClass *)klass;
	widget_class = (GtkWidgetClass *)klass;
	parent_class = gtk_type_class(GNOME_TYPE_CANVAS);

	gtk_object_add_arg_type("MekoClientInfo::client", GTK_TYPE_POINTER,
				GTK_ARG_READWRITE, ARG_CLIENT);
	gtk_object_add_arg_type("MekoClientInfo::with_status", GTK_TYPE_BOOL,
				GTK_ARG_READWRITE, ARG_WITH_STATUS);
	gtk_object_add_arg_type("MekoClientInfo::show_online", GTK_TYPE_BOOL,
				GTK_ARG_READWRITE, ARG_SHOW_ONLINE);

	client_info_signals[REMODEL] = gtk_signal_new(
		"remodel", GTK_RUN_FIRST, object_class->type,
		GTK_SIGNAL_OFFSET(MekoClientInfoClass, remodel),
		gtk_marshal_NONE__NONE,
		GTK_TYPE_NONE, 0
	);

	client_info_signals[RECOLOR] = gtk_signal_new(
		"recolor", GTK_RUN_FIRST, object_class->type,
		GTK_SIGNAL_OFFSET(MekoClientInfoClass, recolor),
		gtk_marshal_NONE__NONE,
		GTK_TYPE_NONE, 0
	);

	gtk_object_class_add_signals(object_class, client_info_signals,
				     LAST_SIGNAL);

	object_class->set_arg = meko_client_info_set_arg;
	object_class->get_arg = meko_client_info_get_arg;
	object_class->destroy = meko_client_info_destroy;

	widget_class->realize = meko_client_info_realize;
	widget_class->size_request = meko_client_info_size_request;
	widget_class->size_allocate = meko_client_info_size_allocate;
	widget_class->state_changed = meko_client_info_state_changed;
	widget_class->drag_data_get = meko_client_info_drag_data_get;

	klass->remodel = meko_real_client_info_remodel;
	klass->recolor = meko_real_client_info_recolor;
}

/* meko_client_info_init
 * Initialise an object instance, creating all of the canvas items that
 * are to go inside.  The setting of locations and contents of these is
 * done on realization.
 */
static void
meko_client_info_init(MekoClientInfo *client_info)
{
	GdkPixbuf *pixbuf;
	gchar     *filename;
	
	static const GtkTargetEntry drag_targets[LAST_DRAG] = {
		{ "text/x-say2-client-id-list", 0, DRAG_CLIENT_ID_LIST }
	};

	g_return_if_fail(client_info != NULL);

	client_info->client = NULL;
	client_info->with_status = FALSE;
	client_info->show_online = FALSE;

	client_info->client_changed_sigh = 0;

	/* Create the tooltips for this */
	client_info->tips = gtk_tooltips_new();
	
	/* Background rectangle so we can change colour
	 * (GnomeCanvas's code sucks and assumes state is NORMAL)
	 */
	client_info->bg = gnome_canvas_item_new(
		gnome_canvas_root(GNOME_CANVAS(client_info)),
		GNOME_TYPE_CANVAS_RECT,
		"x1", 0,0,
		"y1", 0.0,
		"x2", 33.0,
		"y2", 33.0,
		NULL
	);

	/* Rectangle surrounds the image */
	client_info->rect = gnome_canvas_item_new(
		gnome_canvas_root(GNOME_CANVAS(client_info)),
		GNOME_TYPE_CANVAS_RECT,
		"x1", 0.0,
		"y1", 0.0,
		"x2", 33.0,
		"y2", 33.0,
		"fill_color", "white",
		"outline_color", "black",
		NULL
	);

	/* Nickname text */
	client_info->nick = gnome_canvas_item_new(
		gnome_canvas_root(GNOME_CANVAS(client_info)),
		GNOME_TYPE_CANVAS_TEXT,
		"x", 36.0,
		"y", 2.0,
		"anchor", GTK_ANCHOR_NORTH_WEST,
		"font", NICK_FONT,
		NULL
	);

	/* E-Mail text, this can move */
	client_info->email = gnome_canvas_item_new(
		gnome_canvas_root(GNOME_CANVAS(client_info)),
		GNOME_TYPE_CANVAS_TEXT,
		"anchor", GTK_ANCHOR_SOUTH_WEST,
		"font", TEXT_FONT,
		NULL
	);

	/* Status or tagline text, this might dissapear */
	client_info->text = gnome_canvas_item_new(
		gnome_canvas_root(GNOME_CANVAS(client_info)),
		GNOME_TYPE_CANVAS_TEXT,
		"x", 36.0,
		"y", 31.0,
		"anchor", GTK_ANCHOR_SOUTH_WEST,
		"font", TEXT_FONT,
		NULL
	);

	/* Client icon image */
	client_info->icon = gnome_canvas_item_new(
		gnome_canvas_root(GNOME_CANVAS(client_info)),
		GNOME_TYPE_CANVAS_PIXBUF,
		"x", 1.0,
		"y", 1.0,
		NULL
	);

	/* Status image, might not be there */
	client_info->status = gnome_canvas_item_new(
		gnome_canvas_root(GNOME_CANVAS(client_info)),
		GNOME_TYPE_CANVAS_PIXBUF,
		NULL
	);

	/* Load the hand icon */
	filename = g_concat_dir_and_file(MEKO_IMAGESDIR, "hand.png");
	pixbuf = gdk_pixbuf_new_from_file(filename);
	g_free(filename);

	/* Status hand image, might not be there */
	client_info->hand = gnome_canvas_item_new(
		gnome_canvas_root(GNOME_CANVAS(client_info)),
		GNOME_TYPE_CANVAS_PIXBUF,
		"pixbuf", pixbuf,
		NULL
	);
	gnome_canvas_item_hide(client_info->hand);
	gdk_pixbuf_unref(pixbuf);

	/* Initialise drag-and-drop support */
	gtk_drag_source_set(GTK_WIDGET(client_info), GDK_BUTTON1_MASK,
			    drag_targets, LAST_DRAG, GDK_ACTION_COPY);
}

/* meko_client_info_new
 * Create a new object instance, just a wrapper around gtk_type_new and
 * gtk_object_set really
 */
GtkWidget *
meko_client_info_new(Say2Client *client)
{
	GtkWidget *client_info;

	client_info = GTK_WIDGET(gtk_type_new(MEKO_TYPE_CLIENT_INFO));

	if (client)
		gtk_object_set(GTK_OBJECT(client_info), "client", client, NULL);

	return client_info;
}

/* meko_client_info_destroy
 * Destroy an object instance, freeing and unreferencing any data we've
 * been hanging on to
 */
static void
meko_client_info_destroy(GtkObject *object)
{
	MekoClientInfo *client_info;

	g_return_if_fail(object != NULL);
	g_return_if_fail(MEKO_IS_CLIENT_INFO(object));
	
	client_info = MEKO_CLIENT_INFO(object);

	gtk_object_set(object, "client", NULL, NULL);

	/* Chain up */
	if (GTK_OBJECT_CLASS(parent_class)->destroy)
		GTK_OBJECT_CLASS(parent_class)->destroy(object);
}

/* meko_client_info_set_arg
 * Set object arguments
 */
static void
meko_client_info_set_arg(GtkObject *object, GtkArg *arg, guint arg_id)
{
	MekoClientInfo *client_info;

	g_return_if_fail(object != NULL);
	g_return_if_fail(MEKO_IS_CLIENT_INFO(object));

	client_info = MEKO_CLIENT_INFO(object);

	switch (arg_id) {
	case ARG_CLIENT:
		meko_client_info_set_client(client_info,
					    GTK_VALUE_POINTER(*arg));
		break;
	case ARG_WITH_STATUS:
		client_info->with_status = GTK_VALUE_BOOL(*arg);
		meko_client_info_remodel(client_info);
		break;
	case ARG_SHOW_ONLINE:
		client_info->show_online = GTK_VALUE_BOOL(*arg);
		meko_client_info_remodel(client_info);
		break;
	}
}

/* meko_client_info_get_arg
 * Get object arguments
 */
static void
meko_client_info_get_arg(GtkObject *object, GtkArg *arg, guint arg_id)
{
	MekoClientInfo *client_info;

	g_return_if_fail(object != NULL);
	g_return_if_fail(MEKO_IS_CLIENT_INFO(object));

	client_info = MEKO_CLIENT_INFO(object);

	switch (arg_id) {
	case ARG_CLIENT:
		GTK_VALUE_POINTER(*arg) = client_info->client;
		break;
	case ARG_WITH_STATUS:
		GTK_VALUE_BOOL(*arg) = client_info->with_status;
		break;
	case ARG_SHOW_ONLINE:
		GTK_VALUE_BOOL(*arg) = client_info->show_online;
		break;
	default:
		arg->type = GTK_TYPE_INVALID;
		break;
	}
}

/* meko_client_info_realize
 * Ensure that everything is painted in the right colours and in the right
 * place
 */
static void
meko_client_info_realize(GtkWidget *widget)
{
	MekoClientInfo *client_info;

	g_return_if_fail(widget != NULL);
	g_return_if_fail(MEKO_IS_CLIENT_INFO(widget));

	client_info = MEKO_CLIENT_INFO(widget);

	/* Chain up */
	if (GTK_WIDGET_CLASS(parent_class)->realize)
		GTK_WIDGET_CLASS(parent_class)->realize(widget);

	meko_client_info_remodel(client_info);
	meko_client_info_recolor(client_info);
}

/* meko_client_info_size_request
 * Return how big we want to be
 */
static void
meko_client_info_size_request(GtkWidget *widget, GtkRequisition *requisition)
{
	MekoClientInfo *client_info;
	double		x1, y1, x2, y2;
	double		x_bound = 159, y_bound = 33;

	g_return_if_fail(widget != NULL);
	g_return_if_fail(MEKO_IS_CLIENT_INFO(widget));

	client_info = MEKO_CLIENT_INFO(widget);

	/* Chain up */
	if (GTK_WIDGET_CLASS(parent_class)->size_request) {
		GTK_WIDGET_CLASS(parent_class)->size_request(widget,
							     requisition);
	}

	/* Compare current bounds to the various components */
	gnome_canvas_item_get_bounds(client_info->rect, &x1, &y1, &x2, &y2);
	x_bound = MAX(x_bound, x2);
	y_bound = MAX(y_bound, y2);
	gnome_canvas_item_get_bounds(client_info->nick, &x1, &y1, &x2, &y2);
	x_bound = MAX(x_bound, x2);
	y_bound = MAX(y_bound, y2);
	gnome_canvas_item_get_bounds(client_info->email, &x1, &y1, &x2, &y2);
	x_bound = MAX(x_bound, x2);
	y_bound = MAX(y_bound, y2);
	gnome_canvas_item_get_bounds(client_info->text, &x1, &y1, &x2, &y2);
	x_bound = MAX(x_bound, x2);
	y_bound = MAX(y_bound, y2);
	gnome_canvas_item_get_bounds(client_info->status, &x1, &y1, &x2, &y2);
	x_bound = MAX(x_bound, x2);
	y_bound = MAX(y_bound, y2);

	/* Set the requisition so it fits in everything */
	requisition->width = x_bound + 1;
	requisition->height = y_bound + 1;
}

/* meko_client_info_size_allocate
 * Set how much space we've actually got and adjust things that need
 * adjusting for that
 */
static void
meko_client_info_size_allocate(GtkWidget *widget, GtkAllocation *allocation)
{
	MekoClientInfo *client_info;
	
	g_return_if_fail(widget != NULL);
	g_return_if_fail(MEKO_IS_CLIENT_INFO(widget));

	client_info = MEKO_CLIENT_INFO(widget);

	/* Store the allocation */
	widget->allocation = *allocation;

	/* Set the canvas scroll region */
	gnome_canvas_set_scroll_region(GNOME_CANVAS(widget), 0.0, 0.0,
				       (double)(allocation->width - 1),
				       (double)(allocation->height - 1));

	/* Set the size of the background item */
	gnome_canvas_item_set(client_info->bg, "x1", 0.0, "y1", 0.0,
			      "x2", (double)(allocation->width - 1),
			      "y2", (double)(allocation->height - 1),
			      NULL);

	/* Chain up */
	if (GTK_WIDGET_CLASS(parent_class)->size_allocate) {
		GTK_WIDGET_CLASS(parent_class)->size_allocate(widget,
							      allocation);
	}
}

/* meko_client_info_state_changed
 * Adjusts the colours to that of the new state
 */
static void
meko_client_info_state_changed(GtkWidget *widget, GtkStateType old_state)
{
	g_return_if_fail(widget != NULL);
	g_return_if_fail(MEKO_IS_CLIENT_INFO(widget));

	/* Chain up */
	if (GTK_WIDGET_CLASS(parent_class)->state_changed) {
		GTK_WIDGET_CLASS(parent_class)->state_changed(widget,
							      old_state);
	}

	/* Sort it out */
	meko_client_info_recolor(MEKO_CLIENT_INFO(widget));
}

/* meko_client_info_drag_data_get
 * Provide the ID of our client to whatever widget just got the widget dragged
 * and dropped onto it.
 */
static void
meko_client_info_drag_data_get(GtkWidget *widget, GdkDragContext *drag_context,
			       GtkSelectionData *selection_data, guint info,
			       guint time)
{
	MekoClientInfo *client_info;
	const gchar    *id;
	
	g_return_if_fail(widget != NULL);
	g_return_if_fail(MEKO_IS_CLIENT_INFO(widget));

	client_info = MEKO_CLIENT_INFO(widget);

	/* Chain up */
	if (GTK_WIDGET_CLASS(parent_class)->drag_data_get) {
		GTK_WIDGET_CLASS(parent_class)->drag_data_get(
			widget, drag_context, selection_data, info, time
		);
	}

	/* Pass the id of our client as the selection data */
	if (info == DRAG_CLIENT_ID_LIST) {
		GString *buff;

		/* Terminate the id of the client with \r\n */
		buff = g_string_new("");
		gtk_object_get(GTK_OBJECT(client_info->client),
			       "id", &id, NULL);
		g_string_sprintfa(buff, "%s\r\n", id);

		/* Set the selection data */
		gtk_selection_data_set(selection_data, selection_data->target,
				       8, buff->str, buff->len);
		g_string_free(buff, TRUE);
	}
}

/* meko_real_client_info_remodel
 * Remodel the widget based on it's arguments, always leaves the widget in a
 * "beautiful" state.
 */
static void
meko_real_client_info_remodel(MekoClientInfo *client_info)
{
	Say2Client  *client;
	const gchar *str;
	char	    *icon;
	int	     icon_sz;

	g_return_if_fail(client_info != NULL);

	/* Clear every entry, we reset them all in a minute.  Seems less
	 * efficient, but it's just plain damned easier
	 */
	gnome_canvas_item_set(client_info->nick, "text", "", NULL);
	gnome_canvas_item_set(client_info->email, "text", "", NULL);
	gnome_canvas_item_set(client_info->text, "text", "", NULL);
	gnome_canvas_item_set(client_info->icon, "pixbuf", NULL, NULL);
	gnome_canvas_item_set(client_info->status, "pixbuf", NULL, NULL);

	/* Don't bother if there's no client */
	if (!client_info->client)
		return;
	client = SAY2_CLIENT(client_info->client);

	/* Set the e-mail, that's easy */
	gtk_object_get(GTK_OBJECT(client), "email", &str, NULL);
	if (str) {
		GString *buff;

		buff = g_string_new("");
		g_string_sprintfa(buff, "<%s>", str);

		gnome_canvas_item_set(client_info->email,
				      "text", buff->str, NULL);

		g_string_free(buff, TRUE);
	}

	/* Get what to display it as */
	str = meko_say2_client_label(client);
	gnome_canvas_item_set(client_info->nick, "text", str, NULL);

	/* Need to set the longer status text, and possibly things like
	 * tooltips.  This is a lot more complex for users than it is for
	 * channels, because users have several things we need to work
	 * out
	 */
	if (SAY2_IS_USER(client)) {
		Say2UserStatus  user_status;
		gboolean	messages;
		GString        *buff;
		gchar          *msg;

		buff = g_string_new("");

		/* Switch based on status enum */
		gtk_object_get(GTK_OBJECT(client), "status", &user_status,
			       "status_msg", &msg, "messages", &messages, NULL);
		switch (user_status) {
		case SAY2_USER_ACTIVE:
			/* Only say ACTIVE if they've specifically given a
			 * message.  Otherwise we leave buff empty, which
			 * means we'll magically replace it with their
			 * tagline below.
			 */
			if (msg)
				buff = g_string_append(buff, _("ACTIVE"));
			break;
		case SAY2_USER_OFFLINE:
			buff = g_string_append(buff, _("OFFLINE"));
			break;
		case SAY2_USER_BUSY:
			buff = g_string_append(buff, _("BUSY"));
			break;
		case SAY2_USER_AWAY:
			buff = g_string_append(buff, _("AWAY"));
			break;
		case SAY2_USER_IDLE:
			buff = g_string_append(buff, _("IDLE"));
			break;
		default:
			/* Unknown status, *sigh* */
			buff = g_string_append(buff, _("unknown"));
			break;
		}

		/* Append the status message to the description */
		if (msg) {
			buff = g_string_append(buff, ": ");
			buff = g_string_append(buff, msg);
		}

		/* Only show an online icon if we want to, or if they're
		 * not receiving messages
		 */
		if ((user_status != SAY2_USER_ACTIVE) || (messages == FALSE)
		    || (client_info->show_online)) {
			GdkPixbuf *pixbuf;

			pixbuf = meko_create_status_image(user_status);
			if (pixbuf) {
				gnome_canvas_item_show(client_info->status);
				gnome_canvas_item_set(client_info->status,
					              "pixbuf", pixbuf, NULL);
				gdk_pixbuf_unref(pixbuf);
			} else {
				gnome_canvas_item_hide(client_info->status);
			}

			if (messages) {
				gnome_canvas_item_hide(client_info->hand);
			} else {
				gnome_canvas_item_show(client_info->hand);
			}
		} else {
			gnome_canvas_item_hide(client_info->status);
			gnome_canvas_item_hide(client_info->hand);
		}

		/* Get their tagline, we'll need it in a sec */
		gtk_object_get(GTK_OBJECT(client), "tagline", &str, NULL);

		/* Set the status text if we need to */
		if (client_info->with_status)  {
			/* If there's nothing in the buffer, usually when we're
			 * active with no message, show their tagline instead
			 */
			gnome_canvas_item_set(client_info->text, "text",
					      (buff->len ? buff->str : str),
					      NULL);
		}

		/* Stick the tagline on the front of the tooltip */
		if (buff->len && str)
			buff = g_string_prepend(buff, "\n");
		if (str)
			buff = g_string_prepend(buff, str);

		/* Set the tooltip */
		gtk_tooltips_set_tip(client_info->tips, GTK_WIDGET(client_info),
				     (buff->len ? buff->str : NULL), NULL);

		/* Free the buffer */
		g_string_free(buff, TRUE);

	} else {
		/* Get the tagline and set the tooltip */
		gtk_object_get(GTK_OBJECT(client), "tagline", &str, NULL);
		gtk_tooltips_set_tip(client_info->tips, GTK_WIDGET(client_info),
				     str, NULL);
		
		/* Set the text to the tagline too */
		if (client_info->with_status)
			gnome_canvas_item_set(client_info->text, "text",
					      str, NULL);
	}

	/* Users and channels can have icons, either of faces or logos and
	 * the like.
	 */
	icon_sz = say2_client_get_icon(client, &icon);
	if (icon && icon_sz) {
		GdkPixbufLoader *loader;
		GdkPixbuf	*pixbuf, *scaled;

		/* Because we've already got that data in memory,
		 * we'll load and convert it using a GdkPixbufLoader
		 */
		loader = gdk_pixbuf_loader_new();
		gdk_pixbuf_loader_write(loader, icon, icon_sz);
		pixbuf = gdk_pixbuf_loader_get_pixbuf(loader);
		gdk_pixbuf_loader_close(loader);

		/* Scale it */
		if (pixbuf) {
			scaled = gdk_pixbuf_scale_simple(pixbuf, 32, 32,
							 GDK_INTERP_HYPER);
			if (scaled) {
				gdk_pixbuf_unref(pixbuf);
				pixbuf = scaled;
			}

			/* Add it to the canvas */
			gnome_canvas_item_show(client_info->icon);
			gnome_canvas_item_set(client_info->icon,
					      "pixbuf", pixbuf, NULL);
			gdk_pixbuf_unref(pixbuf);
		} else {
			gnome_canvas_item_hide(client_info->icon);
		}
	}

	/* Shuffle everything around so it's in the right place */
	if (client_info->with_status) {
		GdkPixbuf *pixbuf;
		double	   x1, y1, x2, y2;

		/* E-mail follows nickname */
		gnome_canvas_item_get_bounds(client_info->nick,
					     &x1, &y1, &x2, &y2);
		gnome_canvas_item_set(client_info->email,
				      "x", x2 + 2, "y", y2 - 2, NULL);

		/* If there's a status icon, position it before the textual
		 * description
		 */
		gtk_object_get(GTK_OBJECT(client_info->status),
			       "pixbuf", &pixbuf, NULL);
		gnome_canvas_item_set(client_info->text,
				      "x", (pixbuf ? 54.0 : 36.0), NULL);
		gnome_canvas_item_set(client_info->status,
				      "x", 36.0, "y", 17.0, NULL);
		gnome_canvas_item_set(client_info->hand,
				      "x", 42.0, "y", 17.0, NULL);
		
	} else {
		double x1, y1, x2, y2;

		/* E-mail at the bottom */
		gnome_canvas_item_set(client_info->email,
				      "x", 36.0, "y", 31.0, NULL);

		/* Status goes just after the nickname */
		gnome_canvas_item_get_bounds(client_info->nick,
					     &x1, &y1, &x2, &y2);
		gnome_canvas_item_set(client_info->status,
				      "x", x2 + 4, "y", 2.0, NULL);
		gnome_canvas_item_set(client_info->hand,
				      "x", x2 + 10, "y", 2.0, NULL);
	}
}

/* meko_real_client_info_recolor
 * Adjust the colours of the widget to match those of the current state
 */
static void
meko_real_client_info_recolor(MekoClientInfo *client_info)
{
	GdkColor     *fg, *bg;
	GtkStateType  state;

	g_return_if_fail(client_info != NULL);

	/* Get the new foreground and background colours */
	state = GTK_WIDGET(client_info)->state;
	fg = &(GTK_WIDGET(client_info)->style->fg[state]);
	if (GTK_IS_LIST_ITEM(GTK_WIDGET(client_info)->parent)
	    && (state == GTK_STATE_NORMAL)) {
		bg = &(GTK_WIDGET(client_info)->style->base[state]);
	} else {
		bg = &(GTK_WIDGET(client_info)->style->bg[state]);
	}

	/* Adjust the colours of text widgets */
	gnome_canvas_item_set(client_info->nick, "fill_color_gdk", fg, NULL);
	gnome_canvas_item_set(client_info->email, "fill_color_gdk", fg, NULL);
	gnome_canvas_item_set(client_info->text, "fill_color_gdk", fg, NULL);

	/* Adjust the colour of the background widget */
	gnome_canvas_item_set(client_info->bg, "fill_color_gdk", bg, NULL);
}

/* meko_client_info_remodel
 * Emit the remodel signal
 */
static void
meko_client_info_remodel(MekoClientInfo *client_info)
{
	g_return_if_fail(client_info != NULL);

	gtk_signal_emit(GTK_OBJECT(client_info), client_info_signals[REMODEL]);
}

/* meko_client_info_recolor
 * Emit the recolor signal
 */
static void
meko_client_info_recolor(MekoClientInfo *client_info)
{
	g_return_if_fail(client_info != NULL);

	gtk_signal_emit(GTK_OBJECT(client_info), client_info_signals[RECOLOR]);
}

/* meko_client_info_set_client
 * Convenience wrapper around gtk_object_set
 */
void
meko_client_info_set_client(MekoClientInfo *client_info, Say2Client *client)
{
	g_return_if_fail(client_info != NULL);

	if (client_info->client) {
		gtk_signal_disconnect(GTK_OBJECT(client_info->client),
				      client_info->client_changed_sigh);
		gtk_object_unref(GTK_OBJECT(client_info->client));
	}

	client_info->client = client;
	
	if (client_info->client) {
		gtk_object_ref(GTK_OBJECT(client_info->client));
		gtk_object_sink(GTK_OBJECT(client_info->client));
		client_info->client_changed_sigh = gtk_signal_connect(
			GTK_OBJECT(client_info->client), "changed",
			GTK_SIGNAL_FUNC(_client_changed_cb), client_info
		);
	}

	meko_client_info_remodel(client_info);
}

/* meko_client_info_set_with_status
 * Convenience wrapper around gtk_object_set
 */
void
meko_client_info_set_with_status(MekoClientInfo *client_info,
				 gboolean with_status)
{
	g_return_if_fail(client_info != NULL);

	gtk_object_set(GTK_OBJECT(client_info), "with_status", with_status,
		       NULL);
}

/* meko_client_info_set_show_online
 * Convenience wrapper around gtk_object_set
 */
void
meko_client_info_set_show_online(MekoClientInfo *client_info,
				 gboolean show_online)
{
	g_return_if_fail(client_info != NULL);

	gtk_object_set(GTK_OBJECT(client_info), "show_online", show_online,
		       NULL);
}

/* meko_client_info_get_client
 * Convenience wrapper around gtk_object_get
 */
Say2Client *
meko_client_info_get_client(MekoClientInfo *client_info)
{
	Say2Client *client;

	g_return_val_if_fail(client_info != NULL, NULL);

	gtk_object_get(GTK_OBJECT(client_info), "client", &client, NULL);

	return client;
}

/* meko_client_info_get_with_status
 * Convenience wrapper around gtk_object_get
 */
gboolean
meko_client_info_get_with_status(MekoClientInfo *client_info)
{
	gboolean with_status;

	g_return_val_if_fail(client_info != NULL, FALSE);

	gtk_object_get(GTK_OBJECT(client_info), "with_status", &with_status,
		       NULL);

	return with_status;
}

/* meko_client_info_get_show_online
 * Convenience wrapper around gtk_object_get
 */
gboolean
meko_client_info_get_show_online(MekoClientInfo *client_info)
{
	gboolean show_online;

	g_return_val_if_fail(client_info != NULL, FALSE);

	gtk_object_get(GTK_OBJECT(client_info), "show_online", &show_online,
		       NULL);

	return show_online;
}

/* _client_changed_cb
 * Remodel the widget because the client has had information changed
 */
static void
_client_changed_cb(Say2Client *client, MekoClientInfo *client_info)
{
	g_return_if_fail(client_info != NULL);

	meko_client_info_remodel(client_info);
}
