/* meko
 * Copyright (C) 2001 Scott James Remnant <scott@netsplit.com>.
 * All Rights Reserved.
 *
 * meko-client-list.c
 *  - MekoClientList class
 *  - A widget, derived from GtkList that holds and displays a number of
 *    Say2Client structures in order using MekoClientInfo's to display them.
 * -- 
 * $Id: meko-client-list.c,v 1.16 2001/09/17 10:31:02 scott Exp $
 * 
 * This file is distributed according to the GNU General Public
 * License.  For full details, read the top of 'main.c' or the
 * file called COPYING that was distributed with this code.
 */

#include <meko.h>

#include <glib.h>
#include <gdk/gdk.h>
#include <gtk/gtkstyle.h>
#include <gtk/gtkobject.h>
#include <gtk/gtksignal.h>
#include <gtk/gtkwidget.h>
#include <gtk/gtkcontainer.h>
#include <gtk/gtklist.h>
#include <gtk/gtklistitem.h>
#include <gtk/gtkselection.h>
#include <gtk/gtkdnd.h>

#include "say2/say2-client.h"
#include "say2/say2-user.h"
#include "util/say2/client-label.h"
#include "widgets/client-info/meko-client-info.h"

#include "meko-client-list.h"

/* Arguments
 * gboolean	 show_online	show the online status glyph, don't hide it
 */
enum {
	ARG_0,
	ARG_SHOW_ONLINE
};

/* Signals */
enum {
	ACTIVATE,
	DROPPED,
	LAST_SIGNAL
};

/* Drag types */
enum {
	DRAG_CLIENT_ID_LIST,
	LAST_DRAG
};


/* Forward prototypes for required functions */
static void meko_client_list_init(MekoClientList *);
static void meko_client_list_class_init(MekoClientListClass *);

/* Forward prototypes for GtkObject methods */
static void meko_client_list_set_arg(GtkObject *, GtkArg *, guint);
static void meko_client_list_get_arg(GtkObject *, GtkArg *, guint);

/* Forward prototypes for GtkWidget methods */
static void meko_client_list_size_request(GtkWidget *, GtkRequisition *);
static void meko_client_list_drag_data_received(GtkWidget *, GdkDragContext *,
						gint, gint, GtkSelectionData *,
						guint, guint);

/* Forward prototypes for internal functions */
static void _set_show_online(GtkListItem *, MekoClientList *);
static gint _get_position(MekoClientList *, Say2Client *);
static gint _info_button_cb(MekoClientInfo *, GdkEventButton *,
			    MekoClientList *);
static void _client_changed_cb(Say2Client *, MekoClientList *);

/* Our parent class */
static GtkListClass *parent_class = NULL;

/* Signal array */
static guint client_list_signals[LAST_SIGNAL] = { 0 };


/* meko_client_list_get_type
 * Returns our GtkType id, allocating it if necessary.
 */
guint
meko_client_list_get_type(void)
{
	static guint client_list_type = 0;

	if (!client_list_type) {
		static const GtkTypeInfo client_list_info = {
			"MekoClientList",
			sizeof(MekoClientList),
			sizeof(MekoClientListClass),
			(GtkClassInitFunc)meko_client_list_class_init,
			(GtkObjectInitFunc)meko_client_list_init,
			/* reserved_1 */ NULL,
			/* reserved_2 */ NULL,
			(GtkClassInitFunc)NULL
		};

		client_list_type = gtk_type_unique(GTK_TYPE_LIST,
						   &client_list_info);
	}

	return client_list_type;
}

/* meko_client_list_class_init
 * Initialise the object class, setting our arguments and overriding methods
 * of parent object classes
 */
static void
meko_client_list_class_init(MekoClientListClass *klass)
{
	GtkObjectClass *object_class;
	GtkWidgetClass *widget_class;

	g_return_if_fail(klass != NULL);

	object_class = (GtkObjectClass *)klass;
	widget_class = (GtkWidgetClass *)klass;
	parent_class = gtk_type_class(gtk_list_get_type());

	gtk_object_add_arg_type("MekoClientList::show_online", GTK_TYPE_BOOL,
				GTK_ARG_READWRITE, ARG_SHOW_ONLINE);

	client_list_signals[ACTIVATE] = gtk_signal_new(
		"activate", GTK_RUN_FIRST, object_class->type,
		GTK_SIGNAL_OFFSET(MekoClientListClass, activate),
		gtk_marshal_NONE__POINTER,
		GTK_TYPE_NONE, 1, GTK_TYPE_POINTER
	);

	client_list_signals[DROPPED] = gtk_signal_new(
		"dropped", GTK_RUN_FIRST, object_class->type,
		GTK_SIGNAL_OFFSET(MekoClientListClass, dropped),
		gtk_marshal_NONE__STRING,
		GTK_TYPE_NONE, 1, GTK_TYPE_STRING
	);

	gtk_object_class_add_signals(object_class, client_list_signals,
				     LAST_SIGNAL);

	object_class->set_arg = meko_client_list_set_arg;
	object_class->get_arg = meko_client_list_get_arg;

	widget_class->size_request = meko_client_list_size_request;
	widget_class->drag_data_received = meko_client_list_drag_data_received;

	klass->activate = NULL;
}

/* meko_client_list_init
 * Initialise an object instance, very little to do here other than set our
 * members to their default values
 */
static void
meko_client_list_init(MekoClientList *client_list)
{
	static const GtkTargetEntry drag_targets[LAST_DRAG] = {
		{ "text/x-say2-client-id-list", 0, DRAG_CLIENT_ID_LIST }
	};

	g_return_if_fail(client_list != NULL);

	client_list->show_online = FALSE;

	/* Initialise drag-and-drop support */
	gtk_drag_dest_set(GTK_WIDGET(client_list), GTK_DEST_DEFAULT_ALL,
			  drag_targets, LAST_DRAG, GDK_ACTION_COPY);
}

/* meko_client_list_new
 * Create a new object instance, just a wrapper around gtk_type_new really
 */
GtkWidget *
meko_client_list_new(void)
{
	GtkWidget *widget;

	widget = GTK_WIDGET(gtk_type_new(MEKO_TYPE_CLIENT_LIST));

	return widget;
}

/* _set_show_online
 * foreach callback to set the show_online argument of children to the
 * same as the client_list given
 */
static void
_set_show_online(GtkListItem *list_item, MekoClientList *client_list)
{
	GList	       *clist;
	gboolean	show_online;
	MekoClientInfo *client_info;
	
	g_return_if_fail(list_item != NULL);
	g_return_if_fail(GTK_IS_LIST_ITEM(list_item));

	g_return_if_fail(client_list != NULL);
	g_return_if_fail(MEKO_IS_CLIENT_LIST(client_list));

	clist = gtk_container_children(GTK_CONTAINER(list_item));
	if (clist == NULL)
		return;
	g_return_if_fail(MEKO_IS_CLIENT_INFO(clist->data));

	client_info = MEKO_CLIENT_INFO(clist->data);

	show_online = meko_client_list_get_show_online(client_list);
	meko_client_info_set_show_online(client_info, show_online);
}

/* meko_client_list_set_arg
 * Set object arguments
 */
static void
meko_client_list_set_arg(GtkObject *object, GtkArg *arg, guint arg_id)
{
	MekoClientList *client_list;

	g_return_if_fail(object != NULL);
	g_return_if_fail(MEKO_IS_CLIENT_LIST(object));

	client_list = MEKO_CLIENT_LIST(object);

	switch (arg_id) {
	case ARG_SHOW_ONLINE:
		client_list->show_online = GTK_VALUE_BOOL(*arg);
		gtk_container_foreach(GTK_CONTAINER(client_list),
				      (GtkCallback)_set_show_online,
				      (gpointer)client_list);
		break;
	}
}

/* meko_client_list_get_arg
 * Get object arguments
 */
static void
meko_client_list_get_arg(GtkObject *object, GtkArg *arg, guint arg_id)
{
	MekoClientList *client_list;

	g_return_if_fail(object != NULL);
	g_return_if_fail(MEKO_IS_CLIENT_LIST(object));

	client_list = MEKO_CLIENT_LIST(object);

	switch (arg_id) {
	case ARG_SHOW_ONLINE:
		GTK_VALUE_BOOL(*arg) = client_list->show_online;
		break;
	default:
		arg->type = GTK_TYPE_INVALID;
		break;
	}
}

/* meko_client_list_size_request
 * GtkList does a good job of figuring out how tall this widget needs to be,
 * but we want to restrict width to a specific value.
 */
static void
meko_client_list_size_request(GtkWidget *widget, GtkRequisition *requisition)
{
	MekoClientList *client_list;

	g_return_if_fail(widget != NULL);
	g_return_if_fail(MEKO_IS_CLIENT_LIST(widget));

	client_list = MEKO_CLIENT_LIST(widget);

	/* Chain up */
	if (GTK_WIDGET_CLASS(parent_class)->size_request) {
		GTK_WIDGET_CLASS(parent_class)->size_request(widget,
							     requisition);
	}

	/* Width is 160 + border_width * 2 */
	requisition->width = 160 + GTK_CONTAINER(client_list)->border_width * 2;
}

/* meko_client_list_drag_data_received
 * Get the ID of the Say2Client that was just dropped onto us
 */
static void
meko_client_list_drag_data_received(GtkWidget *widget,
				    GdkDragContext *drag_context,
				    gint x, gint y,
				    GtkSelectionData *selection_data,
				    guint info, guint time)
{
	MekoClientList *client_list;

	g_return_if_fail(widget != NULL);
	g_return_if_fail(MEKO_IS_CLIENT_LIST(widget));

	client_list = MEKO_CLIENT_LIST(widget);

	/* Chain up */
	if (GTK_WIDGET_CLASS(parent_class)->drag_data_received) {
		GTK_WIDGET_CLASS(parent_class)->drag_data_received(
			widget, drag_context, x, y, selection_data, info, time
		);
	}

	/* Get the data */
	if ((selection_data->length >= 0) && (selection_data->format == 8)) {
		/* FIXME Need to do something intelligent here, sigh */
		gtk_drag_finish(drag_context, TRUE, FALSE, time);
		return;
	}

	gtk_drag_finish(drag_context, FALSE, FALSE, time);
}

/* meko_client_list_set_show_online
 * Convenience wrapper around gtk_object_set
 */
void
meko_client_list_set_show_online(MekoClientList *client_list,
				 gboolean show_online)
{
	g_return_if_fail(client_list != NULL);

	gtk_object_set(GTK_OBJECT(client_list), "show_online", show_online,
		       NULL);
}

/* meko_client_list_get_show_online
 * Convenience wrapper around gtk_object_get
 */
gboolean
meko_client_list_get_show_online(MekoClientList *client_list)
{
	gboolean show_online;

	g_return_val_if_fail(client_list != NULL, FALSE);

	gtk_object_get(GTK_OBJECT(client_list), "show_online", &show_online,
		       NULL);

	return show_online;
}

/* _get_position
 * Figure out where to put the item in the list
 */
static gint
_get_position(MekoClientList *client_list, Say2Client *client)
{
	const gchar *label;
	GList	    *list;
	gint	     position;

	g_return_val_if_fail(client_list != NULL, -1);
	g_return_val_if_fail(client != NULL, -1);

	/* Figure out where in the list to put it */
	label = meko_say2_client_label(client);
	list = gtk_container_children(GTK_CONTAINER(client_list));
	for (position = 0; list; list = g_list_next(list), position++) {
		GList	       *cl_item;
		MekoClientInfo *cl_info;
		Say2Client     *cl;
		const gchar    *cl_label;

		cl_item = gtk_container_children(GTK_CONTAINER(list->data));
		if ((cl_item == NULL) || !MEKO_IS_CLIENT_INFO(cl_item->data))
			continue;

		cl_info = MEKO_CLIENT_INFO(cl_item->data);
		cl = meko_client_info_get_client(cl_info);
		if (cl == NULL)
			continue;

		cl_label = meko_say2_client_label(cl);
		if (g_strcasecmp(label, cl_label) < 0)
			break;
	}

	return position;
}

/* meko_client_list_add
 * Add a child to the list, placing at the right point so the list is sorted.
 */
void
meko_client_list_add(MekoClientList *client_list, Say2Client *client)
{
	GtkWidget *client_info;
	GtkWidget *item;

	g_return_if_fail(client_list != NULL);
	g_return_if_fail(client != NULL);

	/* We need a packing GtkListItem to insert something into the list */
	item = gtk_list_item_new();
	gtk_container_set_border_width(GTK_CONTAINER(item), 1);
	gtk_widget_show(item);

	/* Create the widget to represent the client given */
	client_info = gtk_widget_new(MEKO_TYPE_CLIENT_INFO,
				     "client", client,
				     "with_status", FALSE,
				     "show_online", client_list->show_online,
				     NULL);
	gtk_signal_connect(GTK_OBJECT(client_info), "button-press-event",
			   GTK_SIGNAL_FUNC(_info_button_cb), client_list);
	gtk_container_add(GTK_CONTAINER(item), client_info);
	gtk_widget_show(client_info);

	/* Hook the changed signal of the client all the time it's parent
	 * is alive
	 */
	gtk_signal_connect_while_alive(GTK_OBJECT(client), "changed",
				       GTK_SIGNAL_FUNC(_client_changed_cb),
				       client_list, GTK_OBJECT(client_info));

	/* Insert into the right place */
	gtk_list_insert_items(GTK_LIST(client_list),
			      g_list_append(NULL, item),
			      _get_position(client_list, client));
}

/* meko_client_list_remove
 * Remove a child from the list.
 */
void
meko_client_list_remove(MekoClientList *client_list, Say2Client *client)
{
	GList *list;

	g_return_if_fail(client_list != NULL);

	/* Figure out where in the list it is */
	list = gtk_container_children(GTK_CONTAINER(client_list));
	for (; list; list = g_list_next(list)) {
		GList	       *cl_item;
		MekoClientInfo *cl_info;
		Say2Client     *cl;

		cl_item = gtk_container_children(GTK_CONTAINER(list->data));
		if ((cl_item == NULL) || !MEKO_IS_CLIENT_INFO(cl_item->data))
			continue;

		cl_info = MEKO_CLIENT_INFO(cl_item->data);
		cl = meko_client_info_get_client(cl_info);
		if (client == cl) {
			gtk_container_remove(GTK_CONTAINER(client_list),
					     GTK_WIDGET(list->data));
			break;
		}
	}
}

/* _info_button_cb
 * Check to see whether it was a double-click, and if so emit an activate
 * signal for the client involved
 */
static gint
_info_button_cb(MekoClientInfo *client_info, GdkEventButton *event,
		MekoClientList *client_list)
{
	Say2Client *client;
	
	g_return_val_if_fail(client_info != NULL, FALSE);
	g_return_val_if_fail(client_list != NULL, FALSE);
	g_return_val_if_fail(event != NULL, FALSE);

	if ((event->type == GDK_2BUTTON_PRESS) && (event->button == 1)) {
		client = meko_client_info_get_client(client_info);
		gtk_signal_emit(GTK_OBJECT(client_list),
				client_list_signals[ACTIVATE], client);
		return TRUE;
	}

	return FALSE;
}

/* _client_changed_cb
 * Reposition the client in the list when it changes
 */
static void
_client_changed_cb(Say2Client *client, MekoClientList *client_list)
{
	gint       position;
	GtkWidget *item;
	GList     *list;

	g_return_if_fail(client != NULL);
	g_return_if_fail(client_list != NULL);

	/* Figure out where in the list it is at the moment */
	item = NULL;
	list = gtk_container_children(GTK_CONTAINER(client_list));
	for (position = 0; list; list = g_list_next(list), position++) {
		GList	       *cl_item;
		MekoClientInfo *cl_info;
		Say2Client     *cl;

		cl_item = gtk_container_children(GTK_CONTAINER(list->data));
		if ((cl_item == NULL) || !MEKO_IS_CLIENT_INFO(cl_item->data))
			continue;

		cl_info = MEKO_CLIENT_INFO(cl_item->data);
		cl = meko_client_info_get_client(cl_info);
		if (client == cl) {
			item = list->data;
			break;
		}
	}
	g_return_if_fail(item != NULL);

	/* Move the item */
	gtk_widget_ref(item);
	gtk_container_remove(GTK_CONTAINER(client_list), item);
	gtk_list_insert_items(GTK_LIST(client_list), g_list_append(NULL, item),
			      _get_position(client_list, client));
	gtk_widget_unref(item);
}
