/* meko
 * Copyright (C) 2001 Scott James Remnant <scott@netsplit.com>.
 * All Rights Reserved.
 *
 * meko-status-menu-dialog.c
 *  - MekoStatusMenuDialog class
 *  - A dialog box that prompts the user for a message while they're offline
 *    and asks whether they'd like to receive messages too
 * -- 
 * $Id: meko-status-menu-dialog.c,v 1.3.2.1 2001/11/23 22:19:55 scott Exp $
 * 
 * This file is distributed according to the GNU General Public
 * License.  For full details, read the top of 'main.c' or the
 * file called COPYING that was distributed with this code.
 */

#include <meko.h>

#include <ctype.h>

#include <glib.h>
#include <gdk/gdk.h>
#include <gtk/gtkobject.h>
#include <gtk/gtksignal.h>
#include <gtk/gtkwidget.h>
#include <gtk/gtklabel.h>
#include <gtk/gtkentry.h>
#include <gtk/gtkcheckbutton.h>
#include <libgnome/gnome-defs.h>
#include <libgnome/gnome-i18n.h>
#include <libgnome/gnome-util.h>
#include <libgnomeui/gnome-dialog.h>
#include <libgnomeui/gnome-stock.h>
#include <gdk-pixbuf/gdk-pixbuf.h>
#include <glade/glade-xml.h>

#include "say2/say2-user.h"
#include "util/gui/pixbuf-canvas.h"
#include "util/gui/status-image-widget.h"

#include "meko-status-menu-dialog.h"


/* Signals */
enum {
	CONSTRUCT,
	LAST_SIGNAL
};

/* Fonts */
#define STATUS_FONT "-*-helvetica-bold-r-normal-*-14-*-*-*-p-*-iso8859-1"


/* Forward prototypes for required functions */
static void meko_status_menu_dialog_init(MekoStatusMenuDialog *);
static void meko_status_menu_dialog_class_init(MekoStatusMenuDialogClass *);

/* Forward prototypes for GtkObject methods */
static void meko_status_menu_dialog_set_arg(GtkObject *, GtkArg *, guint);
static void meko_status_menu_dialog_get_arg(GtkObject *, GtkArg *, guint);
static void meko_status_menu_dialog_destroy(GtkObject *);

/* Forward prototypes for MekoStatusMenuDialog methods */
static void meko_real_status_menu_dialog_construct(MekoStatusMenuDialog *,
						   Say2UserStatus);

/* Forward prototypes for MekoStatusMenuDialog method wrappers */
static void meko_status_menu_dialog_construct(MekoStatusMenuDialog *,
					      Say2UserStatus);

/* Forward prototypes for internal functions */
static void _change_label(GtkLabel *, const gchar *);


/* Our parent class */
static GnomeDialogClass *parent_class = NULL;

/* Signal array */
static guint status_dialog_signals[LAST_SIGNAL] = { 0 };


/* meko_status_menu_dialog_get_type
 * Returns our GtkType id, allocating it if necessary
 */
guint
meko_status_menu_dialog_get_type(void)
{
	static guint status_dialog_type = 0;

	if (!status_dialog_type) {
		static const GtkTypeInfo status_dialog_info = {
			"MekoStatusMenuDialog",
			sizeof(MekoStatusMenuDialog),
			sizeof(MekoStatusMenuDialogClass),
			(GtkClassInitFunc)meko_status_menu_dialog_class_init,
			(GtkObjectInitFunc)meko_status_menu_dialog_init,
			/* reserved_1 */ NULL,
			/* reserved_2 */ NULL,
			(GtkClassInitFunc)NULL
		};

		status_dialog_type = gtk_type_unique(GNOME_TYPE_DIALOG,
						     &status_dialog_info);
	}

	return status_dialog_type;
}

/* meko_status_menu_dialog_class_init
 * Initialise the object class, setting our arguments, signals and overriding
 * methods of parent object classes
 */
static void
meko_status_menu_dialog_class_init(MekoStatusMenuDialogClass *klass)
{
	GtkObjectClass *object_class;

	g_return_if_fail(klass != NULL);

	object_class = (GtkObjectClass *)klass;
	parent_class = gtk_type_class(GNOME_TYPE_DIALOG);

	status_dialog_signals[CONSTRUCT] = gtk_signal_new(
		"construct", GTK_RUN_FIRST, object_class->type,
		GTK_SIGNAL_OFFSET(MekoStatusMenuDialogClass, construct),
		gtk_marshal_NONE__INT,
		GTK_TYPE_NONE, 1, GTK_TYPE_INT
	);

	gtk_object_class_add_signals(object_class, status_dialog_signals,
				     LAST_SIGNAL);

	object_class->set_arg = meko_status_menu_dialog_set_arg;
	object_class->get_arg = meko_status_menu_dialog_get_arg;
	object_class->destroy = meko_status_menu_dialog_destroy;

	klass->construct = meko_real_status_menu_dialog_construct;
}

/* meko_status_menu_dialog_init
 * Initialise an object instance, set default values and emit the construct
 * signal to have the dialog itself built
 */
static void
meko_status_menu_dialog_init(MekoStatusMenuDialog *status_dialog)
{
	g_return_if_fail(status_dialog != NULL);

	status_dialog->entry_message = NULL;
	status_dialog->checkbutton_receive = NULL;
	
	/* Set the window stuff */
	gtk_window_set_title(GTK_WINDOW(status_dialog), _("Change Status"));
	gtk_window_set_wmclass(GTK_WINDOW(status_dialog), "status-menu-dialog",
			       "Meko");
	gtk_window_set_policy(GTK_WINDOW(status_dialog), FALSE, FALSE, TRUE);
	
	/* Add the buttons we want */
	gnome_dialog_append_button(GNOME_DIALOG(status_dialog),
				   GNOME_STOCK_BUTTON_OK);
	gnome_dialog_append_button(GNOME_DIALOG(status_dialog),
				   GNOME_STOCK_BUTTON_CANCEL);
	gnome_dialog_set_default(GNOME_DIALOG(status_dialog), 0);
}

/* meko_status_menu_dialog_new
 * Create a new object intsance, just a wrapper around gtk_type_new really
 */
GtkWidget *
meko_status_menu_dialog_new(Say2UserStatus status)
{
	GtkWidget *widget;

	widget = GTK_WIDGET(gtk_type_new(MEKO_TYPE_STATUS_MENU_DIALOG));

	/* Call construct to get stuff from Glade and stuff it in */
	meko_status_menu_dialog_construct(MEKO_STATUS_MENU_DIALOG(widget),
					  status);

	return widget;
}

/* meko_status_menu_dialog_destroy
 * Destroy an object instance, freeing and unreferencing anydata we've
 * been hanging on to
 */
static void
meko_status_menu_dialog_destroy(GtkObject *object)
{
	MekoStatusMenuDialog *status_dialog;

	g_return_if_fail(object != NULL);
	g_return_if_fail(MEKO_IS_STATUS_MENU_DIALOG(object));

	status_dialog = MEKO_STATUS_MENU_DIALOG(object);

	/* Chain up */
	if (GTK_OBJECT_CLASS(parent_class)->destroy)
		GTK_OBJECT_CLASS(parent_class)->destroy(object);
}

/* meko_status_menu_dialog_set_arg
 * Set object arguments
 */
static void
meko_status_menu_dialog_set_arg(GtkObject *object, GtkArg *arg, guint arg_id)
{
	MekoStatusMenuDialog *status_dialog;

	g_return_if_fail(object != NULL);
	g_return_if_fail(MEKO_IS_STATUS_MENU_DIALOG(object));

	status_dialog = MEKO_STATUS_MENU_DIALOG(object);
	switch (arg_id) {
	}
}

/* meko_status_menu_dialog_get_arg
 * Get object arguments
 */
static void
meko_status_menu_dialog_get_arg(GtkObject *object, GtkArg *arg, guint arg_id)
{
	MekoStatusMenuDialog *status_dialog;

	g_return_if_fail(object != NULL);
	g_return_if_fail(MEKO_IS_STATUS_MENU_DIALOG(object));

	status_dialog = MEKO_STATUS_MENU_DIALOG(object);
	switch (arg_id) {
	default:
		arg->type = GTK_TYPE_INVALID;
		break;
	}
}

/* _change_label
 * Assume the text in a label widget contains a %s and printf the other
 * parameter into it before setting it again
 */
static void
_change_label(GtkLabel *label, const gchar *str)
{
	gchar *format;
	gchar *text;

	gtk_label_get(label, &format);
		
	text = g_strdup_printf(format, str);
	gtk_label_set(label, text);
	
	g_free(text);
}

/* meko_real_status_menu_dialog_construct
 * Construct the dialog from the Glade source
 */
static void
meko_real_status_menu_dialog_construct(MekoStatusMenuDialog *status_dialog,
				       Say2UserStatus status)
{
	GladeXML    *gui;
	GtkWidget   *widget;
	gchar	    *filename;
	const gchar *statustext;

	g_return_if_fail(status_dialog != NULL);

	/* Load the interface from the MEKO_GLADEDIR directory */
	filename = g_concat_dir_and_file(MEKO_GLADEDIR,
					 "status-menu-dialog.glade");
	gui = glade_xml_new(filename, NULL);
	g_free(filename);

	/* Grab the vbox from the glade interface, and stuff it into ourselves
	 */
	widget = glade_xml_get_widget(gui, "dialog-vbox");
	if (widget != NULL) {
		gtk_widget_ref(widget);
		gtk_container_remove(GTK_CONTAINER(widget->parent), widget);
		gtk_box_pack_start(GTK_BOX(GNOME_DIALOG(status_dialog)->vbox),
				   widget, TRUE, TRUE, 0);
		gtk_widget_unref(widget);
		gtk_widget_show(widget);
	}

	/* Get the status as text */
	statustext = say2_user_status_to_str(status);

	/* The status label at the top is done in a large font with the
	 * first letter capitalised
	 */
	widget = glade_xml_get_widget(gui, "label-status");
	if ((widget != NULL) && GTK_IS_LABEL(widget)) {
		GtkStyle *style;
		gchar    *ustatus;

		/* Set the style to the resource and copy it */
		gtk_widget_set_rc_style(widget);
		style = gtk_style_copy(gtk_widget_get_style(widget));

		/* Load the new font, and if we could, set the style */
		style->font = gdk_font_load(STATUS_FONT);
		if (style->font != NULL)
			gtk_widget_set_style(widget, style);
		gtk_style_unref(style);

		/* Uppercase the first character and set the label */
		ustatus = g_strdup(statustext);
		ustatus[0] = toupper(ustatus[0]);
		_change_label(GTK_LABEL(widget), ustatus);
		g_free(ustatus);
	}

	/* Just fix the blurb label to contain the status */
	widget = glade_xml_get_widget(gui, "label-blurb");
	if ((widget != NULL) && GTK_IS_LABEL(widget)) {
		_change_label(GTK_LABEL(widget), statustext);
	}

	/* Get the entry box widget, give it keyboard focus and save it */
	widget = glade_xml_get_widget(gui, "entry-msg");
	if ((widget != NULL) && GTK_IS_ENTRY(widget)) {
		gtk_widget_grab_focus(widget);
		status_dialog->entry_message = widget;

		/* Close the dialog when enter is pressed here */
		gnome_dialog_editable_enters(GNOME_DIALOG(status_dialog),
					     GTK_EDITABLE(widget));
	}

	/* Get the check box, grab it's child (which should be a label) and
	 * add our status to it.  Then save the checkbox
	 */
	widget = glade_xml_get_widget(gui, "checkbutton-receive");
	if ((widget != NULL) && GTK_IS_CHECK_BUTTON(widget)) {
		GList *children;

		/* First child should be it's label */
		children = gtk_container_children(GTK_CONTAINER(widget));
		if ((children != NULL) && (children->data != NULL)
		    && GTK_IS_LABEL(children->data)) {
			_change_label(GTK_LABEL(children->data), statustext);
		}

		status_dialog->checkbutton_receive = widget;
	}

	/* Show an image for the new status, scaled to a bigger size */
	widget = glade_xml_get_widget(gui, "vbox-status-image");
	if ((widget != NULL) && GTK_IS_CONTAINER(widget)) {
		GdkPixbuf *pixbuf;
		GtkWidget *canvas;

		pixbuf = meko_create_status_image(status);
		if (pixbuf) {
			GdkPixbuf *scaled;

			scaled = gdk_pixbuf_scale_simple(pixbuf, 24, 24,
							 GDK_INTERP_HYPER);
			if (scaled) {
				gdk_pixbuf_unref(pixbuf);
				pixbuf = scaled;
			}
		}

		canvas = meko_create_pixbuf_canvas(pixbuf);
		gtk_container_add(GTK_CONTAINER(widget), canvas);
	}

	/* Don't need glade anymore */
	gtk_object_unref(GTK_OBJECT(gui));
}

/* meko_status_menu_dialog_construct
 * Emit the construct signal
 */
static void
meko_status_menu_dialog_construct(MekoStatusMenuDialog *status_dialog,
				  Say2UserStatus status)
{
	g_return_if_fail(status_dialog != NULL);

	gtk_signal_emit(GTK_OBJECT(status_dialog),
			status_dialog_signals[CONSTRUCT], status);
}

/* meko_status_menu_dialog_get_message
 * Return the contents of the message entry box
 */
const gchar *
meko_status_menu_dialog_get_message(MekoStatusMenuDialog *status_dialog)
{
	g_return_val_if_fail(status_dialog != NULL, NULL);

	if (status_dialog->entry_message == NULL)
		return NULL;

	return gtk_entry_get_text(GTK_ENTRY(status_dialog->entry_message));
}

/* meko_status_menu_dialog_get_receive
 * Return whether the "receive messages" (don't cork) box ix checked
 */
gboolean
meko_status_menu_dialog_get_receive(MekoStatusMenuDialog *status_dialog)
{
	g_return_val_if_fail(status_dialog != NULL, TRUE);

	if (status_dialog->checkbutton_receive == NULL)
		return TRUE;

	return gtk_toggle_button_get_active(
		GTK_TOGGLE_BUTTON(status_dialog->checkbutton_receive)
	);
}
