/* meko
 * Copyright (C) 2001 Scott James Remnant <scott@netsplit.com>.
 * All Rights Reserved.
 *
 * meko-status-menu.c
 *  - MekoStatusMenu class
 *  - A widget, derived from GtkButton that allows a user to select their
 *    state from the list of possible states.  Displays an icon for the state
 *    and their nickname
 * -- 
 * $Id: meko-status-menu.c,v 1.8.4.1 2001/11/23 22:19:55 scott Exp $
 * 
 * This file is distributed according to the GNU General Public
 * License.  For full details, read the top of 'main.c' or the
 * file called COPYING that was distributed with this code.
 */

#include <meko.h>

#include <glib.h>
#include <gdk/gdk.h>
#include <gtk/gtkobject.h>
#include <gtk/gtksignal.h>
#include <gtk/gtkaccelgroup.h>
#include <gtk/gtkwidget.h>
#include <gtk/gtkbutton.h>
#include <gtk/gtkcontainer.h>
#include <gtk/gtkmenu.h>
#include <gtk/gtkmenuitem.h>
#include <gtk/gtkbox.h>
#include <gtk/gtkhbox.h>
#include <gtk/gtklabel.h>
#include <gtk/gtkmisc.h>
#include <libgnome/gnome-defs.h>
#include <libgnome/gnome-i18n.h>
#include <libgnomeui/gtkpixmapmenuitem.h>

#include "say2/say2-user.h"
#include "util/gui/pixbuf-pixmap.h"
#include "util/gui/status-image-widget.h"
#include "util/say2/client-label.h"

#include "meko-status-menu.h"

/* Arguments
 * Say2User	  *user		user to copy data from and display (optional)
 * gchar	  *text		text to display on label
 */
enum {
	ARG_0,
	ARG_USER,
	ARG_TEXT
};

/* Signals */
enum {
	CHANGED,
	LAST_SIGNAL
};


/* Forward prototypes for required functions */
static void meko_status_menu_init(MekoStatusMenu *);
static void meko_status_menu_class_init(MekoStatusMenuClass *);

/* Forward prototypes for GtkObject methods */
static void meko_status_menu_set_arg(GtkObject *, GtkArg *, guint);
static void meko_status_menu_get_arg(GtkObject *, GtkArg *, guint);
static void meko_status_menu_destroy(GtkObject *);

/* Forward prototypes for GtkWidget methods */
static gint meko_status_menu_button_press_event(GtkWidget *, GdkEventButton *);

/* Forward prototypes for GtkContainer methods */
static GtkType meko_status_menu_child_type(GtkContainer *);

/* Forward prototypes for MekoStatusMenu method wrappers */
static void meko_status_menu_changed(MekoStatusMenu *);

/* Forward prototypes for internal functions */
static GtkWidget *_create_menu_item(GtkMenu *, GtkAccelGroup *, Say2UserStatus,
				    const gchar *);
static GtkWidget *_create_menu(MekoStatusMenu *);
static void	  _menu_detach_cb(GtkWidget *, GtkMenu *);
static void	  _menu_deactivate_cb(GtkWidget *, MekoStatusMenu *);
static void	  _menu_position_cb(GtkWidget *, gint *, gint *,
				    MekoStatusMenu *);
static void	  _user_changed_cb(Say2User *, MekoStatusMenu *);


/* Our parent class */
static GtkButtonClass *parent_class = NULL;

/* Signal array */
static guint status_menu_signals[LAST_SIGNAL] = { 0 };


/* meko_status_menu_get_type
 * Returns our GtkType id, allocating it if necessary.
 */
guint
meko_status_menu_get_type(void)
{
	static guint status_menu_type = 0;

	if (!status_menu_type) {
		static const GtkTypeInfo status_menu_info = {
			"MekoStatusMenu",
			sizeof(MekoStatusMenu),
			sizeof(MekoStatusMenuClass),
			(GtkClassInitFunc)meko_status_menu_class_init,
			(GtkObjectInitFunc)meko_status_menu_init,
			/* reserved_1 */ NULL,
			/* reserved_2 */ NULL,
			(GtkClassInitFunc)NULL
		};

		status_menu_type = gtk_type_unique(GTK_TYPE_BUTTON,
						   &status_menu_info);
	}

	return status_menu_type;
}

/* meko_status_menu_class_init
 * Initialise the object class, setting our arguments and overriding methods
 * of parent object classes
 */
static void
meko_status_menu_class_init(MekoStatusMenuClass *klass)
{
	GtkObjectClass    *object_class;
	GtkWidgetClass    *widget_class;
	GtkContainerClass *container_class;

	g_return_if_fail(klass != NULL);

	object_class = (GtkObjectClass *)klass;
	widget_class = (GtkWidgetClass *)klass;
	container_class = (GtkContainerClass *)klass;
	parent_class = gtk_type_class(GTK_TYPE_BUTTON);

	gtk_object_add_arg_type("MekoStatusMenu::user", GTK_TYPE_POINTER,
				GTK_ARG_READWRITE, ARG_USER);
	gtk_object_add_arg_type("MekoStatusMenu::text", GTK_TYPE_STRING,
				GTK_ARG_READWRITE, ARG_TEXT);

	status_menu_signals[CHANGED] = gtk_signal_new(
		"changed", GTK_RUN_FIRST, object_class->type,
		GTK_SIGNAL_OFFSET(MekoStatusMenuClass, changed),
		gtk_marshal_NONE__NONE,
		GTK_TYPE_NONE, 0
	);

	gtk_object_class_add_signals(object_class, status_menu_signals,
				     LAST_SIGNAL);

	object_class->set_arg = meko_status_menu_set_arg;
	object_class->get_arg = meko_status_menu_get_arg;
	object_class->destroy = meko_status_menu_destroy;

	widget_class->button_press_event = meko_status_menu_button_press_event;

	container_class->child_type = meko_status_menu_child_type;

	klass->changed = NULL;
}

/* _create_menu_item
 * Create a menu item in the popup
 */
static GtkWidget *
_create_menu_item(GtkMenu *menu, GtkAccelGroup *accel_group,
		  Say2UserStatus status, const gchar *text)
{
	GtkWidget *item, *w;
	GdkPixbuf *pixbuf;
	gint       keyval;

	g_return_val_if_fail(menu != NULL, NULL);
	g_return_val_if_fail(accel_group != NULL, NULL);
	
	item = gtk_pixmap_menu_item_new();

	/* Load a pixbuf containing the status image, and render it into a
	 * GtkPixmap before adding to the menu (does alpha the right way)
	 */
	pixbuf = meko_create_status_image(status);
	if (pixbuf) {
		w = meko_create_pixbuf_pixmap(pixbuf, 128);
		gtk_pixmap_menu_item_set_pixmap(GTK_PIXMAP_MENU_ITEM(item), w);
	}

	/* Set the label */
	w = gtk_label_new(text);
	gtk_misc_set_alignment(GTK_MISC(w), 0.0, 0.5);
	gtk_container_add(GTK_CONTAINER(item), w);
	gtk_widget_show(w);

	/* Parse the label for underlines and set an accelerator */
	keyval = gtk_label_parse_uline(GTK_LABEL(w), text);
	gtk_widget_add_accelerator(item, "activate_item", accel_group,
				   keyval, 0, 0);

	/* Set the data */
	gtk_object_set_data(GTK_OBJECT(item), "status",
			    GUINT_TO_POINTER(status));

	/* Append, show and return */
	gtk_menu_append(menu, item);
	gtk_widget_show(item);
	return item;
}

/* _create_menu
 * Create the popup menu that gets shown when the button is clicked
 */
static GtkWidget *
_create_menu(MekoStatusMenu *status_menu)
{
	GtkAccelGroup *accel_group;
	GtkWidget     *menu;

	g_return_val_if_fail(status_menu != NULL, NULL);

	menu = gtk_menu_new();

	/* Create an accelerator group */
	accel_group = gtk_accel_group_new();
	gtk_accel_group_attach(accel_group, GTK_OBJECT(menu));

	/* Add the menu items */
	_create_menu_item(GTK_MENU(menu), accel_group, SAY2_USER_ACTIVE,
			  _("_Active"));
	_create_menu_item(GTK_MENU(menu), accel_group, SAY2_USER_BUSY,
			  _("_Busy"));
	_create_menu_item(GTK_MENU(menu), accel_group, SAY2_USER_AWAY,
			  _("A_way"));
	_create_menu_item(GTK_MENU(menu), accel_group, SAY2_USER_IDLE,
			  _("_Idle"));

	return menu;
}

/* meko_status_menu_init
 * Initialise an object instance, setting our default values and creating the
 * popup menu we use
 */
static void
meko_status_menu_init(MekoStatusMenu *status_menu)
{
	g_return_if_fail(status_menu != NULL);

	GTK_WIDGET_UNSET_FLAGS(GTK_WIDGET(status_menu), GTK_CAN_FOCUS);

	gtk_button_set_relief(GTK_BUTTON(status_menu), GTK_RELIEF_NONE);

	status_menu->user = NULL;
	status_menu->text = NULL;

	status_menu->status = status_menu->set_status = SAY2_USER_UNKNOWN;

	status_menu->user_changed_sigh = 0;

	/* Create and Attach the menu */
	status_menu->menu = _create_menu(MEKO_STATUS_MENU(status_menu));
	gtk_menu_attach_to_widget(GTK_MENU(status_menu->menu),
				  GTK_WIDGET(status_menu), _menu_detach_cb);
	gtk_signal_connect(GTK_OBJECT(status_menu->menu), "deactivate",
			   GTK_SIGNAL_FUNC(_menu_deactivate_cb), status_menu);

	/* Construct the contents */

	status_menu->box = gtk_hbox_new(FALSE, 4);
	gtk_container_add(GTK_CONTAINER(status_menu), status_menu->box);
	gtk_widget_show(status_menu->box);

	status_menu->label = gtk_label_new("");
	gtk_misc_set_alignment(GTK_MISC(status_menu->label), 0.0, 0.5);
	gtk_box_pack_end(GTK_BOX(status_menu->box), status_menu->label,
			 TRUE, TRUE, 0);
	gtk_widget_show(status_menu->label);

	status_menu->pixmap = NULL;
}

/* meko_status_menu_new
 * Create a new object instance, just a wrapper around gtk_type_new and
 * gtk_object_set really
 */
GtkWidget *
meko_status_menu_new(Say2User *user)
{
	GtkWidget *status_menu;

	status_menu = GTK_WIDGET(gtk_type_new(MEKO_TYPE_STATUS_MENU));

	if (user)
		gtk_object_set(GTK_OBJECT(status_menu), "user", user, NULL);

	return status_menu;
}

/* meko_status_menu_destroy
 * Destroy an object instance, freeing and unreferencing any data we've
 * been hanging on to
 */
static void
meko_status_menu_destroy(GtkObject *object)
{
	MekoStatusMenu *status_menu;

	g_return_if_fail(object != NULL);
	g_return_if_fail(MEKO_IS_STATUS_MENU(object));
	
	status_menu = MEKO_STATUS_MENU(object);

	if (status_menu->menu)
		gtk_widget_destroy(status_menu->menu);

	gtk_object_set(object, "user", NULL, NULL);

	/* Chain up */
	if (GTK_OBJECT_CLASS(parent_class)->destroy)
		GTK_OBJECT_CLASS(parent_class)->destroy(object);
}

/* meko_status_menu_set_arg
 * Set object arguments, just call the real function
 */
static void
meko_status_menu_set_arg(GtkObject *object, GtkArg *arg, guint arg_id)
{
	MekoStatusMenu *status_menu;

	g_return_if_fail(object != NULL);
	g_return_if_fail(MEKO_IS_STATUS_MENU(object));

	status_menu = MEKO_STATUS_MENU(object);

	switch (arg_id) {
	case ARG_USER:
		meko_status_menu_set_user(status_menu, GTK_VALUE_POINTER(*arg));
		break;
	case ARG_TEXT:
		meko_status_menu_set_text(status_menu, GTK_VALUE_STRING(*arg));
		break;
	}
}

/* meko_status_menu_get_arg
 * Get object arguments, just call the real function
 */
static void
meko_status_menu_get_arg(GtkObject *object, GtkArg *arg, guint arg_id)
{
	MekoStatusMenu *status_menu;

	g_return_if_fail(object != NULL);
	g_return_if_fail(MEKO_IS_STATUS_MENU(object));

	status_menu = MEKO_STATUS_MENU(object);

	switch (arg_id) {
	case ARG_USER:
		GTK_VALUE_POINTER(*arg) = status_menu->user;
		break;
	case ARG_TEXT:
		GTK_VALUE_STRING(*arg) = status_menu->text;
		break;
	default:
		arg->type = GTK_TYPE_INVALID;
		break;
	}
}

/* meko_status_menu_button_press_event
 * User pressed the button, so popup the menu
 */
static gint
meko_status_menu_button_press_event(GtkWidget *widget, GdkEventButton *event)
{
	MekoStatusMenu *status_menu;

	g_return_val_if_fail(widget != NULL, FALSE);
	g_return_val_if_fail(MEKO_IS_STATUS_MENU(widget), FALSE);
	g_return_val_if_fail(event != NULL, FALSE);

	status_menu = MEKO_STATUS_MENU(widget);

	if ((event->type == GDK_BUTTON_PRESS) && (event->button == 1)) {
		gtk_menu_popup(GTK_MENU(status_menu->menu), NULL, NULL,
			       (GtkMenuPositionFunc)_menu_position_cb,
			       status_menu, event->button, event->time);
		return TRUE;
	}

	return FALSE;
}

/* meko_status_menu_child_type
 * Tell the world that we don't want children!
 */
static GtkType
meko_status_menu_child_type(GtkContainer *container)
{
	return GTK_TYPE_NONE;
}

/* meko_status_menu_changed
 * Emit the status changed signal
 */
static void
meko_status_menu_changed(MekoStatusMenu *status_menu)
{
	g_return_if_fail(status_menu != NULL);

	gtk_signal_emit(GTK_OBJECT(status_menu), status_menu_signals[CHANGED]);
}

/* meko_status_menu_set_user
 * Set what user we're going to automatically model and watch for changes
 */
void
meko_status_menu_set_user(MekoStatusMenu *status_menu, Say2User *user)
{
	g_return_if_fail(status_menu != NULL);

	/* Disconnect the existing user signal */
	if (status_menu->user) {
		gtk_signal_disconnect(GTK_OBJECT(status_menu->user),
				      status_menu->user_changed_sigh);
		gtk_object_unref(GTK_OBJECT(status_menu->user));
	}

	status_menu->user = user;

	/* Connect a signal */
	if (status_menu->user) {
		gtk_object_ref(GTK_OBJECT(status_menu->user));
		gtk_object_sink(GTK_OBJECT(status_menu->user));
		status_menu->user_changed_sigh = gtk_signal_connect(
			GTK_OBJECT(status_menu->user), "changed",
			GTK_SIGNAL_FUNC(_user_changed_cb), status_menu
		);

		/* Update our status as if the user had changed */
		_user_changed_cb(status_menu->user, status_menu);
	} else {
		meko_status_menu_set_text(status_menu, "");
		status_menu->set_status = SAY2_USER_UNKNOWN;
		meko_status_menu_set_value(status_menu, SAY2_USER_UNKNOWN);
	}
}

/* meko_status_menu_set_text
 * Set the label of the button
 */
void
meko_status_menu_set_text(MekoStatusMenu *status_menu, const gchar *text)
{
	g_return_if_fail(status_menu != NULL);
	g_return_if_fail(text != NULL);

	if (status_menu->text)
		g_free(status_menu->text);

	status_menu->text = g_strdup(text);

	if (status_menu->label)
		gtk_label_set_text(GTK_LABEL(status_menu->label), text);
}

/* meko_status_menu_get_user
 * Return what user we're automatically modelling
 */
Say2User *
meko_status_menu_get_user(MekoStatusMenu *status_menu)
{
	Say2User *user;
	
	g_return_val_if_fail(status_menu != NULL, NULL);

	gtk_object_get(GTK_OBJECT(status_menu), "user", &user, NULL);

	return user;
}

/* meko_status_menu_get_text
 * Return what our current label is
 */
gchar *
meko_status_menu_get_text(MekoStatusMenu *status_menu)
{
	gchar *text;
	
	g_return_val_if_fail(status_menu != NULL, NULL);

	gtk_object_get(GTK_OBJECT(status_menu), "text", &text, NULL);

	return text;
}

/* meko_status_menu_set_value
 * Set the value we should currently display
 */
void
meko_status_menu_set_value(MekoStatusMenu *status_menu, Say2UserStatus status)
{
	GdkPixbuf *pixbuf;

	g_return_if_fail(status_menu != NULL);

	if (status_menu->pixmap) {
		gtk_container_remove(GTK_CONTAINER(status_menu->box),
				     status_menu->pixmap);
		status_menu->pixmap = NULL;
	}

	status_menu->status = status;

	/* Create the widget and pack it in */
	pixbuf = meko_create_status_image(status);
	if (pixbuf) {
		status_menu->pixmap = meko_create_pixbuf_pixmap(pixbuf, 128);
		if (status_menu->box) {
			gtk_box_pack_start(GTK_BOX(status_menu->box),
					   status_menu->pixmap,
					   FALSE, FALSE, 0);
		}
	}

	/* If this is a different status to that we were set to, signal */
	if (status_menu->status != status_menu->set_status) {
		status_menu->set_status = status_menu->status;
		meko_status_menu_changed(status_menu);
	}
}

/* meko_status_menu_get_value
 * Get what value we're currently displaying
 */
Say2UserStatus
meko_status_menu_get_value(MekoStatusMenu *status_menu)
{
	g_return_val_if_fail(status_menu != NULL, SAY2_USER_UNKNOWN);

	return status_menu->set_status;
}

/* _menu_detach_cb
 * Someone detached our popup menu!  No fair!
 */
static void
_menu_detach_cb(GtkWidget *widget, GtkMenu *menu)
{
	MekoStatusMenu *status_menu;
	
	g_return_if_fail(widget != NULL);
	g_return_if_fail(MEKO_IS_STATUS_MENU(widget));
	g_return_if_fail(menu != NULL);

	status_menu = MEKO_STATUS_MENU(widget);
	g_return_if_fail(status_menu->menu == GTK_WIDGET(menu));

	status_menu->menu = NULL;
	gtk_signal_disconnect_by_data(GTK_OBJECT(menu), widget);
}

/* _menu_deactivate_cb
 * Menu closed, so update ourselves
 */
static void
_menu_deactivate_cb(GtkWidget *widget, MekoStatusMenu *status_menu)
{
	Say2UserStatus  status;
	GtkWidget      *item;
	
	g_return_if_fail(widget != NULL);
	g_return_if_fail(GTK_MENU(widget));
	g_return_if_fail(status_menu != NULL);

	item = gtk_menu_get_active(GTK_MENU(widget));
	if (item == NULL)
		return;

	status = GPOINTER_TO_UINT(gtk_object_get_data(GTK_OBJECT(item),
						      "status"));
	if (status_menu->status != status) {
		/* Don't do a full update if we're bound to a user */
		if (status_menu->user != NULL) {
			status_menu->set_status = status;
			meko_status_menu_changed(status_menu);
		} else {
			meko_status_menu_set_value(status_menu, status);
		}
	}
}

/* _menu_position_cb
 * Set where we want the menu to appear
 */
static void
_menu_position_cb(GtkWidget *widget, gint *x, gint *y,
		  MekoStatusMenu *status_menu)
{
	g_return_if_fail(widget != NULL);
	g_return_if_fail(status_menu != NULL);

	gdk_window_get_origin(GTK_WIDGET(status_menu)->window, x, y);
	*y += GTK_WIDGET(status_menu)->allocation.height;
}

/* _user_changed_cb
 * See if the user's nickname or state changed, and if so update ourselves
 * accordinly
 */
static void
_user_changed_cb(Say2User *user, MekoStatusMenu *status_menu)
{
	Say2UserStatus  status;
	const gchar    *label;

	g_return_if_fail(user != NULL);
	g_return_if_fail(status_menu != NULL);
	g_return_if_fail(status_menu->user == user);

	label = meko_say2_client_label(SAY2_CLIENT(user));
	gtk_object_get(GTK_OBJECT(user), "status", &status, NULL);

	if ((status_menu->text == NULL) || strcmp(label, status_menu->text))
		meko_status_menu_set_text(status_menu, label);

	status_menu->set_status = status;	
	if (status != status_menu->status)
		meko_status_menu_set_value(status_menu, status);
}
